#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "cell.h"
#include "semantics.h"
#include "map.h"
#include "v1.h"

static int parse_arguments(int argc, char** argv,
			int* version,
			int* x, int* y,
			char** input, char** output);
static int parse_numeric(char* string, int* target, const char* name);

static int open_files(char* input, char* output, 
		FILE** finput, FILE** foutput);
static void close_files(FILE** finput, FILE** foutput);


extern struct cell_desc* first_cell;

int main(int argc, char** argv) {
	int dim[MAX_DIMS];
	int version;
	int empties;
	char* input = NULL;
	char* output = NULL;
	FILE* finput = NULL;
	FILE* foutput = NULL;
	struct cell_desc** map = NULL;
	
	if ( parse_arguments(argc,argv,&version,&dim[0],&dim[1],&input,&output) ) {
		return -1;
	}
	
	if ( open_files(input,output,&finput,&foutput) ) {
		return -1;
	}
	
	/* Front end */
	printf("PARSE O MATIC running %d %d\n",dim[0],dim[1]);
	/* Generate empty cell */
	push_cell();
	yyset_in(finput);
	if (yyparse()) {
		yylex_destroy();
		printf("\n");
		close_files(&finput,&foutput);
		return -1;
	}
	yylex_destroy();
	printf("\nDone\n");
	pop_cell(); /* Remove dummy headercell */
	/* Analysis */
	if ( semantic_analysis(first_cell,dim,version) != 0 ) {
		close_files(&finput,&foutput);
		free_cell_list(first_cell);
		free_map(&map);
		return -1;
	}
	print_cells();
	
	/* Backend */
	/* Generate grid */
	build_grid(first_cell,dim,&map);
	print_map(map,dim);
	if ( (empties = count_empties(map,dim)) != 0 ) {
		printf("Warning: %d empty cells detected\n",empties);
	}
	/* Translate */
	generate_result(map,dim,foutput,version);
	
	close_files(&finput,&foutput);
	free_cell_list(first_cell);
	free_map(&map);
	return 0;
}


static int parse_arguments(int argc, char** argv,
			int* version,
			int* x, int* y,
			char** input, char** output) {

	int ret = 0;
	
	if ( argc != 6 ) {
		printf("Usage: %s arch-version x-size y-size input-file output-file\n",argv[0]);
		return 1;
	}
	
	if ( (ret = parse_numeric(argv[1],version,"version")) != 0 ) {
		return ret;
	}
	if ( *version < 0 || *version > 2 ) {
		fprintf(stderr,"Unsupported version: %d\n",*version);
		return 1;
	}
	
	if ( (ret = parse_numeric(argv[2],x,"x")) != 0 ) {
		return ret;
	}
	if ( (ret =parse_numeric(argv[3],y,"y")) != 0 ) {
		return ret;
	}
	*input = argv[4];
	*output = argv[5];
	
	return 0;
}

static int parse_numeric(char* string, int* target, const char* name) {
	char* end = NULL;
	errno = 0;
	*target = (int)strtol(string,&end,10);
	
	if ( end == string || end != string + strlen(string)) {
		fprintf(stderr,"unable to parse %s\n",name);
		return 1;
	} else if ( errno == ERANGE || errno == EINVAL) {
		perror(name);
		return 1;
	}
	return 0;
}

static int open_files(char* input, char* output, 
		FILE** finput, FILE** foutput) {

	if ( input == NULL || output == NULL ) {
		fprintf(stderr,"Missing files\n");
		return 1;
	}

	*finput = fopen(input,"r");
	
	if ( *finput == NULL ) {
		fprintf(stderr,
			"Unable to open input file: %s\n", 
			strerror(errno)
		);
		return 1;
	}
	
	*foutput = fopen(output,"w");
	
	if ( *foutput == NULL ) {
		fprintf(stderr,
			"Unable to open output file: %s\n",
			strerror(errno)
		);
		fclose(*finput);
		return 1;
	}
	return 0;
}

static void close_files(FILE** finput, FILE** foutput) {
	
	if ( *finput != NULL ) {
		fclose(*finput);
		*finput = NULL;
	}
	if ( *foutput != NULL ) {
		fclose(*foutput);
		*finput = NULL;
	}
}
