#include <stdio.h>
#include <string.h>

#include "semantics.h"
#include "cell.h"
#include "defs.h"

static int alu_statement_analysis(struct statements* alu_stmnt);
static int md_statement_analysis(struct statements* md_stmnt);
static int mov_statement_analysis(struct statements* mov_stmnt);
static int init_statement_analysis(struct statements* init_stmnt);
static int simple_statement_analysis(struct statements* single);
static int statement_analysis(struct statements* list);
static int dup_statement_analysis(struct statements* list);
static int is_duplicate(struct statements* s1, struct statements* s2);
static int cell_analysis(struct cell_desc* first, int dim[]);
static int cell_dupcheck(struct cell_desc* first);
static int comp_pos(int dim1[], int dim2[] );
static int cell_pos_check(struct cell_desc* first, int dim[]);
static int bounds_check(const int bound[], int dim[]);
static int version_check(struct statements* list,int version);
static int statement_version(struct statements* s);
static int branch_statement_analysis(struct statements* branch_stmnt);
static int switch_statement_analysis(struct statements* switch_stmnt);
static int select_statement_analysis(struct statements* overwrite_stmnt);
static int validwrite_statement_analysis(struct statements* valid_stmnt);
static int writepolicy_statement_analysis(struct statements* wrp_stmnt);
static int is_output(enum operands op);
static int is_input(enum operands op);
static int branch_block_analysis(struct statements* parent);
static int input_branch_analysis(struct statements* parent);
static int output_branch_analysis(struct statements* parent);
static int is_output_branch(struct statements* smnt);
static int is_input_branch(struct statements* smnt);


int semantic_analysis(struct cell_desc* first, int dim[],int version) {
	int ret = 0;
	struct cell_desc* next = first;
	if ( (ret = cell_analysis(first,dim)) != 0 ) {
		return ret;
	}
	while ( next != NULL ) {
		printf("Analysing cell: %d %d\n",next->pos[0],next->pos[1]);
		if ( (ret = statement_analysis(next->list)) != 0 ) {
			return ret;
		}
		if ( (ret = version_check(next->list,version)) != 0 ) {
			return ret;
		}
		next = next->next;
	}
	return ret;
}

static int cell_analysis(struct cell_desc* first, int dim[]) {
	int ret = 0;
	if ( (ret = cell_dupcheck(first)) != 0 ) {
		return ret;
	}
	if ( (ret = cell_pos_check(first,dim)) != 0 ) {
		return ret;
	}
	
}

static int cell_dupcheck(struct cell_desc* first) {
	struct cell_desc* current = first;
	struct cell_desc* l2_cur = NULL;
	while ( current != NULL ) {
		l2_cur = current->next;
		while ( l2_cur != NULL ) {
			if ( comp_pos(current->pos,l2_cur->pos) ) {
				fprintf(stderr,"Duplicate cells found. Pos: ");
				print_cell_pos(stderr,current);
				fprintf(stderr,"\n");
				return 1;
			}
			l2_cur = l2_cur->next;
		}
		current = current->next;
	}
	return 0;
}

static int comp_pos(int dim1[], int dim2[]) {
	int i = 0;
	for ( i = 0 ; i < MAX_DIMS ; i++ ) {
		if ( dim1[i] - dim2[i] != 0) {
			return 0;
		}
	}
	return 1;
}

static int cell_pos_check(struct cell_desc* first, int dim[]) {
	struct cell_desc* current = first;
	while ( current != NULL ) {
		if ( bounds_check(dim,current->pos) ) {
			fprintf(stderr,"Out of bounds cell. Pos: ");
			print_cell_pos(stderr,current);
			fprintf(stderr,"\n");
			return 1;
		}
		current = current->next;
	}
	return 0;
}


static int bounds_check(const int bound[], int dim[]) {
	int edge[MAX_DIMS];
	int i;
	memcpy(edge,bound,sizeof(int)*MAX_DIMS);
	for ( i = 0 ; i < MAX_DIMS; i++ ) {
		edge[i] -= 1;
	}
	for ( i = 0 ; i < MAX_DIMS ; i++ ) {
		if ( edge[i] - dim[i] < 0) {
			return 1;
		}
	}
	return 0;
}

static int version_check(struct statements* list,int version) {
	struct statements* s = list;
	int ret = 0;
	
	while ( s != NULL ) {
		ret = statement_version(s);
		
		if ( ret > version || ret == 0 ) {
			fprintf(stderr,"Statement not supported by version: %d\n",version);
			print_statement(stderr,s);
			return 1;
		}
		s->version = ret;
		if ( s->children != NULL ) {
			version_check(s->children,version);
		}
		s = s->next;
	}
	return 0;
}


static int statement_version(struct statements* s) {
	switch (s->stype) {
	case s_init:
	case s_add:
	case s_sub:
	case s_mul:
	case s_div:
	case s_slr:
	case s_sll:
	case s_and:
	case s_or:
	case s_nor:
	case s_xor:
	case s_mov:
		return 1;
		break;
	
	case s_branch:
	case s_switch:
	case s_select:
	case s_validwrite:
	case s_writepolicy:
		return 2;
		break;
	default:
		fprintf(stderr,"Unknown statement version\n");
		return 0;
	}
}

static int statement_analysis(struct statements* list) {
	int ret = 0;
	struct statements* next = list;
	if ( (ret=dup_statement_analysis(list)) != 0 ) {
		return ret;
	}
	while ( next != NULL ) {
		if ( (ret = statement_analysis(next->children) ) != 0 ) {
			return ret;
		}
		if ( (ret = simple_statement_analysis(next)) != 0 ) {
			return ret;
		}
		if ( next->stype == s_branch && (ret = branch_block_analysis(next)) != 0 ) {
			return ret;
		}
		next = next->next;
	}
	return ret;
}

static int dup_statement_analysis(struct statements* list) {
	struct statements* current = list;
	struct statements* l2_cur = NULL;
	
	while ( current != NULL ) {
		l2_cur = current->next;
		while ( l2_cur != NULL ) {
			/*if ( current->target == l2_cur->target 
				&& !(current->stype == s_init ^ l2_cur->stype == s_init)
				&& !(
					(current->stype == s_branch ^ l2_cur->stype == s_branch) 
					|| (current->flag != l2_cur->flag)
				)
			) */
			if ( is_duplicate(current, l2_cur) ) 
			{
				
				fprintf(stderr, 
					"Duplicate target assignment %s %s %s %s %s\n",
					str_op(current->target),
					str_type(current->stype),
					str_type(l2_cur->stype),
					str_flag(current->flag),
					str_flag(l2_cur->flag)
				);
				return 1;
			}
			l2_cur = l2_cur->next;
		}
		current = current->next;
	}
	return 0;
}

static int is_duplicate(struct statements* s1, struct statements* s2) {
	if ( s1->stype == s_switch && s2->stype == s_switch ) {
		if ( (s1->op1 == s2->op1 || s1->op1 == s2->op2 )
			&& (s1->op2 == s2->op1 || s1->op2 == s2->op2)
		) {
			return 1;
		} else {
			return 0;
		}
	} else if ( s1->target == s2->target ) {
		/* Not a duplicate if only one of the statements is a init statement */
		if ( s1->stype == s_init ^ s2->stype == s_init ) {
			return 0;
		/* Not a duplicate if only one of the statements is a branch statement */
		} else if ( s1->stype == s_branch ^ s2->stype == s_branch ) {
			return 0;
		} else if ( s1->stype == s_branch && s2->stype == s_branch ) {
			/* Two branches */
			if ( s1->flag == br_valid && s2->flag == br_valid ) {
				/* Both branches are valid branches with the same target */
				/* Check invert flag */
				if ( s1->data != s2->data ) {
					return 0;
				} else {
					return 1;
				}
			} else if (s1->flag != s2->flag ) {
				/* Not using the same flag */
				return 0;
			} else {
				return 1;
			}
		/* Not a duplicate if only one of the statements is a write policy statement */
		} else if (s1->stype == s_writepolicy ^ s2->stype == s_writepolicy ) {
			return 0;
		/* Not a duplicate if only one of the statements is a valid policy statement */
		} else if (s1->stype == s_validwrite ^ s2->stype == s_validwrite ) {
			return 0;
		} else {
			return 1;
		}
	}
	return 0;
}


static int simple_statement_analysis(struct statements* single) {
	
	switch(single->stype) {
	/* ALU ops */
	case s_add:
	case s_sub:
	case s_slr:
	case s_sll:
	case s_and:
	case s_or:
	case s_nor:
	case s_xor:
		return alu_statement_analysis(single);
		break;
	case s_mul:
	case s_div:
		return md_statement_analysis(single);
		break;
	case s_mov:
		return mov_statement_analysis(single);
		break;
	case s_init:
		return init_statement_analysis(single);
		break;
	case s_branch:
		return branch_statement_analysis(single);
		break;
	case s_switch:
		return switch_statement_analysis(single);
		break;
	case s_select:
		return select_statement_analysis(single);
		break;
	case s_validwrite:
		return validwrite_statement_analysis(single);
		break;
	case s_writepolicy:
		return writepolicy_statement_analysis(single);
		break;
	default:
		fprintf(stderr,"Invalid statement type: %d\n",single->stype);
		return 1;
		break;
	}
	return 0;
}

static int alu_statement_analysis(struct statements* alu_stmnt) {
	if (alu_stmnt == NULL ) {
		fprintf(stderr,
			"Internal error. Null pointer passed to alu analysis\n"
		);
		return -1;
	}

	if ( alu_stmnt->target != alu_out ) {
		fprintf(stderr,"Invalid target for alu operation\n");
		return 1;
	}
	if ( alu_stmnt->op1 == alu_out || alu_stmnt->op2 == alu_out ) {
		fprintf(stderr,"Direct alu feedback not allowed\n");
		return 2;
	}
	if ( alu_stmnt->op1 == unknown || alu_stmnt->op2 == unknown ) {
		fprintf(stderr,
			"Unknown operand not allowed on alu operations\n"
		);
		return 3;
	}
	return 0;
}

static int md_statement_analysis(struct statements* md_stmnt) {
	if (md_stmnt == NULL ) {
		fprintf(stderr,
			"Internal error. Null pointer passed to mul analysis\n"
		);
		return -1;
	}

	if ( md_stmnt->target != mul_out ) {
		fprintf(stderr,"Invalid target for mul operation\n");
		return 1;
	}
	if ( md_stmnt->op1 == mul_out || md_stmnt->op2 == mul_out ) {
		fprintf(stderr,"Direct mul feedback not allowed\n");
		return 2;
	}
	if ( md_stmnt->op1 == unknown || md_stmnt->op2 == unknown ) {
		fprintf(stderr,
			"Unknown operand not allowed on mul operations\n"
		);
		return 3;
	}
	return 0;
}

static int mov_statement_analysis(struct statements* mov_stmnt) {
	if ( !is_output(mov_stmnt->target) ) {
		fprintf(stderr,"Invalid target for mov operation\n");
		return 1;
	}
	switch ( mov_stmnt->op1 ) {
	case north:
	case west:
	case alu_out:
	case mul_out:
	case state:
		break;
	
	default:
		fprintf(stderr,"Invalid source for mov operation\n");
		return 2;
	}
	return 0;
}

static int init_statement_analysis(struct statements* init_stmnt) {
	if ( init_stmnt->target != state ) {
		fprintf(stderr,"Invalid target for init statement\n");
		return 1;
	}
	return 0;
}

static int branch_statement_analysis(struct statements* branch_stmnt) {
	if ( branch_stmnt->target != alu_out && !is_input(branch_stmnt->target)) {
		fprintf(stderr,"Invalid branch expression\n");
		return 1;
	}
	
	switch ( branch_stmnt->flag ) {
	case br_valid:
		if ( !is_input(branch_stmnt->target) ) {
			fprintf(stderr,"Can not perform branch select on non-input\n");
			return 2;
		} 
		break;
	case br_zero:
	case br_negative:
		if ( branch_stmnt->target != alu_out ) {
			fprintf(stderr,"Can not branch upon non-flag source\n");
			return 2;
		}
		
		break;
	default:
		fprintf(stderr,"Unknown branch expression\n");
		return 0;
	}
	
	
	return 0;
}

static int switch_statement_analysis(struct statements* switch_stmnt) {
	if ( !is_output(switch_stmnt->op1) || !is_output(switch_stmnt->op2)) {
		fprintf(stderr,"Trying to switch non-output targets\n");
		return 1;
	} else if ( switch_stmnt->op1 == switch_stmnt->op2 ) {
		fprintf(stderr,"Unable to switch a target with itself\n");
		return 2;
	}
	
	return 0;
}

static int select_statement_analysis(struct statements* select_stmnt) {
	if ( !is_input(select_stmnt->target) || !is_input(select_stmnt->op1) ) {
		fprintf(stderr,"Invalid input branch\n");
		return 1;
	}
	return 0;
}

static int validwrite_statement_analysis(struct statements* valid_stmnt) {
	if ( !is_output(valid_stmnt->target) || (valid_stmnt->op1 != alu_out && valid_stmnt->flag != br_never)) {
		fprintf(stderr,"Invalid output valid target\n");
		return 1;
	}
	return 0;
}

static int writepolicy_statement_analysis(struct statements* wrp_stmnt) {
	if ( wrp_stmnt->target != state ) {
		fprintf(stderr,"Invalid target for write policy change\n");
		return 1;
	}
	return 0;
}


static int is_output(enum operands op) {
	return (op == south || op == east || op == state );
}

static int is_input(enum operands op) {
	return (op == north || op == west);
}


static int branch_block_analysis(struct statements* parent) {
	if ( parent->stype != s_branch ) {
		/* Not a branch block */
		return 0;
	}
	
	if ( is_output_branch(parent) ) {
		/* children must be switch statements */
		return output_branch_analysis(parent);
	} else if ( is_input_branch(parent) ) {
		/* children must be overwrite statements */
		return input_branch_analysis(parent);
	} else {
		fprintf(stderr,"Unknown branch block\n");
		return 1;
	}
	

	return 0;
}

static int is_output_branch(struct statements* stmnt) {
	return (stmnt->target == alu_out);
}

static int is_input_branch(struct statements* stmnt) {
	return (stmnt->flag == br_valid);
}

static int input_branch_analysis(struct statements* parent) {
	struct statements* children = parent->children;
	
	while ( children != NULL ) {
		if ( children->stype != s_select ) {
			fprintf(stderr,"Non input overwrite statement in branch block\n");
			return 1;
		} else if ( parent->target == children->target && parent->data != invert) {
			fprintf(stderr,"Non invert branch cannot contain such operations\n");
			return 2;
		} else if ( parent->target != children->target && parent->data != no_invert ) {
			fprintf(stderr,"Invert branch cannot contain such operations\n");
			return 3;
		}
		
		children = children->next;
	}
	return 0;
}

static int output_branch_analysis(struct statements* parent) {
	struct statements* children = parent->children;
	while ( children != NULL ) {
		if ( children->stype != s_switch ) {
			fprintf(stderr,"Non output switch statement in branch block\n");
			return 1;
		}
		children = children->next;
	}
	return 0;
}


