%{
#include <stdio.h>
#include "defs.h"
#include "cell.h"
#define YYERROR_VERBOSE 1

extern char yytext[];
extern int yylineno;
extern struct cell_desc* first_cell;

void yyerror(const char* str) {
	fprintf(stderr,"Error %d: %s\n",yylineno,str);
}

int yywrap() {
	return 1;
}

%}
%union {
	struct statements* s;
	int value;
}

%token CELL INIT IF 
%token NORTH WEST EAST SOUTH
%token ALU_OUT MUL_OUT STATE
%token SWITCH VALID NEGATIVE ZERO NOT WRITE ALWAYS NEVER
%token AND OR NOR XOR ADD SUB SLR SLL MUL DIV
%token DIGITS

%type<s> initialization assignment statement statement_list 
%type<s> branch branch_statement branch_statement_list branch_expression valid write
%type<value> op integer number source target outputs inputs flags flag_source write_policy

%%
description: cell_list

cell_list: 
	cell
	| cell_list cell 
	;

cell:
	CELL dim_list '{' statement_list '}' {
					first_cell->list = $4;
					push_cell();
				}
	;

statement_list:
	statement	{ $$ = $1; }
	| statement_list statement { 
				$2->next = $1; 
				$$ = $2;
			}
	;

statement:
	initialization	{ $$ = $1 }
	| assignment	{ $$ = $1 }
	| branch	{ $$ = $1 }
	| valid		{ $$ = $1 }
	| write		{ $$ = $1 }
	;

initialization:
	INIT '{' STATE '=' number ';' '}' {
				struct statements* s = new_statement();
				s->stype = s_init;
				s->target = state;
				s->data = $5;
				$$ = s;
			}
	;

assignment:
	target '=' source op source ';' {
				struct statements* s = new_statement();
				s->stype = $4;
				s->op1 = $3;
				s->op2 = $5;
				s->target = $1;
				$$ = s;
			}
	| target '=' source ';' {
				struct statements* s = new_statement();
				s->stype = s_mov;
				s->op1 = $3;
				s->target = $1;
				$$ = s;
			}
	;

branch:
	IF '(' branch_expression ')' '{' branch_statement_list '}' {
				$3->children = $6;
				$$ = $3;
			}
	;

branch_expression:
	flags '(' flag_source ')'	{
				struct statements* s = new_statement();
				s->stype = s_branch;
				s->target = alu_out;
				s->flag = $1;
				$$ = s;
			}
	| NOT VALID '(' inputs ')'	{
				struct statements* s = new_statement();
				s->stype = s_branch;
				s->flag = br_valid;
				s->target = $4;
				s->data = invert;
				$$ = s;
			}
	| VALID	'(' inputs ')'		{
				struct statements* s = new_statement();
				s->stype = s_branch;
				s->flag = br_valid;
				s->target = $3;
				s->data = no_invert;
				$$ = s;
			}
	;

branch_statement_list:
	branch_statement	{ $$ = $1; }
	| branch_statement_list branch_statement {
				$2->next = $1;
				$$ = $2;
			}
	;

branch_statement:
	SWITCH '(' outputs ',' outputs ')' ';' {
				struct statements* s = new_statement();
				s->stype = s_switch;
				s->op1 = $3;
				s->op2 = $5;
				$$ = s;
			}
	| inputs '=' inputs ';' {
				struct statements* s = new_statement();
				s->stype = s_select;
				s->target = $1;
				s->op1 = $3;
				$$ = s;
			}
	;

valid:
	VALID '(' outputs ')' '=' flags '(' flag_source ')' ';' {
				struct statements* s = new_statement();
				s->stype = s_validwrite;
				s->target = $3;
				s->op1 = $8;
				s->flag = $6;
				$$ = s;
			}
	| VALID '(' outputs ')' '=' NEVER ';' {
				struct statements* s = new_statement();
				s->stype = s_validwrite;
				s->target = $3;
				s->flag = br_never;
				$$ = s;
			}
	;

write:
	WRITE '(' STATE ')' '=' write_policy ';' {
				struct statements* s = new_statement();
				s->stype = s_writepolicy;
				s->target = state;
				s->wr_flag = $6;
				$$ = s;
			}
	;

write_policy:
	VALID		{ $$ = wr_valid; }
	| ALWAYS	{ $$ = wr_always; }
	;

target:
	outputs		{ $$ = $1; }
	| ALU_OUT	{ $$ = alu_out; }
	| MUL_OUT	{ $$ = mul_out; }
	;

op:
	AND	{ $$ = s_and; }
	| OR	{ $$ = s_or;  }
	| XOR	{ $$ = s_xor; }
	| NOR	{ $$ = s_nor; }
	| ADD	{ $$ = s_add; }
	| SUB	{ $$ = s_sub; }
	| DIV	{ $$ = s_div; }
	| MUL	{ $$ = s_mul; }
	| SLL	{ $$ = s_sll; }
	| SLR	{ $$ = s_slr; }
	;

source:
	inputs		{ $$ = $1; }
	| ALU_OUT	{ $$ = alu_out; }
	| STATE		{ $$ = state; }
	| MUL_OUT	{ $$ = mul_out; }
	;

inputs:
	NORTH		{ $$ = north; }
	| WEST		{ $$ = west; }
	;

outputs:
	STATE		{ $$ = state; }
	| EAST		{ $$ = east; }
	| SOUTH		{ $$ = south; }
	;

flags:
	NEGATIVE	{ $$ = br_negative; }
	| ZERO		{ $$ = br_zero; }
	;

flag_source:
	ALU_OUT		{ $$ = alu_out; }
	;

dim_list:
	dim
	| dim_list dim
	;

dim: 
	'[' integer ']' { 
			if (first_cell->dims < MAX_DIMS) {
				first_cell->pos[first_cell->dims++] = $2; 
			} else {
				yyerror("Unsupported dimension\n");
				YYERROR;
			}
		}
	;

number:
	SUB integer	{ $$ = -1*$2; }
	| integer	{ $$ = $1; }
	;

integer:
	DIGITS { $$ = strtol(yytext, NULL, 10); }
	;


%%