#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "v1.h"
#include "defs.h"
#include "bits.h"

static int convert_statement(struct v1_part* target, struct statements* s);
static int alu_statement(struct v1_part* target, struct statements* s);
static int alu_mux_to_bits(int32_t* target, enum operands op);
static int alu_op_to_bits(int32_t* target, enum statement_type op);

static int md_statement(struct v1_part* target, struct statements* s);
static int md_mux_to_bits(int32_t* target, enum operands op);
static int md_op_to_bits(int32_t* target, enum statement_type op);

static int mov_statement(struct v1_part* target , struct statements* s);
static int output_mux_to_bits(int32_t* target, enum statement_type op);

static int init_statement(struct v1_part* target, struct statements* s);


static int new_v1_part(struct v1_part** target){
	
	*target = malloc(sizeof(struct v1_part));
	
	if ( *target == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		return 1;
	}
	memset(*target,INVALID_SIGN,sizeof(struct v1_part));
	(*target)->init_state = 0;
	
	return 0;
}

int destroy_v1_part(struct v1_part** target) {
	free(*target);
	*target = NULL;
}


int generate_v1_part(struct v1_part** target, struct statements* list) {
	struct statements* current = list;
	int res = 0;
	
	if ( (res = new_v1_part(target)) != 0 ) {
		return res;
	}
	
	while ( current != NULL ) {
		if ( current->version > 1 ) {
			current = current->next;
			continue;
		}
		
		if ( (res = convert_statement(*target,current) ) != 0 ) {
			return res;
		}
		current = current->next;
	}
	return 0;
}


static int convert_statement(struct v1_part* target, struct statements* s) {

	switch (s->stype) {
	case s_add:
	case s_sub:
	case s_slr:
	case s_sll:
	case s_or:
	case s_nor:
	case s_xor:
	case s_and:
		return alu_statement(target,s);
		break;
	case s_mul:
	case s_div:
		return md_statement(target,s);
		break;
	case s_mov:
		return mov_statement(target,s);
		break;
	case s_init:
		return init_statement(target,s);
		break;
	
	default:
		fprintf(stderr,
			"Invalid statement found. Building for an old version?\n"
		);
		return 1;
	}
	return 0;
}

static int alu_statement(struct v1_part* target, struct statements* s) {

	if ( s->op1 == unknown || s->op2 == unknown || s->target != alu_out) {
		fprintf(stderr,"Internal error. Invalid ALU statement\n");
		return 1;
	}
	
	if (	target->alu_input_a != INVALID_SIGN ||
		target->alu_input_b != INVALID_SIGN ||
		target->alu_op != INVALID_SIGN ) {
		fprintf(stderr,
			"Tried to allocate more than one task to ALU\n"
		);
		return 2;
	}
	
	if ( alu_mux_to_bits(&(target->alu_input_a),s->op1) ) {
		return 3;
	}
	if ( alu_mux_to_bits(&(target->alu_input_b),s->op2) ) {
		return 3;
	}
	if ( alu_op_to_bits(&(target->alu_op),s->stype) ) {
		return 4;
	}
	return 0;
}

static int alu_mux_to_bits(int32_t* target, enum operands op) {
	switch ( op ) {
	case north:
		*target = V1_ALU_MUX_NORTH;
		break;
	case west:
		*target = V1_ALU_MUX_WEST;
		break;
	case mul_out:
		*target = V1_ALU_MUX_MULOUT;
		break;
	case state:
		*target = V1_ALU_MUX_STATE;
		break;
	default:
		fprintf(stderr,"Invalid ALU mux source\n");
		return 1;
	}
	return 0;
} 

static int alu_op_to_bits(int32_t* target, enum statement_type op) {
	switch ( op ) {
	case s_add:
		*target = V1_ALU_OP_ADD;
		break;
	case s_sub:
		*target = V1_ALU_OP_SUB;
		break;
	case s_slr:
		*target = V1_ALU_OP_SLR;
		break;
	case s_sll:
		*target = V1_ALU_OP_SLL;
		break;
	case s_or:
		*target = V1_ALU_OP_OR;
		break;
	case s_nor:
		*target = V1_ALU_OP_NOR;
		break;
	case s_and:
		*target = V1_ALU_OP_AND;
		break;
	case s_xor:
		*target = V1_ALU_OP_XOR;
		break;
	
	default:
		fprintf(stderr,"Invalid ALU op\n");
		return 1;
	}
	return 0;
}

static int md_statement(struct v1_part* target, struct statements* s) {

	if ( s->op1 == unknown || s->op2 == unknown || s->target != mul_out) {
		fprintf(stderr,"Internal error. Invalid MUL statement\n");
		return 1;
	}
	
	if (	target->md_input_a != INVALID_SIGN ||
		target->md_input_b != INVALID_SIGN ||
		target->md_op != INVALID_SIGN ) {
		fprintf(stderr,
			"Tried to allocate more than one task to MUL\n"
		);
		return 2;
	}
	
	if ( md_mux_to_bits(&(target->md_input_a),s->op1) ) {
		return 3;
	}
	if ( md_mux_to_bits(&(target->md_input_b),s->op2) ) {
		return 3;
	}
	if ( md_op_to_bits(&(target->md_op),s->stype) ) {
		return 4;
	}
	return 0;
}

static int md_mux_to_bits(int32_t* target, enum operands op) {
	switch ( op ) {
	case north:
		*target = V1_MD_MUX_NORTH;
		break;
	case west:
		*target = V1_MD_MUX_WEST;
		break;
	case alu_out:
		*target = V1_MD_MUX_ALUOUT;
		break;
	case state:
		*target = V1_MD_MUX_STATE;
		break;
	default:
		fprintf(stderr,"Invalid MUL mux source\n");
		return 1;
	}
	return 0;
} 

static int md_op_to_bits(int32_t* target, enum statement_type op) {
	switch ( op ) {
	case s_mul:
		*target = V1_MD_OP_MUL;
		break;
	case s_div:
		*target = V1_MD_OP_DIV;
		break;
	default:
		fprintf(stderr,"Invalid MUL op\n");
		return 1;
	}
	return 0;
}


static int mov_statement(struct v1_part* target , struct statements* s) {

	switch(s->target) {
	case east:
		if ( target->east_output != INVALID_SIGN ) {
			fprintf(stderr, 
				"Tried to allocate more than one signal to east\n"
			);
			return 2;
		}
		return output_mux_to_bits(&(target->east_output),s->op1);
		break;
	case south:
		if ( target->south_output != INVALID_SIGN ) {
			fprintf(stderr, 
				"Tried to allocate more than one signal to south\n"
			);
			return 2;
		}
		return output_mux_to_bits(&(target->south_output),s->op1);
		break;
	case state:
		if ( target->state_input != INVALID_SIGN ) {
			fprintf(stderr, 
				"Tried to allocate more than one signal to state\n"
			);
			return 2;
		}
		return output_mux_to_bits(&(target->state_input),s->op1);
		break;
		
	default:
		fprintf(stderr,"Internal error. Invalid mov statement\n");
		return 1;
	
	}
	
	
	return 0;
	
}


static int output_mux_to_bits(int32_t* target, enum statement_type op) {
	switch ( op ) {
	case north:
		*target = V1_OUT_MUX_NORTH;
		break;
	case west:
		*target = V1_OUT_MUX_WEST;
		break;
	case alu_out:
		*target = V1_OUT_MUX_ALUOUT;
		break;
	case mul_out:
		*target = V1_OUT_MUX_MULOUT;
		break;
	case state:
		*target = V1_OUT_MUX_STATE;
		break;
	default:
		fprintf(stderr,"Invalid output source\n");
		return 1;
	}
	return 0;
}


static int init_statement(struct v1_part* target, struct statements* s) {
	if ( s->target != state ) {
		fprintf(stderr,"Unable to initialize non-state component\n");
		return 1;
	}
	/* Include valid bit */
	target->init_state = (1<<PATH_SIZE)| ~((~0)<<PATH_SIZE)&s->data;
	
	return 0;
}

void print_v1_part(struct v1_part* source) {
	if ( source == NULL ) {
		return;
	}
	printf("\nV1 CONFIG:\n");
	printf("ALU MUX A: 0x%x\n",source->alu_input_a);
	printf("ALU MUX B: 0x%x\n",source->alu_input_b);
	printf("ALU OP   : 0x%x\n",source->alu_op);
	printf("MD MUX A : 0x%x\n",source->md_input_a);
	printf("MD MUX B : 0x%x\n",source->md_input_b);
	printf("MD OP    : 0x%x\n",source->md_op);
	printf("EAST MUX : 0x%x\n",source->east_output);
	printf("SOUTH MUX: 0x%x\n",source->south_output);
	printf("STATE MUX: 0x%x\n",source->state_input);
	printf("STATE    : 0x%x\n",source->init_state);
	printf("---------------\n");
}


void serialize_v1_part(struct v1_part* source, char* target_buffer, int version) {
	int buffer_pos = 0;
	pos_update_wrapper(source->alu_input_a,	V1_ALU_INPUT_A_BITS,	target_buffer,&buffer_pos);
	pos_update_wrapper(source->alu_input_b,	V1_ALU_INPUT_B_BITS,	target_buffer,&buffer_pos);
	pos_update_wrapper(source->alu_op,	V1_ALU_OP_BITS,		target_buffer,&buffer_pos);
	pos_update_wrapper(source->md_input_a,	V1_MD_INPUT_A_BITS,	target_buffer,&buffer_pos);
	pos_update_wrapper(source->md_input_b,	V1_MD_INPUT_B_BITS,	target_buffer,&buffer_pos);
	pos_update_wrapper(source->md_op,	V1_MD_OP_BITS,		target_buffer,&buffer_pos);
	pos_update_wrapper(source->state_input,	V1_STATE_INPUT_BITS,	target_buffer,&buffer_pos);
	pos_update_wrapper(source->east_output,	V1_EAST_OUTPUT_BITS,	target_buffer,&buffer_pos);
	pos_update_wrapper(source->south_output,V1_SOUTH_OUTPUT_BITS,	target_buffer,&buffer_pos);
	pos_update_wrapper(source->init_state,	V1_INIT_STATE_BITS,	target_buffer,&buffer_pos);
}

