#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "v2.h"
#include "v1.h"
#include "defs.h"
#include "bits.h"

static int convert_statement(struct v2_part* target, struct statements* s);
static int alu_statement(struct v1_part* target, struct statements* s);
static int alu_mux_to_bits(int32_t* target, enum operands op);
static int alu_op_to_bits(int32_t* target, enum statement_type op);

static int md_statement(struct v1_part* target, struct statements* s);
static int md_mux_to_bits(int32_t* target, enum operands op);
static int md_op_to_bits(int32_t* target, enum statement_type op);

static int mov_statement(struct v1_part* target , struct statements* s);
static int output_mux_to_bits(int32_t* target, enum statement_type op);

static int init_statement(struct v1_part* target, struct statements* s);

static int branch_statement(struct v2_part* target, struct statements *s);
static int convert_branch_statement(struct v2_part* target, struct statements* source, 
							struct statements *s);
static int switch_statement(struct v2_part* target, struct statements* source,
			struct statements* s);
static int switch_to_bits(int32_t* target, struct statements* source);
static int select_statement(struct v2_part* target, struct statements* source,
			struct statements* s);
static int select_to_bits(int32_t* target, enum branch_flags br, enum invert inv );


static int validwrite_statement(struct v2_part* target, struct statements* s );
static int valid_out_to_bits(int32_t* target, enum branch_flags flag);
static int write_policy_statement(struct v2_part* target, struct statements *s);


static int new_v2_part(struct v2_part** target){
	
	*target = malloc(sizeof(struct v2_part));
	
	if ( *target == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		return 1;
	}
	memset(*target,INVALID_SIGN,sizeof(struct v2_part));
	(*target)->v1 = NULL;
	
	return 0;
}

int destroy_v2_part(struct v2_part** target) {
	destroy_v1_part(&((*target)->v1));
	free(*target);
	*target = NULL;
}


int generate_v2_part(struct v2_part** target, struct statements* list, int version) {
	struct statements* current = list;
	int res = 0;
	
	if ( (res = new_v2_part(target)) != 0 ) {
		return res;
	}
	if ( version >= 2 ) {
		while ( current != NULL ) {
			if ( version > 2 || current->version != 2) {
				current = current->next;
				continue;
			}
			
			if ( (res = convert_statement(*target,current) ) != 0 ) {
				return res;
			}
			current = current->next;
		}
	}
	if ( ( res = generate_v1_part(&((*target)->v1),list)) != 0 ) {
		return res;
	}
	
	return 0;
}


static int convert_statement(struct v2_part* target, struct statements* s) {
	
	switch (s->stype) {
	case s_branch:
		return branch_statement(target,s);
		break;
	case s_validwrite:
		return validwrite_statement(target,s);
		break;
	case s_writepolicy:
		return write_policy_statement(target,s);
		break;
	
	default:
		fprintf(stderr,
			"Invalid statement found. Building for an old version?\n"
		);
		print_statement(stderr,s);
		return 1;
	}
	return 0;
}

static int branch_statement(struct v2_part* target, struct statements *s) {
	struct statements* list = s->children;
	
	while ( list != NULL ) {
		
		convert_branch_statement(target,s,list);
		list = list->next;
	}
	
	return 0;
}

static int convert_branch_statement(struct v2_part* target, 
			struct statements* source, struct statements *s) {
	
	/* TODO: Fix if here */
	
	switch ( s->stype ) {
	case s_switch:
		return switch_statement(target,source,s);
		break;
	case s_select:
		return select_statement(target,source,s);
		break;
	default:
		fprintf(stderr,"Invalid branch statement found. \n");
		print_statement(stderr,s);
		return 1;
	}
	return 0;
}

static int switch_statement(struct v2_part* target, struct statements* source,
			struct statements* s) {
	if ( source->target != alu_out ) {
		fprintf(stderr,"Non alu source in switch statement is not supported\n");
		print_statement(stderr,s);
		return 1;
	}
	
	switch ( s->op1 ) {
	case east:
		if ( s->op2 == state ) {
			return switch_to_bits(&(target->east_state_switch),source);
		} else if ( s->op2 == south ) {
			return switch_to_bits(&(target->south_east_switch),source);
		} else {
			fprintf(stderr,"Invalid operand in switch statement.\n");
			return 2;
		}
	
		break;
	case state:
		if ( s->op2 == east ) {
			return switch_to_bits(&(target->east_state_switch),source);
		} else if ( s->op2 == south ) {
			return switch_to_bits(&(target->state_south_switch),source);
		} else {
			fprintf(stderr,"Invalid operand in switch statement.\n");
			return 2;
		}
		break;
	case south:
		if ( s->op2 == state ) {
			return switch_to_bits(&(target->state_south_switch),source);
		} else if ( s->op2 == east ) {
			return switch_to_bits(&(target->south_east_switch),source);
		} else {
			fprintf(stderr,"Invalid operand in switch statement.\n");
			return 2;
		}
		break;
	default:
		fprintf(stderr,"Unknown operand in switch statement.\n");
		print_statement(stderr,s);
		return 1;
	}
	return 0;
}

static int switch_to_bits(int32_t* target, struct statements* source) {
	if ( *target != INVALID_SIGN ) {
		fprintf(stderr,"Dual switch assignment in block\n");
		print_statement(stderr,source);
		return 1;
	}
	
	
	switch (source->flag) {
	case br_zero:
		*target = V2_SWITCH_ZERO;
		break;
	case br_negative:
		*target = V2_SWITCH_NEGATIVE;
		break;
	default:
		fprintf(stderr,"Unknown switch flag\n");
		return 1;
	}
	return 0;
}

static int select_statement(struct v2_part* target, struct statements* source,
			struct statements* s) {
	
	switch(s->target) {
	case north:
		return select_to_bits(&(target->north_select),source->flag,source->data);
		break;
	case west:
		return select_to_bits(&(target->west_select),source->flag,source->data);
		break;
	default:
		fprintf(stderr,"Unknown select target\n");
		return 1;
	}
	return 0;
}

static int select_to_bits(int32_t* target, enum branch_flags br, enum invert inv ) {
	
	if ( br != br_valid ) {
		fprintf(stderr,"Invalid select statement (flag)\n");
		return 1;
	}
	if ( *target != INVALID_SIGN ) {
		fprintf(stderr,"Two select statements on same target\n");
		return 2;
	}
	
	/* TODO: Check if this is correct */
	if ( inv == invert ) {
		*target = V2_IN_SELECT_INV;
		return 0;
	} else if ( inv == no_invert ) {
		*target = V2_IN_SELECT_OTHER;
		return 0;
	} else {
		fprintf(stderr,"Invalid select statement (invert)\n");
		return 2;
	}
	
}


static int validwrite_statement(struct v2_part* target, struct statements* s ) {
	if ( s->target != east && s->target != south && s->target != state ) {
		fprintf(stderr,"Internal error. Invalid valid statement\n");
		return 1;
	}
	/* A bit redudant */
	switch ( s->target) {
	case east:
		return valid_out_to_bits(&(target->east_valid),s->flag);
	case south:
		return valid_out_to_bits(&(target->south_valid),s->flag);
	case state:
		return valid_out_to_bits(&(target->state_valid),s->flag);
	default:
		fprintf(stderr,"Invalid valid statement\n");
		return 1;
	}
	return 0;
	
}

static int valid_out_to_bits(int32_t* target, enum branch_flags flag) {
	if ( *target != INVALID_SIGN ) {
		fprintf(stderr,"Tried to set two valid write conditions\n");
		return 1;
	}
	
	switch ( flag ) {
	case br_zero:
		*target = V2_OUT_VALID_ZERO;
		break;
	case br_negative:
		*target = V2_OUT_VALID_NEGATIVE;
		break;
	case br_never:
		*target = V2_OUT_VALID_NEVER;
		break;
	default:
		fprintf(stderr,"Invalid valid flag\n");
		return 1;
	}
	return 0;
}


static int write_policy_statement(struct v2_part* target, struct statements *s) {
	if ( s->target != state ) {
		fprintf(stderr,"Internal error. Invalid WP statement\n");
		return 1;
	}
	if ( target->write_policy != INVALID_SIGN ) {
		fprintf(stderr,"Tried to assert two write policies\n");
		return 1;
	}
	
	switch ( s->wr_flag ) {
	case wr_always:
		target->write_policy = V2_STATE_WRITE_ALWAYS;
		break;
	case wr_valid:
		target->write_policy = V2_STATE_WRITE_VALID;
		break;
	default:
		fprintf(stderr,"Invalid write policy flag\n");
		return 2;
	}
	return 0;
}

void print_v2_part(struct v2_part* source, int version) {
	if ( version >= 2 ) {
	printf("\nV2 CONFIG:\n");
	printf("SOUTH EAST SW  : 0x%x\n",source->south_east_switch);
	printf("EAST STATE SW  : 0x%x\n",source->east_state_switch);
	printf("STATE SOUTH SW : 0x%x\n",source->state_south_switch);
	printf("SOUTH VALID    : 0x%x\n",source->south_valid);
	printf("EAST VALID     : 0x%x\n",source->east_valid);
	printf("STATE VALID    : 0x%x\n",source->state_valid);
	printf("NORTH SELECT   : 0x%x\n",source->north_select);
	printf("WEST SELECT    : 0x%x\n",source->west_select);
	printf("WRITE POLICY   : 0x%x\n",source->write_policy);
	printf("---------------\n");
	}
	print_v1_part(source->v1);
}

void serialize_v2_part(struct v2_part* source, char* target_buffer, int version) {
	int buffer_pos = 0;
	if ( version >= 2 ) {
		pos_update_wrapper(0, 				2,			target_buffer,&buffer_pos);
		pos_update_wrapper(source->north_select,	V2_IN_SELECT_BITS,	target_buffer,&buffer_pos);
		pos_update_wrapper(source->west_select,		V2_IN_SELECT_BITS,	target_buffer,&buffer_pos);
		pos_update_wrapper(source->state_south_switch,	V2_OUT_VALID_BITS,	target_buffer,&buffer_pos);
		pos_update_wrapper(source->east_state_switch,	V2_OUT_VALID_BITS,	target_buffer,&buffer_pos);
		pos_update_wrapper(source->south_east_switch,	V2_OUT_VALID_BITS,	target_buffer,&buffer_pos);
		pos_update_wrapper(source->state_valid,		V2_OUT_VALID_BITS,	target_buffer,&buffer_pos);
		pos_update_wrapper(source->east_valid,		V2_OUT_VALID_BITS,	target_buffer,&buffer_pos);
		pos_update_wrapper(source->south_valid,		V2_OUT_VALID_BITS,	target_buffer,&buffer_pos);
		pos_update_wrapper(source->write_policy,	V2_STATE_WRITE_BITS,	target_buffer,&buffer_pos);
	}
	serialize_v1_part(source->v1,target_buffer+buffer_pos,version);
}
