#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>

#include "library.h"
#include "memory.h"
#include "cell.h"

static void try_library_init(void *library, int expected_x, int expected_y);
static int verify_configuration(void* library, int expected_x, int expected_y);
static int load_memory_configuration(struct libraries* library, int expected_x,
				int expected_y, int memory_delta);
static int load_fabric_configuration(struct libraries* library, int expected_x,
				int expected_y);

static int load_memory(struct libraries* library, int expected_x,
			int expected_y);

/* Builds a callframe/stateframe/stackframe
 * ------ <- memory_target
 *|memory|
 *--------
 *| state|
 *| +rc  |
 *--------
 */
int load_configuration(const char* file, 
			int expected_x, 
			int expected_y, 
			uint8_t* memory_target,
			struct libraries** list) {
	
	void* library = dlopen(file,RTLD_NOW);
	struct libraries* new_lib = NULL;
	int memory_delta = 0;
	
	if ( memory_target == NULL ) {
		fprintf(stderr, "Invalid memory region passed to libraryload\n");
		return 0;
	}
	
	if ( library == NULL ) {
		fprintf(stderr, 
			"Unable to load library %s : %s\n",
			file,
			dlerror()
		);
		return 0;
	}
	
	try_library_init(library,expected_x,expected_y);
	
	if ( !verify_configuration(library,expected_x,expected_y) ) {
		dlclose(library);
		return 0;
	}
	
	new_lib = malloc(sizeof(struct libraries));
	
	if ( new_lib == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		dlclose(library);
		return 0;
	}
	new_lib->file_name = file;
	new_lib->loaded_library = library;
	new_lib->load_address = find_free_task_space(
				*list,
				expected_x,
				expected_y,
				memory_target
	);
	new_lib->memory_size = 0;
	
	memory_delta = new_lib->load_address 
		+ configuration_size(expected_x,expected_y)
		- memory_target;
		
	printf("Memory delta : %08X\n",memory_delta);
	
	if ( !load_memory_configuration(
			new_lib,expected_x,
			expected_y,memory_delta) 
			) {
		
		free(new_lib);
		dlclose(library);
		return 0;
	}
	
	if ( !load_fabric_configuration(new_lib,expected_x,expected_y) ) {
		free(new_lib);
		dlclose(library);
		return 0;
	}
	
	if ( !load_memory(new_lib,expected_x,expected_y) ) {
		free(new_lib);
		dlclose(library);
		return 0;
	}
	
	find_free_task_space(*list,expected_x,expected_y,memory_target);
	new_lib->next = *list;
	*list = new_lib;
	return 1;
}

int unload_configuration(struct libraries** list) {
	struct libraries* cur = *list;
	struct libraries* next = NULL;
	
	while ( cur != NULL ) {
		next = cur->next;
		printf("Unloading configuration at %p\n",cur->load_address);
		if ( cur->loaded_library != NULL ) {
			dlclose(cur->loaded_library);
		}
		free(cur);
		cur = next;
	}
	
	*list = NULL;
}

static void try_library_init(void* library, int expected_x, int expected_y) {
	void (*init)(grid_size_t dims, grid_size_t* size) = NULL;
	grid_size_t sizes[2] = {expected_x,expected_y};
	
	if ( library == NULL ) {
		return;
	}
	
	init = dlsym(library,STRING(INIT_SYMBOL));
	
	if ( init == NULL ) {
		printf("No init function found for library\n");
		return;
	} else {
		printf("Running library initialization function\n");
		init(2,sizes);
	}
}

static int verify_configuration(void* library, int expected_x, int expected_y) {
	uint32_t* grid_size = NULL;
	uint32_t* grid_dims = NULL;
	
	if ( library == NULL ) {
		fprintf(stderr,"Tried to validate invalid library\n");
		return 0;
	}
	
	printf("%s\n",STRING(GRID_SIZE));
	grid_size = (uint32_t*) dlsym(library,STRING(GRID_SIZE));
	grid_dims = (uint32_t*) dlsym(library,STRING(GRID_DIMENTIONS));
	
	if ( grid_size == NULL || grid_dims == NULL ) {
		fprintf(stderr,
			"Unable to detect grid size/dimentions %p/%p\n",
			grid_size,
			grid_dims
		);
		return 0;
	}
	
	if ( *grid_dims != 2 ) {
		fprintf(stderr,"Unsupported dimention size: %d\n",*grid_dims);
		return 0;
	}
	
	if ( grid_size[0] != expected_x ) {
		fprintf(stderr,
			"Invalid X dimention: %d != %d\n", 
			grid_size[0],
			expected_x
		);
		return 0;
	}
	
	if ( grid_size[1] != expected_y ) {
		fprintf(stderr,
			"Invalid Y dimention: %d != %d\n", 
			grid_size[1],
			expected_y
		);
		return 0;
	}
	return 1;
}

static int load_memory_configuration(struct libraries* library, int expected_x,
				int expected_y, int memory_delta) {
	struct mem_pattern** memory_config = NULL;
	struct mem_pattern* inmemory = NULL;
	
	memory_config = (struct mem_pattern**) dlsym(
						library->loaded_library,
						STRING(MEMCFG_SYMBOL)
	);
	
	if ( memory_config == NULL || *memory_config == NULL) {
		fprintf(stderr,"Memconfig load failed : %s\n",dlerror());
		return 0;
	}
	/* WARNING: BUFFER OVERRUN POSSIBLE */
	/* Do something with this memory */
	memcpy(library->load_address,
		*memory_config,
		2*(expected_x+expected_y)*sizeof(struct mem_pattern)
	);
	
	inmemory = (struct mem_pattern*) library->load_address;
	/* Loop through the memory units and add memorydeltas */
	for ( int i = 0 ; i < (expected_x*2 + expected_y*2) ; i++ ) {
		
		inmemory->address += memory_delta;
		inmemory->address_limit += memory_delta;
		
		inmemory++;
	}
	
	return 1;
}

static int load_fabric_configuration(struct libraries* library, int expected_x,
				int expected_y) {
	
	struct cell_rc_targets** fabric_config = NULL;
	int offset = 2*(expected_x+expected_y)*sizeof(struct mem_pattern);
	
	fabric_config = (struct cell_rc_targets**) dlsym(
						library->loaded_library,
						STRING(FABRIC_SYMBOL)
	);
	
	if ( fabric_config == NULL || *fabric_config == NULL) {
		fprintf(stderr,"Fabric config load failed: %s\n",dlerror());
		return 0;
	}
	
	memcpy(library->load_address + offset,
		*fabric_config,
		expected_x*expected_y*sizeof(struct cell_rc_targets)
	);
	return 1;
}

static int load_memory(struct libraries* library, int expected_x,
			int expected_y) {
	
	memsize_t* memory_size = NULL;
	uint8_t** memory = NULL;
	int config_size = configuration_size(expected_x, expected_y);
	
	memory_size = (memsize_t*) dlsym(
					library->loaded_library,
					STRING(MEMORY_SIZE_SYMBOL)
	);
	
	if ( memory_size == NULL ) {
		fprintf(stderr,"Unable to load memory size: %s\n",dlerror());
		return 0;
	}
	
	memory = (uint8_t**) dlsym(
				library->loaded_library,
				STRING(MEMORY_SYMBOL)
	);
	
	if ( memory == NULL || *memory == NULL) {
		fprintf(stderr,"Unable to load memory: %s\n",dlerror());
		return 0;
	}
	
	memcpy(library->load_address + config_size, *memory, *memory_size);
	
	library->memory_size = *memory_size;
	return 1;
}

/* Calculate memory requirements */
int configuration_size(int expected_x, int expected_y) {
	int size = 0;
	size += memory_configuration_size(expected_x,expected_y);
	size += expected_x*expected_y*sizeof(struct cell_rc_targets);
	return size;
}

int memory_configuration_size(int expected_x, int expected_y) {
	return 2*(expected_x+expected_y)*sizeof(struct mem_pattern);
}

/* This returns aligned task space */
uint8_t* find_free_task_space(struct libraries* library, int expected_x,
				int expected_y, uint8_t* base_address) {
	struct libraries* current = library;
	uint8_t* free_slot = base_address;
	int config_size = configuration_size(expected_x,expected_y);
	
	while ( current != NULL ) {
		/* Check if this is the region with the highest address area */
		if ( current->load_address + current->memory_size + config_size
				> free_slot ) {
			
			free_slot = current->load_address 
				+ current->memory_size 
				+ config_size;
		}
		current = current->next;
	}
	if ( (unsigned int)free_slot % __alignof__(uint64_t) ) {
		printf("Warning: Unaligned memory space\n");
		free_slot += __alignof__(uint64_t) 
			- ((int)free_slot % __alignof__(uint64_t));
	}
	
	printf("freeslot %p\n",free_slot);
	return free_slot;
}

