#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "grid.h"
#include "memory.h"
#include "counters.h"

static int new_memory_interface_rc(struct mem_interface_rc** list,
			uint8_t* memory_base);
static int new_memory_interface(struct mem_interface** list,
			struct mem_interface** end,
			struct grid_cell* gcell,
			cell_valid* ioport_valid,
			cell_transfer* ioport,
			enum mem_interface_type mtype,
			struct mem_rc_bus* bus,
			uint8_t* memory_base,
			int init_idle);
static int memory_grid_west_south_con(struct mem_interface** list,
			struct mem_rc_bus* rc_bus,
			struct grid_cell* NW_corner,
			uint8_t* memory_base,
			int init_idle);
static int memory_grid_north_east_con(struct mem_interface** list,
			struct mem_rc_bus* rc_bus,
			struct grid_cell* NW_corner,
			uint8_t* memory_base,
			int init_idle);
static void free_memory_interfaces(struct mem_interface** list);
static void memory_single_unit_tick(struct mem_interface* munit,
			struct counters* cnt,
			int idle_size);
static void memory_fetch(struct mem_interface* munit,
			struct counters* cnt);
static void single_memory_rc_tick(struct mem_interface_rc* mrc,
			struct counters* cnt);
static void memory_unit_reconfigure(struct mem_interface_rc* master,
			struct mem_interface* unit,
			struct counters* cnt);
static void mem_rc_flush_drain(struct mem_interface_rc* master,
			struct counters* cnt);
static void memory_unit_grid_transfer(struct mem_interface* list);
static void write_idles(FILE* fd, int idles);
static void write_cell_transfer(FILE* fd, char valid, cell_transfer value);
static void drain_mem_unit_stripe(struct mem_interface* list,
				int idles, FILE* target);
static void write_stripe_item(struct mem_interface* cur, 
				int idles, FILE* target);


int initialize_memory(uint8_t** memory_base) {
	*memory_base = malloc(MEMORY_SIZE);
	if ( *memory_base == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		return 0;
	}
	memset(*memory_base,0,MEMORY_SIZE);
	return 1;
}

void free_memory(uint8_t** memory_base) {
	free(*memory_base);
	*memory_base = NULL;
}

int memory_grid_connection(struct mem_interface_rc** list, 
			struct grid_cell* NW_corner,
			uint8_t* memory_base,
			int init_idle) {
	*list = NULL;
	if ( !new_memory_interface_rc(list,memory_base) ) {
		return 0;
	}
	if ( !memory_grid_west_south_con(&((*list)->munit_list),&((*list)->rc_bus),NW_corner,memory_base,init_idle) ) {
		return 0;
	}
	if ( !new_memory_interface_rc(list,memory_base) ) {
		return 0;
	}
	if ( !memory_grid_north_east_con(&((*list)->munit_list),&((*list)->rc_bus),NW_corner,memory_base,init_idle) ) {
		return 0;
	}
	return 1;
}

static int new_memory_interface_rc(struct mem_interface_rc** list,
			uint8_t* memory_base) {
	struct mem_interface_rc* new;
	
	new = malloc(sizeof(struct mem_interface_rc));
	
	if ( new == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		return 0;
	}
	new->state = MEM_RC_IDLE;
	new->ticks_left = 0;
	new->last = NULL;
	new->munit_list = NULL;
	new->memory_base = memory_base;
	new->rc_bus.source.address = 0;
	new->rc_bus.source.address_limit = 0;
	new->rc_bus.source.stride = 0;
	new->rc_bus.source.dummy = 0;
	new->rc_bus.drain.address = 0xFC;
	new->rc_bus.drain.address_limit = 0xFC;
	new->rc_bus.drain.stride = 4;
	new->rc_bus.drain.dummy = 1;
	new->next = *list;
	printf("Created memory rc unit\n");
	*list = new;
}

/* This function assumes square grid */
static int memory_grid_west_south_con(struct mem_interface** list,
			struct mem_rc_bus* rc_bus,
			struct grid_cell* NW_corner,
			uint8_t* memory_base,
			int init_idle) {
	
	struct grid_cell* row = NW_corner;
	struct grid_cell* col = NULL;
	struct mem_interface* end = NULL;
	struct mem_interface** pend = &end;
	
	/* West connections */
	while ( row != NULL ) {
		if ( !new_memory_interface(list,pend,row,&(row->cell.register_valid_west),&(row->cell.register_west),FROM_MEM,rc_bus,memory_base,init_idle) ) {
			return 0;
		}
		col = row;
		row = row->south;
	}
	
	if ( !new_memory_interface(list,pend,NULL,NULL,NULL,UNCONFIGURABLE_MEM,rc_bus,memory_base,init_idle) ) {
		return 0;
	}
	
	/* South connections */
	while ( col != NULL ) {
		if ( !new_memory_interface(list,pend,col,&(col->cell.result_valid_south),&(col->cell.result_south),TO_MEM,rc_bus,memory_base,init_idle) ) {
			return 0;
		}
		
		col = col->east;
	}
	return 1;
}

static int memory_grid_north_east_con(struct mem_interface** list,
			struct mem_rc_bus* rc_bus,
			struct grid_cell* NW_corner,
			uint8_t* memory_base,
			int init_idle) {
	
	struct grid_cell* col = NW_corner;
	struct grid_cell* row = NULL;
	struct mem_interface* end = NULL;
	struct mem_interface** pend = &end;
	
	/* North connections */
	while ( col != NULL ) {
		if ( !new_memory_interface(list,pend,col,&(col->cell.register_valid_north),&(col->cell.register_north),FROM_MEM,rc_bus,memory_base,init_idle) ) {
			return 0;
		}
		
		row = col;
		col = col->east;
	}
	
	if ( !new_memory_interface(list,pend,NULL,NULL,NULL,UNCONFIGURABLE_MEM,rc_bus,memory_base,init_idle) ) {
		return 0;
	}
	
	/* East connections */
	while ( row != NULL ) {
		if ( !new_memory_interface(list,pend,row,&(row->cell.result_valid_east),&(row->cell.result_east),TO_MEM,rc_bus,memory_base,init_idle) ) {
			return 0;
		}
		
		row = row->south;
	}
	return 1;
}

static int new_memory_interface(struct mem_interface** list_start,
			struct mem_interface** list_end,
			struct grid_cell* gcell,
			cell_valid* ioport_valid,
			cell_transfer* ioport,
			enum mem_interface_type mtype,
			struct mem_rc_bus* rc_bus,
			uint8_t* memory_base,
			int init_idle) {
	
	struct mem_interface* munit = malloc(sizeof(struct mem_interface));
	
	
	if ( munit == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		return -1;
	}
	printf("Memory unit created: %p\n", munit);
	
	munit->state = MEM_UNCONFIGURED;
	munit->type = mtype;
	munit->connection = gcell;
	munit->memory_base = memory_base;
	munit->mem_info.address = 0;
	munit->mem_info.address_limit = 0;
	munit->mem_info.stride = 0;
	munit->mem_info.dummy = 0;
	munit->rc_bus = rc_bus;
	munit->output = 0x0;
	munit->ioport_valid = ioport_valid;
	munit->ioport = ioport;
	
	munit->private_dump = tmpfile();
	
	if ( munit->private_dump == NULL ) {
		fprintf(stderr,"Unable to open private dump file\n");
		return -2;
	}
	munit->current_idle = init_idle;
	
	if ( *list_start == NULL ) {
		printf("First in list %p\n",munit);
		munit->next = NULL;
		*list_start = munit;
		*list_end = munit;
	} else {
		printf("Appending %p to %p\n",munit,*list_end);
		(*list_end)->next = munit;
		*list_end = munit;
	}
}

void free_memory_connection(struct mem_interface_rc** list ) {
	struct mem_interface_rc* next = NULL;
	struct mem_interface_rc* cur = *list;
	
	while ( cur != NULL ) {
		next = cur->next;
		free_memory_interfaces(&(cur->munit_list));
		free(cur);
		cur = next;
	}

}

static void free_memory_interfaces(struct mem_interface** list) {
	struct mem_interface* next = NULL;
	struct mem_interface* cur = *list;
	
	while ( cur != NULL ) {
		next = cur->next;
		fclose(cur->private_dump);
		free(cur);
		cur = next;
	}
	*list = NULL;
}


void memory_unit_tick(struct mem_interface_rc* list,
			struct counters* cnt,
			int idle_size) {
	struct mem_interface_rc* rc_unit = list;
	struct mem_interface* munit = NULL;
	
	while ( rc_unit != NULL ) {
		munit = rc_unit->munit_list;
		while ( munit != NULL ) {
			memory_single_unit_tick(munit,cnt,idle_size);
			munit = munit->next;
		}
		rc_unit = rc_unit->next;
	}
}

static void memory_single_unit_tick(struct mem_interface* munit,
				struct counters* cnt,
				int idle_size) {
	printf("MEMORY UNIT TICK at %p\n",munit);
	if ( munit->type == FROM_MEM) {
		munit->current_idle = idle_size;
	} else if ( munit->next != NULL ) {
		munit->next->current_idle = idle_size;
	}
	switch(munit->state) {
	case MEM_UNCONFIGURED:
		printf("Unconfigured\n");
		munit->output_valid = 0;
		break;
	case MEM_RECONFIGURING:
		if ( munit->type == UNCONFIGURABLE_MEM ) {
			if ( munit->next != NULL ) {
				munit->next->current_idle = 0;
			}
			break;
		}
		munit->output_valid = 0;
		if ( munit->ticks_left == 1 ) {
			printf("Reconfiguring(transfer)\n");
			munit->rc_bus->drain = munit->mem_info;
			munit->mem_info = munit->rc_bus->source;
		} else {
			printf("Reconfiguring\n");
			munit->ticks_left--;
		}
		if ( munit->type == FROM_MEM ) {
			munit->current_idle = 0;
		} else if ( munit->next != NULL ) {
			munit->next->current_idle = 0;
		}
		break;
	case MEM_FETCHING:
		memory_fetch(munit,cnt);
		if ( munit->ticks_left == 1 ) {
			printf("Memory: Fetched data now on output port\n");
			munit->ticks_left = MEMORY_ACCESS_TICKS;
		} else {
			/* But some garbage on the outport
			 * Allows for detection of off-transfer-cycle
			 * access to outport */
			if ( munit->type == FROM_MEM ) { 
				/* Don't do this on input-registers 
				 * to output memory modules 
				 */
				munit->output_valid = 0;
				munit->output = 0xF1F1;
			}
			munit->ticks_left--;
		}
		break;
	case MEM_STALLED:
		munit->output = 0;
		munit->output_valid = 0;
		break;
	
	default:
		printf("Memory unit in invalid state: %d\n",munit->state);
		break;
	}
}

static void memory_fetch(struct mem_interface* munit,
				struct counters* cnt) {
	if ( munit->mem_info.address < munit->mem_info.address_limit ) {
		/* Do memory operation */
		if ( munit->mem_info.address >= MEMORY_SIZE ) {
			printf("Invalid memory access from %p\n",munit);
			munit->output_valid = 0;
		} else if ( munit->type == TO_MEM ) {
			/* Only copy on the last tick */
			record_memory_write_size(cnt,MEMORY_ACCESS_SIZE);
			if ( munit->ticks_left == 1 ) {
				printf("Writing 0x%X to 0x%lX\n",
					*(munit->ioport),
					munit->mem_info.address
				);
				memcpy(munit->memory_base 
					+ munit->mem_info.address, 
					&(munit->output), 
					sizeof(cell_transfer)
				);
				munit->output_valid = 0;
				munit->mem_info.address += munit->mem_info.stride;
			}
		} else if ( munit->type == FROM_MEM ) {
			/* Only copy on the last tick */
			record_memory_read_size(cnt,MEMORY_ACCESS_SIZE);
			if ( munit->ticks_left == 1 ) {
				printf("Reading 0x%X from 0%lX\n",
					*((uint16_t*)(munit->memory_base 
					+ munit->mem_info.address)),
					munit->mem_info.address
				);
				memcpy(&(munit->output), 
					munit->memory_base 
					+ munit->mem_info.address, 
					sizeof(cell_transfer)
				);
				munit->output_valid = 1;
				munit->mem_info.address += munit->mem_info.stride;
			}
		} else if ( munit->type == UNCONFIGURABLE_MEM ) {
			/* Always idle */
			munit->output_valid = 0;
		}
	} else {
		if ( munit->ticks_left == 1 ) {
			munit->state = MEM_STALLED;
		}
		printf("Memory unit %p stalled. Limit reached\n", munit);
		munit->output = 0;
		munit->output_valid = 0;
	}
}


void memory_rc_tick(struct mem_interface_rc* list, struct counters* cnt) {
	struct mem_interface_rc* cur = list;
	
	while ( cur != NULL ) {
		single_memory_rc_tick(cur,cnt);
		cur = cur->next;
	}
}

static void single_memory_rc_tick(struct mem_interface_rc* mrc,
				struct counters* cnt) {
	switch ( mrc->state ) {
	
	case MEM_RC_IDLE:
		printf("IDLE MEMRC\n");
		break;
	
	case MEM_RC_DELAYED:
	case MEM_RC_PRESTART:
		if ( mrc->ticks_left > 0 ) {
			mrc->ticks_left--;
			break;
		} else {
			if ( mrc->state == MEM_RC_DELAYED ) {
				if ( mrc->last->type == UNCONFIGURABLE_MEM ) {
					mrc->last->state = MEM_UNCONFIGURED;
				}
				if ( mrc->last->next != NULL ) {
					printf("Progressing to %p\n",mrc->last->next);
					mrc->last = mrc->last->next;
				}
			} else {
				mrc->last = mrc->munit_list;
			}
			mrc->state = MEM_RC_ACTIVE;
			mrc->ticks_left = MEMORY_RC_TICKS;
		}
	
	case MEM_RC_ACTIVE:
		/* Start writing after one tick of reconfiguration */
		if ( mrc->ticks_left < MEMORY_RC_TICKS ) {
			record_rc_memory_write_size(cnt,MEMORY_RC_ACCESS_SIZE);
		}
		if ( mrc->ticks_left == 0 ) {
			mrc->last->state = MEM_FETCHING;
			mrc->last->ticks_left = MEMORY_ACCESS_TICKS;
			printf("Copy drain\n");
			mem_rc_flush_drain(mrc,cnt);
			mrc->source_address += sizeof(struct mem_pattern);
			mrc->drain_address += sizeof(struct mem_pattern);
			
			if ( mrc->last->next != NULL ) {
				printf("Progressing to %p\n",mrc->last->next);
				mrc->last = mrc->last->next;
				mrc->ticks_left = MEMORY_RC_TICKS;
				if ( mrc->last->type == UNCONFIGURABLE_MEM) {
					mrc->state = MEM_RC_DELAYED;
				}
			}
		}
		if ( mrc->ticks_left > 0 ) {
			memory_unit_reconfigure(mrc,mrc->last,cnt);
			mrc->ticks_left--;
		} else {
			mrc->state = MEM_RC_IDLE;
		}
		break;
		
	
	default:
		printf("ERROR: Memory RC Manager: %p is in invalid state %d\n",
			mrc,
			mrc->state
		);
	}
}

void memory_rc_setup(struct mem_interface_rc* list,
			int x_size, int y_size,
			uint32_t source_memory_address,
			uint32_t drain_memory_address) {
	int memory_offset = 0; /* Offset in drain and source 
				  for a given mem_rc */
	struct mem_interface_rc* cur = list;
	
	while ( cur != NULL ) {
		cur->state = MEM_RC_PRESTART;
		cur->ticks_left = 0;
		cur->last = NULL;
		
		cur->source_address = source_memory_address + memory_offset;
		cur->drain_address = drain_memory_address + memory_offset;
		
		memory_offset += (x_size + y_size)*sizeof(struct mem_pattern);
		
		cur = cur->next;
	}
}

static void memory_unit_reconfigure(struct mem_interface_rc* master, 
			struct mem_interface* munit,
			struct counters* cnt) {
	struct mem_pattern* new_conf = NULL;
	
	if ( munit->type == UNCONFIGURABLE_MEM) {
		munit->state = MEM_RECONFIGURING;
		return;
	}
	
	record_rc_memory_read_size(cnt,MEMORY_RC_ACCESS_SIZE);
	/* Put data on bus the first cycle */
	if ( master->ticks_left == MEMORY_RC_TICKS ) {
		new_conf = (struct mem_pattern*) 
			(master->memory_base + master->source_address);
		printf("Reconfiguring: %p with 0x%X\n",
			munit,
			master->source_address
		);
		master->rc_bus.source = *new_conf;
		munit->ticks_left = MEMORY_RC_TICKS;
	}
	munit->state = MEM_RECONFIGURING;
}

static void mem_rc_flush_drain(struct mem_interface_rc* master,
			struct counters* cnt) {
	printf("Draining: %p to %X\n",master,master->drain_address);
	memcpy(master->memory_base + master->drain_address,
		&(master->rc_bus.drain),
		sizeof(struct mem_pattern)
	);
}


void memory_grid_transfer(struct mem_interface_rc* list) {
	struct mem_interface_rc* cur = list;
	
	while( cur != NULL ) {
		memory_unit_grid_transfer(cur->munit_list);
		cur = cur->next;
	}
}

static void memory_unit_grid_transfer(struct mem_interface* list) {
	struct mem_interface* cur = list;
	
	while( cur != NULL ) {
		
		if ( cur->type == FROM_MEM ) {
			printf("Copying %x to %p\n",cur->output,cur->ioport);
			*(cur->ioport) = cur->output;
			*(cur->ioport_valid) = cur->output_valid;
			printf("Valid is : %hhu\n",cur->output_valid);
			write_cell_transfer(
				cur->private_dump,
				cur->output_valid,
				cur->output
			);
			write_idles(
				cur->private_dump,
				cur->current_idle
			);
		
		} else if ( cur->type == TO_MEM) {
			printf("Copying %x from %p\n", 
				*(cur->ioport), 
				cur->ioport
			);
			cur->output = *(cur->ioport);
			cur->output_valid = *(cur->ioport_valid);
			printf("Valid is : %hhu\n",cur->output_valid);
			write_cell_transfer(
				cur->private_dump,
				cur->output_valid,
				cur->output
			);
			write_idles(
				cur->private_dump,
				cur->current_idle
			);
			
		} else if ( cur->type == UNCONFIGURABLE_MEM ) {
			/* This type does nothing */
		} else {
			fprintf(stderr,"Unknown memory unit type: %d\n", 
				cur->type
			);
		}
		
		cur = cur->next;
	}
}

static void write_idles(FILE* fd, int idles) {
	int i = 0;
	for (; i < idles; i++ ) {
		write_cell_transfer(fd,0,0);
	}
}


static void write_cell_transfer(FILE* fd, char valid, cell_transfer value) {
	int i = (sizeof(cell_transfer)*8)-1;
	fprintf(fd,"%c",valid+'0');
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((value>>i)&0x01));
	}
	fprintf(fd," ");
	fflush(fd);
}

void private_data_fseek(struct mem_interface_rc* list) {
	struct mem_interface_rc* cur = list;
	while ( cur != NULL ) {
		struct mem_interface* munit = cur->munit_list;
		while ( munit != NULL ) {
			fseek(munit->private_dump,0L,SEEK_SET);
			
			munit = munit->next;
		}
		cur = cur->next;
	}
}

void drain_stripe(struct mem_interface_rc* list, 
				int idles, FILE* target) {
	struct mem_interface_rc* cur = list;
	
	while ( cur != NULL ) {
		drain_mem_unit_stripe(
			cur->munit_list,
			idles, 
			target
		);
		//break;
		cur = cur->next;
	}
}

static void drain_mem_unit_stripe(struct mem_interface* list,
				int idles, FILE* target) {
	
	struct mem_interface* cur = list;
	while ( cur != NULL ) {
		if ( cur->type != UNCONFIGURABLE_MEM ) {
			write_stripe_item(cur,idles,target);
		}
		cur = cur->next;
	}
}

static void write_stripe_item(struct mem_interface* cur, 
				int idles, FILE* target) {
	int i = 0;
	int current = '\0';
	
	while ( (current = fgetc(cur->private_dump)) != EOF ) {
		if ( current == ' ' ) {
			
			break;
		}
		fputc(current,target);
	}
	fputc(' ', target);
	for ( ; i < idles; i++ ) {
		write_cell_transfer(target,0,0);
	}
}

int stripe_exists(struct mem_interface_rc* list) {
	struct mem_interface_rc* cur = list;
	struct mem_interface* munit = NULL;
	int current = 0;
	while ( cur != NULL ) {
		munit = cur->munit_list;
		while ( munit != NULL ) {
			if ( munit->type != UNCONFIGURABLE_MEM ) {
				current = fgetc(munit->private_dump);
				if ( current == EOF ) {
					return 1;
				}
				ungetc(current,munit->private_dump);
			}
			munit = munit->next;
		}
		cur = cur->next;
	}
	return 0;
}
