#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rc_manager.h"
#include "grid.h"
#include "cell.h"
#include "counters.h"

static void rc_configure(struct rc_manager* master, 
			struct grid_cell* target,
			struct counters* cnt);
static void rc_flush_drain(struct rc_manager* master,
			struct counters* cnt);

static void rc_managers_bus_connection(struct rc_manager* rc_m);


int rc_managers_create(struct grid_cell* NW_corner, uint8_t* memory_base,
			int x_size, struct rc_manager** list) {
	struct rc_manager* next = NULL;
	struct rc_manager* first = NULL;
	struct rc_manager* prev = NULL;
	struct grid_cell* col = NW_corner; /* Just renaming */
	
	while ( col != NULL ) {
		next = malloc(sizeof(struct rc_manager));
		if ( first == NULL ) {
			first = next;
			*list = first;
		}
		if ( next == NULL ) {
			fprintf(stderr,"Memory allocation error\n");
			return 0;
		}
		
		next->ticks_left = 0;
		next->stride = x_size*sizeof(struct cell_rc_targets);
		next->source_memory_address = 0;
		next->drain_memory_address = 0;
		next->memory_base = memory_base;
		next->state = IDLE;
		next->bus.source.state = 0xFAFA;
		next->bus.source.fabric = NULL;
		next->bus.drain.state = 0;
		next->bus.drain.fabric = NULL;
		next->rc_list = col;
		
		rc_managers_bus_connection(next);
		if ( prev != NULL ) {
			prev->next = next;
			prev = next;
		} else {
			prev = next;
		}
		col = col->east;
	}
	return 1;
}

static void rc_managers_bus_connection(struct rc_manager* rc_m) {
	struct grid_cell* next = rc_m->rc_list;
	while ( next != NULL ) {
		next->cell.config_bus = &(rc_m->bus);
		next = next->south;
	}
}

void rc_managers_free(struct rc_manager** list) {
	struct rc_manager* cur = *list;
	struct rc_manager* next = NULL;
	
	while ( cur != NULL ) {
		next = cur->next;
		free(cur);
		cur = next;
	}
	*list = NULL;
}

/* Setup managers to start reconfiguring */
void rc_setup(struct rc_manager* list, uint32_t source_memory_address, 
		uint32_t drain_memory_address) {
	int prestart_time = DATA_TRANSFER_TICKS; /* Memory rc is at zero */
	int memory_offset = 0; /* Offset in drain and source for a given
				* rc_manager 
				*/
	struct rc_manager* cur = list;
	
	while ( list != NULL ) {
		list->state = PRESTART;
		list->ticks_left = prestart_time;
		list->last = NULL;
		
		/* Generate read/write addresses for the configuration */
		list->source_memory_address = source_memory_address+memory_offset;
		list->drain_memory_address = drain_memory_address+memory_offset;
		
		memory_offset+=sizeof(struct cell_rc_targets);
		prestart_time+=DATA_TRANSFER_TICKS;
		list = list->next;
	}
}

void rc_tick(struct rc_manager* list, struct counters* cnt) {
	struct rc_manager* cur = list;
	
	while ( cur != NULL ) {
		switch ( cur->state ) {
		
		case IDLE:
			break;
		case PRESTART:
			if ( cur->ticks_left > 0 ) {
				cur->ticks_left--;
				break;
			} else {
				cur->last = cur->rc_list;
				cur->state = ACTIVE;
				cur->ticks_left = RC_TICKS;
			}
		case ACTIVE:
			/* Writing starts after first tick */
			if ( cur->ticks_left < RC_TICKS ) {
				record_rc_memory_write_size(cnt,RC_BUS_WIDTH);
			}
			if ( cur->ticks_left == 0 ) {
				cur->last->cell.config_state = COMPUTING;
				cur->last->cell.ticks_left = DATA_TRANSFER_TICKS;
				printf("Copy drain\n");
				rc_flush_drain(cur,cnt);
				cur->source_memory_address += cur->stride;
				cur->drain_memory_address += cur->stride;
				if ( cur->last->south != NULL ) {
					printf("Progressing to %p\n",
						cur->last->south
					);
					cur->last = cur->last->south;
					cur->ticks_left = RC_TICKS;
				}
			}
			if ( cur->ticks_left > 0 ) {
				rc_configure(cur,cur->last,cnt);
				cur->ticks_left--;
			} else {
				cur->state = IDLE;
			}
			break;
			
		default:
			printf("ERROR: RC Manager: %p is in invalid state %d\n",
				cur,
				cur->state
			);
		}
		cur = cur->next;
	}
}

static void rc_configure(struct rc_manager* master, struct grid_cell* target,
			struct counters* cnt){
	struct cell_rc_targets* new_conf = NULL; 
	
	record_rc_memory_read_size(cnt,RC_BUS_WIDTH);
	/* Last rc-tick. Do the accual reconfiguration */
	if ( master->ticks_left ==  RC_TICKS ) {
		new_conf = (struct cell_rc_targets*) 
			(master->memory_base + master->source_memory_address);
		printf("Reconfiguring: %p with %X\n",
			&(target->cell),
			master->source_memory_address
		);
		master->bus.source = *new_conf;
		target->cell.ticks_left = RC_TICKS;
	}
	target->cell.config_state = RECONFIGURING;
}

static void rc_flush_drain(struct rc_manager* master,
			struct counters* cnt) {
	printf("Draining: %p to %X\n", master, master->drain_memory_address);
	memcpy(master->memory_base + master->drain_memory_address, 
		&(master->bus.drain),
		sizeof(struct cell_rc_targets)
	);
}
