library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common_defs.all;

entity alu is 
	generic(
		width		:positive := 32
	);
	port(
		clk		:in	std_logic;
		reset		:in	std_logic;
		
		data_a		:in	std_logic_vector(width-1 downto 0);
		data_b		:in	std_logic_vector(width-1 downto 0);
		
		valid_a		:in	std_logic;
		valid_b		:in	std_logic;
		
		-- Hardcoded width 
		alu_op		:in	std_logic_vector(alu_op_size-1 downto 0);
		
		-- Flags Negative / Zero
		flags		:out	std_logic_vector(alu_flags_size-1 downto 0);
		
		output		:out	std_logic_vector(width-1 downto 0);
		valid_output	:out	std_logic
	);
end entity;

architecture registered_alu of alu is 
	signal reg_data_a	: std_logic_vector(width -1 downto 0);
	signal reg_data_b	: std_logic_vector(width -1 downto 0);
	signal reg_alu_op	: std_logic_vector(alu_op_size-1 downto 0);
	signal reg_valid_a	: std_logic;
	signal reg_valid_b	: std_logic;
	signal temp_output	: std_logic_vector(width -1 downto 0);
	signal temp_valid	: std_logic;
	
	signal shifter_dir	: std_logic;
	signal shifter_out	: std_logic_vector(width -1 downto 0);
	
begin
	sub_shifter : shifter
		generic map (
			width => width
		)
		port map(
			data => reg_data_a,
			shift_amount => reg_data_b,
			direction => shifter_dir,
			output => shifter_out
		);

	input : process(clk, reset) 
	begin
		if reset = '0' then
			reg_data_a <= (reg_data_a'range => '0');
			reg_data_b <= (reg_data_b'range => '0');
			reg_valid_a <= '0';
			reg_valid_b <= '0';
			reg_alu_op <= (reg_alu_op'range => '0');
		elsif rising_edge(clk) then
			if valid_a = '1' and valid_b = '1' and reg_valid_a = '0' and reg_valid_b = '0' then	
				reg_data_a <= data_a;
				reg_data_b <= data_b;
				reg_valid_a <= valid_a;
				reg_valid_b <= valid_b;
				reg_alu_op <= alu_op;
			end if;
		end if;
	end process;

	operation : process(reg_data_a,reg_data_b,reg_alu_op,shifter_out)
	begin 
		shifter_dir <= '-';
		case reg_alu_op is
		-- ADD
		when "000" =>
			temp_output <= std_logic_vector(signed(reg_data_a) + signed(reg_data_b));
		-- SUB
		when "001" =>
			temp_output <= std_logic_vector(signed(reg_data_a) - signed(reg_data_b));
		-- SRL
		when "010" =>
			shifter_dir <= '0';
			temp_output <= shifter_out;
		-- SLL
		when "011" => 
			shifter_dir <= '1';
			temp_output <= shifter_out;
		-- OR
		when "100" => 
			temp_output <= reg_data_a or reg_data_b;
		
		-- NOR
		when "101" => 
			temp_output <= reg_data_a nor reg_data_b;
		-- AND
		when "110" =>
			temp_output <= reg_data_a and reg_data_b;
		-- XOR
		when "111" =>
			temp_output <= reg_data_a xor reg_data_b;
		when others =>
			temp_output <= (temp_output'range => '0');
		end case;
	end process;
	
	-- Zero 
	flags(0) <= ('1' and temp_valid) when (temp_output = (temp_output'range => '0')) else '0';
	-- Negative
	flags(1) <= temp_output(temp_output'left) and temp_valid;
	-- Set valid flag
	temp_valid <= reg_valid_a and reg_valid_b;
	valid_output <= temp_valid;
	output <= temp_output;
end architecture;