library ieee;
use ieee.std_logic_1164.all;
use ieee.math_real.all;

package common_defs is 
	-- Constants
	constant alu_op_size	: positive	:= 3;
	constant alu_flags_size	: positive	:= 2;
	
	-- Common components
	component shift_register is 
	generic(
		width		: positive	:= 8
	);
	port(
		clk		: in std_logic;
		reset		: in std_logic;
		
		write		: in std_logic;
		
		shift_left 	: in std_logic;
		shift_right	: in std_logic;
		
		data_in		: in std_logic_vector(width-1 downto 0);
		data_out	: out std_logic_vector(width-1 downto 0);
		
		carry_in	: in std_logic;
		carry_out	: out std_logic
	);
	end component;
	
	component addersubtracter is 
	generic(
		width		: positive	:= 8
	);
	port(
		data_a		: in std_logic_vector(width-1 downto 0);
		data_b		: in std_logic_vector(width-1 downto 0);
		
		operation	: in std_logic;
		
		output		: out std_logic_vector(width-1 downto 0)
		
	);
	end component;
	
	
	component multiplydivider is 
	generic(
		width		:positive := 32
	);
	port(
		clk		:in	std_logic;
		reset		:in	std_logic;
		
		-- restart	:in	std_logic; ?
		-- 0 - MUL -- 1 - DIV
		opcode		:in	std_logic;
		
		valid_a		:in	std_logic;
		valid_b		:in	std_logic;
		
		-- Dividend
		data_a		:in	std_logic_vector(width-1 downto 0);
		-- Divider
		data_b		:in	std_logic_vector(width-1 downto 0);
		
		valid_output	:out	std_logic;
		output		:out	std_logic_vector(width-1 downto 0)
	);
	end component;
	
	component shifter is 
	generic(
		width		:positive := 32
	);
	port (
		data		:in	std_logic_vector(width -1 downto 0);
		
		shift_amount	:in	std_logic_vector(width -1 downto 0);
		direction	:in	std_logic;
		
		output		:out	std_logic_vector(width -1 downto 0)
	);
	end component;
	
	component alu is 
	generic(
		width		:positive := 32
	);
	port(
		clk		:in	std_logic;
		reset		:in	std_logic;
		
		data_a		:in	std_logic_vector(width-1 downto 0);
		data_b		:in	std_logic_vector(width-1 downto 0);
		
		valid_a		:in	std_logic;
		valid_b		:in	std_logic;
		
		-- Hardcoded width 
		--alu_op		:in	std_logic_vector(2 downto 0);
		alu_op		:in	std_logic_vector(alu_op_size-1 downto 0);
		
		-- Flags Negative / Zero
		flags		:out	std_logic_vector(alu_flags_size-1 downto 0);
		
		output		:out	std_logic_vector(width-1 downto 0);
		valid_output	:out	std_logic
	);
	end component;
	
	component mux is 
	generic(
		path_width	: positive	:= 8;
		sel_width	: positive	:= 2
	);
	port(
		input	: in std_logic_vector(((2**sel_width)*path_width)-1 downto 0);
		
		sel	: in std_logic_vector(sel_width -1  downto 0);
		output	: out std_logic_vector(path_width -1 downto 0)
	);
	end component;
	
	component sc_register is
	generic(
		width		: positive := 8;
		config_width	: positive := 2;
		allow_prohibit	: std_logic := '1'
	);
	port(
		config_clk	: in std_logic;
		data_clk	: in std_logic;
		reset		: in std_logic;
		
		config_in	: in std_logic_vector(config_width -1 downto 0);
		config_out	: out std_logic_vector(config_width -1 downto 0);
		config_enable	: in std_logic;
		
		data_in		: in std_logic_vector(width -1 downto 0);
		data_out	: out std_logic_vector(width -1 downto 0);
		write		: in std_logic
	);
	end component;
	
	component rc_master_rom is
	generic(
		elements	:positive	:= 4;
		data_width	:positive	:= 6;
		init_source	:string		:= "roms/default"
	);
	port (
		addr	: in std_logic_vector(integer(ceil(log2(real(elements))))-1 downto 0);
		output	: out std_logic_vector(data_width-1 downto 0)
	);
	end component;
	
	component rc_ram is
	generic(
		data_width	:positive := 38;
		addr_width	:positive := 8
	);
	port(
		clk		: in std_logic;
		
		write		: in std_logic;
		
		waddr		: in std_logic_vector(addr_width -1 downto 0);
		wdata		: in std_logic_vector(data_width -1 downto 0);
		
		raddr		: in std_logic_vector(addr_width -1 downto 0);
		rdata		: out std_logic_vector(data_width -1 downto 0)
	); 
	end component;
	
	component runtime_manager is
	generic(
		axis		:natural	:= 4
	);
	port(
		clk		:in std_logic;
		reset		:in std_logic;
		
		override	:in std_logic;
		rc_active	:in std_logic;
		vertical	:in std_logic;
		idle_out	:out std_logic
	);
	end component;

end package;
