library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common_defs.all;

entity multiplydivider is 
	generic(
		width		:positive := 16
	);
	port(
		clk		:in	std_logic;
		reset		:in	std_logic;
		
		-- restart	:in	std_logic; ?
		-- 0 - MUL -- 1 - DIV
		opcode		:in	std_logic;
		
		valid_a		:in	std_logic;
		valid_b		:in	std_logic;
		
		-- Dividend
		data_a		:in	std_logic_vector(width-1 downto 0);
		-- Divider
		data_b		:in	std_logic_vector(width-1 downto 0);
		
		valid_output	:out	std_logic;
		output		:out	std_logic_vector(width-1 downto 0)
	);
end entity;

architecture multicycle_multiply_divide of multiplydivider is 
	signal shifter_left	:std_logic;
	signal shifter_right	:std_logic;
	signal shifter_write	:std_logic;
	signal shifter_in	:std_logic_vector((2*width)-1 downto 0);
	signal shifter_out	:std_logic_vector((2*width)-1 downto 0);
	signal shifter_cin	:std_logic;
	signal shifter_cout	:std_logic;
	signal temp_valid	:std_logic;
	
	
	signal data_b_reg	:std_logic_vector(width -1 downto 0);
	signal data_a_reg	:std_logic_vector(width -1 downto 0);
	signal op_reg		:std_logic;
	signal alu_op_reg	:std_logic;
	signal sign_reg		:std_logic;
	
	type muldiv_states	is (start,convert,compute, done);
	signal md_state		:muldiv_states;
	signal next_state	:muldiv_states;
	
	signal counter		:integer range 0 to width;
	
	signal addsub_a		:std_logic_vector(width -1 downto 0);
	signal addsub_b		:std_logic_vector(width -1 downto 0);
	signal addsub_out	:std_logic_vector(width -1 downto 0);
	signal addsub_op	:std_logic;
	
	signal signed_output	:std_logic_vector(width -1 downto 0);
	signal zero_output	:std_logic_vector(width -1 downto 0);
	-- Active low
	signal div_by_zero	:std_logic;
begin
	shift_reg	:component shift_register
		generic map(
			width => 2*width
		)
		port map(
			clk => clk,
			reset => reset,
			write => shifter_write,
			shift_left => shifter_left,
			shift_right => shifter_right,
			data_in => shifter_in,
			data_out => shifter_out,
			carry_out => shifter_cout,
			carry_in => shifter_cin
		);
	
	addsub		:component addersubtracter
		generic map(
			width => width
		)
		port map(
			data_a => addsub_a,
			data_b => addsub_b,
			operation => addsub_op,
			output => addsub_out
		);
	
	md_control: process(clk, reset)
	begin
		if reset = '0' then
			-- Reset statemachine
			-- Shift-register clears itself
			md_state <= start;
			counter <= 0;
			data_a_reg <= (data_a_reg'range => '0');
			data_b_reg <= (data_b_reg'range => '0');
			op_reg <= '0';
			alu_op_reg <= '0';
			sign_reg <= '0';
			div_by_zero <= '0';
		elsif rising_edge(clk) then
			md_state <= next_state;
			if md_state = start then
				counter <= 0; --(counter'range => '0');
				data_b_reg <= data_b;
				data_a_reg <= data_a;
				op_reg <= opcode;
				alu_op_reg <= data_b(data_b'left) xor opcode;
				-- Check if the output should be manipulated
				sign_reg <= data_b(data_b'left) xor data_a(data_a'left);
			elsif md_state = convert then
				if data_b_reg = (data_b_reg'range => '0') then
					div_by_zero <= '1';
				else
					div_by_zero <= '0';
				end if;
				counter <= 0; --(counter'range => '0');
			elsif next_state = compute then
				counter <= counter +1;
			end if;
		end if;
	end process;
	
	
	md_compute: process(md_state,shifter_out,valid_a,valid_b,clk,shifter_cout,opcode,
				data_a,data_b_reg,counter,op_reg,addsub_out,div_by_zero,
				sign_reg,alu_op_reg,data_a_reg
	)
	variable temp_result	:signed(width -1 downto 0);
	begin
		-- Default values
		temp_valid <= '0';
		shifter_left <= '0';
		shifter_right <= '0';
		shifter_write <= '0';
		shifter_cin <= '0';
		
		addsub_a <= shifter_out((2*width)-1 downto width);
		addsub_b <= data_b_reg;
		addsub_op <= alu_op_reg;
		
		case md_state is 
		when start =>
			shifter_in <= (others => '0');
			-- Only start the compute stage if both inputs are valid 
			if valid_a = '1' and valid_b = '1' then
				next_state <= convert;
			else
				next_state <= start;
			end if;
		when convert => 
			shifter_write <= '1';
			shifter_in((2*width)-1 downto width) <= ((2*width)-1 downto width => '0');
			addsub_a <= (others => '0');
			addsub_b <= data_a_reg;
			addsub_op <= data_a_reg(data_a_reg'high);
			shifter_in(width-1 downto 0) <= addsub_out;
			
			next_state <= compute;
			
		when compute => 
			shifter_write <= '1';
			-- Naive 
			if op_reg = '0' then
				-- MUL
				shifter_right <= '1';
				if shifter_cout = '1' then
					temp_result := signed(addsub_out);
				else 
					temp_result := signed(shifter_out((2*width)-1 downto width));
				end if;
			else 
				-- DIV
				shifter_left <= '1';
				temp_result := signed(addsub_out);
				if temp_result < 0 then
					temp_result := signed(shifter_out((2*width)-1 downto width));
					shifter_cin <= (div_by_zero) and sign_reg;
				else
					shifter_cin <= '1';
				end if;
			end if;
			shifter_in((2*width)-1 downto width ) <= std_logic_vector(temp_result);
			shifter_in(width -1 downto 0 ) <= shifter_out(width -1 downto 0);
			
			if counter = width then
				next_state <= done;
			else 
				next_state <= compute;
			end if;
			
		when done =>
			-- Synth result is smaller with this.. 
			shifter_right <= not op_reg; 
			
			shifter_in <= (shifter_in'range => '0');
			temp_valid <= '1';
			
			addsub_a <= (others => '0');
			addsub_b <= shifter_out(width -1 downto 0);
			addsub_op <= sign_reg;
			
			next_state <= done;
		end case;
	end process;
	
	valid_output <= temp_valid;
	signed_output <= addsub_out;

	-- The core of the output should pass trough
	zero_output(zero_output'left-1 downto 1) <= signed_output(signed_output'left -1 downto 1);
	zero_fix : process(div_by_zero, sign_reg, signed_output)
	begin
		-- When a negative number by zero the result is ~(11..11)+1=00..01
		-- However the result should be 10..00
		if div_by_zero = '1' and sign_reg ='1' then
			-- Switch LSB and MSB on div by zero signed	
			zero_output(0) <= signed_output(signed_output'left);
			zero_output(zero_output'left) <= signed_output(0);
		else
		-- When dividing a positive by zero and the result is 11..11 = 11..11
		-- However the result should be 01..11 
			zero_output(0) <= signed_output(0);
			zero_output(zero_output'left) <= signed_output(signed_output'left) and (not div_by_zero);
		end if;
	end process;
	output <= zero_output;
end architecture;