#define _GNU_SOURCE
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define PROM_FILE "master_prom"
#define OROM_FILE "master_orom"

static int generate_pattern(int size, char mode);
static void write_grouped(FILE* fd,int line, int size);
static void write_multiplum(FILE* fd, int line, int size);
static int parse_numeric(char* string, int* target, const char* name);
static int generate_offsets(int addr_width, int size, int cfg_size, char mode);
static int generate_offset(int addr_width, int size, int cfg_size, 
			char mode, char* filename);
static void write_offset(FILE* fd, int addr_width, int line, 
			char mode, int cfg_size, int size);

int main(int argc, char** argv) {
	int size = -1;
	int addr_width = -1;
	int cfg_size = -1;
	char mode = '0';
	
	if ( argc != 5 ) {
		printf("usage: %s mode addr_width cfg_size size\n",argv[0]);
		return -1;
	}
	
	mode = argv[1][0];
	if ( mode != 'm' && mode != 'g' ) {
		fprintf(stderr,"Invalid mode specified\n");
		return -1;
	}
	
	if ( parse_numeric(argv[2],&addr_width,"addr_width") ) {
		return -1;
	}
	if ( parse_numeric(argv[3],&cfg_size,"cfg_size") ) {
		return -1;
	}
	if ( parse_numeric(argv[4],&size,"size") ) {
		return -1;
	}
	
	if ( generate_pattern(size,mode) ) {
		return -1;
	}
	if ( generate_offsets(addr_width,size,cfg_size,mode) ) {
		return -1;
	}
	
	return 0;
}


static int generate_pattern(int size, char mode) {
	FILE* pfd = NULL;
	char* filename = NULL;
	int i = 0;
	int pow2_size = 1<<(int)(ceilf(log2f((float)size+1)));
	
	if (asprintf(&filename,"%s_%d_%d.txt",PROM_FILE,size+1,size) < 0) {
		fprintf(stderr,"Memory allocation error\n");
		return -1;
	}
	
	if ( (pfd = fopen(filename,"w")) == NULL ) {
		perror("pattern:");
		return -1;
	}
	for ( i = 0 ; i < pow2_size; i ++ ) {
		if ( mode == 'm' ) {
			write_multiplum(pfd,i,size);
		} else {
			write_grouped(pfd,i,size);
		}
	}
	free(filename);
	fclose(pfd);
	return 0;
}

static void write_grouped(FILE* fd,int line, int size) {
	int i = 0;
	for ( i = 0 ; i < size; i++ ) {
		if ( i < size - line || line > size) {
			/* Zero any line larger than the grid size*/
			fprintf(fd,"0");
		} else {
			fprintf(fd,"1");
		}
	}
	fprintf(fd,"\n");
}

static void write_multiplum(FILE* fd, int line, int size) {
	int i = 0;
	int factor = -1;
	if ( line == 0 || (size/line)*line != size ) {
		/* The grid size is not dividable by this subsize */
		factor = -1;
	} else {
		/* The grid size is dividable */
		factor = (size/line);
	}
	for ( i = 0 ; i < size; i++ ) {
		if ( factor < 0 ) {
			fprintf(fd,"0");
		} else if ( ((i+1)/factor)*factor == (i+1) ) {
			fprintf(fd,"1");
		} else {
			fprintf(fd,"0");
		}
	}
	fprintf(fd,"\n");
}

static int generate_offsets(int addr_width, int size, int cfg_size, char mode) {
	char* filename = NULL;
	if (asprintf(&filename,"%s_%d_%d.txt",OROM_FILE,size+1,addr_width) < 0) {
		fprintf(stderr,"Memory allocation error\n");
		return -1;
	}
	if ( generate_offset(addr_width,size,cfg_size,mode,filename) ) {
		return -1;
	}
	free(filename);
	if (asprintf(&filename,"%s_v2_%d_%d.txt",OROM_FILE,size+1,addr_width) < 0) {
		fprintf(stderr,"Memory allocation error\n");
		return -1;
	}
	if ( generate_offset(addr_width,size,2*cfg_size,mode,filename) ) {
		return -1;
	}
	free(filename);
	return 0;
}
static int generate_offset(int addr_width, int size, int cfg_size, 
				char mode, char* filename) {
	FILE* ofd = NULL;
	int i = 0;
	int pow2_size = 1<<(int)(ceilf(log2f((float)size+1)));
	
	if ( (ofd = fopen(filename,"w")) == NULL ) {
		perror("offset:");
		return -1;
	}
	for ( i = 0 ; i < pow2_size; i ++ ) {
		write_offset(ofd,addr_width,i,mode,cfg_size,size);
	}
	fclose(ofd);
	return 0;
}

static void write_offset(FILE* fd, int addr_width, int line, 
			char mode, int cfg_size, int size) {
	int i = addr_width-1;
	int value = 0;
	if ( mode == 'g' ) {
		value = line*cfg_size;
	} else {
		if ( line == 0 || (size/line)*line != size ) {
			value = 0;
		} else {
			value = (size/(size/line))*cfg_size;
		}
	}
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",(char)('0'+((value>>i)&0x01)));
	}
	fprintf(fd,"\n");
}

static int parse_numeric(char* string, int* target, const char* name) {
	char* end = NULL;
	errno = 0;
	*target = (int)strtol(string,&end,10);
	
	if ( end == string || end != string + strlen(string)) {
		fprintf(stderr,"unable to parse %s\n",name);
		return 1;
	} else if ( errno == ERANGE || errno == EINVAL) {
		perror(name);
		return 1;
	}
	return 0;
}

