library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use work.common_defs.all;

entity alu_testbench is
end entity;

architecture alu_test of alu_testbench is
	constant size		: integer := 32;

	signal clk		: std_logic;
	signal reset		: std_logic;
	
	signal opcode_i		: std_logic_vector(alu_op_size -1 downto 0);
	signal valid_a_i	: std_logic;
	signal valid_b_i	: std_logic;
	
	signal data_a_i		: std_logic_vector(size -1 downto 0);
	signal data_b_i		: std_logic_vector(size -1 downto 0);

	signal valid_output_o	: std_logic;
	signal output_o		: std_logic_vector(size -1 downto 0);
	signal flags_o		: std_logic_vector(alu_flags_size -1 downto 0);
	signal running		: std_logic := '1';
begin
	uut : alu
		generic map(
			width => size
		)
		port map(
			clk => clk,
			reset => reset,
			alu_op => opcode_i,
			valid_a => valid_a_i,
			valid_b => valid_b_i,
			data_a => data_a_i,
			data_b => data_b_i,
			valid_output => valid_output_o,
			flags => flags_o,
			output => output_o
		);
	
	
	clock_gen : process is
	begin
		while running = '1' loop
			clk <= '1';
			wait for 50 ns;
			clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	manual_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(clk);
			wait until rising_edge(clk);
			reset <= '1';
			valid_a_i <= '0';
			valid_b_i <= '0';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		procedure perform_simulation(
			constant va	:in std_logic;
			constant vb	:in std_logic;
			constant da	:in std_logic_vector(size-1 downto 0);
			constant db	:in std_logic_vector(size-1 downto 0);
			constant op	:in std_logic_vector(alu_op_size -1 downto 0);
			constant result	:in std_logic_vector(size-1 downto 0);
			constant flags	:in std_logic_vector(alu_flags_size -1 downto 0)
			)
		is
			variable counter	:integer	:= 0;
		begin
			reset_uut;
			opcode_i  <= op;
			valid_a_i <= va;
			data_a_i  <= da;
			
			valid_b_i   <= vb;
			data_b_i  <= db;
			
			wait until rising_edge(clk);
			
			wait for 1 ns; -- Allow for propagation of signals 
			assert valid_output_o = (va and vb) report "Valid output not set" severity failure;
			
			assert output_o = result report "Incorrect result " 
				& integer'image(to_integer(signed(result))) 
				& " != " & integer'image(to_integer(signed(output_o))) 
				severity failure;
			assert flags = flags_o report "Incorrect flags. Expected "
				& Image(flags)
				& " was "
				& Image(flags_o)
				severity failure;
			-- End simulationstep
			valid_a_i <= '0';
			valid_b_i <= '0';
		
		end procedure;
	begin
		data_a_i <= (data_a_i'range => '0');
		data_b_i <= (data_b_i'range => '0');
		valid_a_i <= '0';
		valid_b_i <= '0';
		opcode_i <= "000";
		reset_uut;
		--random_stimuli;
		
		
		-- Basic positive I/O-test X+0=X and 0+X=X
		perform_simulation('1','1',X"00000001",X"00000000","000",X"00000001","00");
		perform_simulation('1','1',X"00000000",X"00000001","000",X"00000001","00");
		perform_simulation('1','1',X"AAAAAAAA",X"00000000","000",X"AAAAAAAA","10");
		perform_simulation('1','1',X"55555555",X"00000000","000",X"55555555","00");
		perform_simulation('1','1',X"00000000",X"AAAAAAAA","000",X"AAAAAAAA","10");
		perform_simulation('1','1',X"00000000",X"55555555","000",X"55555555","00");

		-- Some random number
		perform_simulation('1','1',X"0000BEEF",X"00000000","000",X"0000BEEF","00");
		-- Negative passtrough
		perform_simulation('1','1',X"8002323E",X"00000000","000",X"8002323E","10");
		perform_simulation('1','1',X"00000000",X"80000342","000",X"80000342","10");
		
		
		-- Test addition
		-- 1 + 1 = 2 - Simple carry
		perform_simulation('1','1',X"00000001",X"00000001","000",X"00000002","00");
		-- 3 + 3 = 6
		perform_simulation('1','1',X"00000003",X"00000003","000",X"00000006","00");
		-- 1 + -1 = 0 (zero flag)
		perform_simulation('1','1',X"00000001",X"FFFFFFFF","000",X"00000000","01");
		-- -1 + 2 = 1
		perform_simulation('1','1',X"FFFFFFFF",X"00000002","000",X"00000001","00");
		-- Operand order should be irrelevant
		perform_simulation('1','1',X"00000543",X"FFFFFFE2","000",X"00000525","00");
		perform_simulation('1','1',X"FFFFFFE2",X"00000543","000",X"00000525","00");
		-- Overflow
		perform_simulation('1','1',X"7FFFFFFF",X"00000004","000",X"80000003","10");
		
		
		-- Test subtraction
		-- 2 - 1 = 1 
		perform_simulation('1','1',X"00000002",X"00000001","001",X"00000001","00");
		-- 15 - 6
		perform_simulation('1','1',X"0000000F",X"00000006","001",X"00000009","00");
		-- 1 -1  = 0
		perform_simulation('1','1',X"00000001",X"00000001","001",X"00000000","01");
		-- -1 - 1 = -2
		perform_simulation('1','1',X"FFFFFFFF",X"00000001","001",X"FFFFFFFE","10");
		-- Subtraction of negative number should be addition
		perform_simulation('1','1',X"00000004",X"FFFFFFFE","001",X"00000006","00");
		-- Operand order should matter
		perform_simulation('1','1',X"000000EF",X"00000054","001",X"0000009B","00");
		perform_simulation('1','1',X"00000054",X"000000EF","001",X"FFFFFF65","10");
		-- Underflow
		perform_simulation('1','1',X"80000005",X"00000023","001",X"7FFFFFE2","00");
		-- Overflow
		perform_simulation('1','1',X"7FFFFF34",X"FFFFF2EF","001",X"80000C45","10");
		
		-- Test SRL and SLL
		-- Basic tests, is covered by subcomponent tests
		-- SRL
		-- Passtrough
		perform_simulation('1','1',X"FFFFFFFF",X"00000000","010",X"FFFFFFFF","10");
		perform_simulation('1','1',X"00000010",X"00000001","010",X"00000008","00");
		perform_simulation('1','1',X"00000010",X"00000002","010",X"00000004","00");
		perform_simulation('1','1',X"00000030",X"00000002","010",X"0000000C","00");
		-- SLL
		perform_simulation('1','1',X"FFFFFFFF",X"00000000","011",X"FFFFFFFF","10");
		perform_simulation('1','1',X"00000010",X"00000001","011",X"00000020","00");
		perform_simulation('1','1',X"00000010",X"00000002","011",X"00000040","00");
		perform_simulation('1','1',X"00000030",X"00000002","011",X"000000C0","00");
		-- Wiring test SRL 
		perform_simulation('1','1',X"00000002",X"00000000","010",X"00000002","00");
		perform_simulation('1','1',X"00000002",X"00000001","010",X"00000001","00");
		perform_simulation('1','1',X"00000002",X"00000002","010",X"00000000","01");
		-- Wiring test SLL 
		perform_simulation('1','1',X"40000000",X"00000000","011",X"40000000","00");
		perform_simulation('1','1',X"40000000",X"00000001","011",X"80000000","10");
		perform_simulation('1','1',X"40000000",X"00000002","011",X"00000000","01");
		
		-- Test OR
		perform_simulation('1','1',X"34534E65",X"00000000","100",X"34534E65","00");
		perform_simulation('1','1',X"00000002",X"00000001","100",X"00000003","00");
		perform_simulation('1','1',X"00000012",X"00000105","100",X"00000117","00");
		
		perform_simulation('1','1',X"34534E65",X"00000001","100",X"34534E65","00");
		perform_simulation('1','1',X"34534E65",X"00000054","100",X"34534E75","00");
		perform_simulation('1','1',X"AAAAAAAA",X"AAAAAAAA","100",X"AAAAAAAA","10");
		perform_simulation('1','1',X"55555555",X"55555555","100",X"55555555","00");
		perform_simulation('1','1',X"AAAAAAAA",X"55555555","100",X"FFFFFFFF","10");
		
		-- Operand order should not matter
		perform_simulation('1','1',X"000000E4",X"00000011","100",X"000000F5","00");
		perform_simulation('1','1',X"00000011",X"000000E4","100",X"000000F5","00");
		
		-- Test NOR
		perform_simulation('1','1',X"00000011",X"00000000","101",X"FFFFFFEE","10");
		perform_simulation('1','1',X"FFFFFFFF",X"00000000","101",X"00000000","01");
		perform_simulation('1','1',X"AAAAAAAA",X"AAAAAAAA","101",X"55555555","00");
		perform_simulation('1','1',X"55555555",X"55555555","101",X"AAAAAAAA","10");
		-- Operand order should not matter
		perform_simulation('1','1',X"DEADBEEF",X"CAFEBABE","101",X"21004100","00");
		perform_simulation('1','1',X"CAFEBABE",X"DEADBEEF","101",X"21004100","00");
		
		-- Test AND
		perform_simulation('1','1',X"00000001",X"FFFFFFFF","110",X"00000001","00");
		perform_simulation('1','1',X"00000001",X"FFFFFFFF","110",X"00000001","00");
		perform_simulation('1','1',X"AAAAAAAA",X"AAAAAAAA","110",X"AAAAAAAA","10");
		perform_simulation('1','1',X"AAAAAAAA",X"0F0F0F0F","110",X"0A0A0A0A","00");
		perform_simulation('1','1',X"55555555",X"55555555","110",X"55555555","00");
		perform_simulation('1','1',X"55555555",X"AAAAAAAA","110",X"00000000","01");
		perform_simulation('1','1',X"55555555",X"FFFFFFFF","110",X"55555555","00");
		-- Operand order should not matter
		perform_simulation('1','1',X"DEADBEEF",X"CAFEBABE","110",X"CAACBAAE","10");
		perform_simulation('1','1',X"CAFEBABE",X"DEADBEEF","110",X"CAACBAAE","10");
		
		-- Test XOR
		perform_simulation('1','1',X"00000001",X"00000000","111",X"00000001","00");
		perform_simulation('1','1',X"00000001",X"00000001","111",X"00000000","01");
		perform_simulation('1','1',X"AAAAAAAA",X"00000000","111",X"AAAAAAAA","10");
		perform_simulation('1','1',X"AAAAAAAA",X"AAAAAAAA","111",X"00000000","01");
		perform_simulation('1','1',X"55555555",X"55555555","111",X"00000000","01");
		perform_simulation('1','1',X"AAAAAAAA",X"55555555","111",X"FFFFFFFF","10");
		perform_simulation('1','1',X"AAAAAAAA",X"FFFFFFFF","111",X"55555555","00");
		-- Operand order should not matter
		perform_simulation('1','1',X"DEADBEEF",X"CAFEBABE","111",X"14530451","00");
		perform_simulation('1','1',X"CAFEBABE",X"DEADBEEF","111",X"14530451","00");
		
		-- Instruction mix
		perform_simulation('1','1',X"00000543",X"FFFFFFE2","000",X"00000525","00");
		perform_simulation('1','1',X"DEADBEEF",X"CAFEBABE","101",X"21004100","00");
		perform_simulation('1','1',X"AAAAAAAA",X"FFFFFFFF","111",X"55555555","00");
		perform_simulation('1','1',X"00000010",X"00000002","010",X"00000004","00");
		perform_simulation('1','1',X"34534E65",X"00000054","100",X"34534E75","00");
		perform_simulation('1','1',X"7FFFFF34",X"FFFFF2EF","001",X"80000C45","10");
		perform_simulation('1','1',X"AAAAAAAA",X"0F0F0F0F","110",X"0A0A0A0A","00");
		perform_simulation('1','1',X"00000010",X"00000002","011",X"00000040","00");
		perform_simulation('1','1',X"FFFFFFE2",X"00000543","000",X"00000525","00");
		
		-- Flag output on invalid data
		perform_simulation('1','1',X"00000001",X"00000001","000",X"00000002","00");
		perform_simulation('0','0',X"00000001",X"00000001","000",X"00000000","00");
		perform_simulation('0','0',X"80000001",X"00000001","000",X"00000000","00");
		
		-- End of test bench
		running <= '0';
		wait for 500 ns;
		wait;
	end process;

end architecture;