library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use work.common_defs.all;

entity auto_alu_testbench is
end entity;

architecture alu_test of auto_alu_testbench is
	constant size		: integer := 32;

	signal clk		: std_logic;
	signal reset		: std_logic;
	
	signal opcode_i		: std_logic_vector(alu_op_size -1 downto 0);
	signal valid_a_i	: std_logic;
	signal valid_b_i	: std_logic;
	
	signal data_a_i		: std_logic_vector(size -1 downto 0);
	signal data_b_i		: std_logic_vector(size -1 downto 0);

	signal valid_output_o	: std_logic;
	signal output_o		: std_logic_vector(size -1 downto 0);
	signal flags_o		: std_logic_vector(alu_flags_size -1 downto 0);
	signal running		: std_logic := '1';
begin
	uut : alu
		generic map(
			width => size
		)
		port map(
			clk => clk,
			reset => reset,
			alu_op => opcode_i,
			valid_a => valid_a_i,
			valid_b => valid_b_i,
			data_a => data_a_i,
			data_b => data_b_i,
			valid_output => valid_output_o,
			flags => flags_o,
			output => output_o
		);
	
	
	clock_gen : process is
	begin
		while running = '1' loop
			clk <= '1';
			wait for 50 ns;
			clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(clk);
			wait until rising_edge(clk);
			reset <= '1';
			valid_a_i <= '0';
			valid_b_i <= '0';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		procedure perform_simulation(
			constant da	:in std_logic_vector(size-1 downto 0);
			constant db	:in std_logic_vector(size-1 downto 0);
			constant op	:in std_logic_vector(alu_op_size -1 downto 0);
			constant result	:in std_logic_vector(size-1 downto 0);
			constant flags	:in std_logic_vector(alu_flags_size -1 downto 0)
			)
		is
			variable counter	:integer	:= 0;
		begin
			reset_uut;
			opcode_i  <= op;
			valid_a_i <= '1';
			data_a_i  <= da;
			
			valid_b_i   <= '1';
			data_b_i  <= db;
			
			wait until rising_edge(clk);
			
			wait for 1 ns; -- Allow for propagation of signals 
			assert valid_output_o = '1' report "Valid output not set" severity failure;
			
			assert output_o = result report "Incorrect result " 
				& integer'image(to_integer(signed(result))) 
				& " != " & integer'image(to_integer(signed(output_o))) 
				severity failure;
			assert flags = flags_o report "Incorrect flags. Expected "
				& Image(flags)
				& " was "
				& Image(flags_o)
				severity failure;
			-- End simulationstep
			valid_a_i <= '0';
			valid_b_i <= '0';
		
		end procedure;

	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-alu.txt";
			variable L: line;
			
			variable operation	:std_logic_vector(alu_op_size -1 downto 0);
			variable first_operand	:std_logic_vector(size -1 downto 0);
			variable second_operand	:std_logic_vector(size -1 downto 0);
			variable result		:std_logic_vector(size -1 downto 0);
			variable flags		:std_logic_vector(alu_flags_size -1 downto 0);
		
			--variable op		:integer;
		
--			procedure perform_simulation is 
--		
--			begin
--				opcode_i  <= std_logic_vector(to_unsigned(operation,alu_op_size));
--				valid_a_i <= '1';
--				data_a_i  <= std_logic_vector(to_signed(first_operand,size));
--				
--				valid_b_i   <= '1';
--				data_b_i  <= std_logic_vector(to_signed(second_operand,size));
--				
--				wait until rising_edge(clk);
--				wait for 1 ns; -- shouldn't this be valid at the same time ?
--				assert valid_output_o = '1' report "Valid signal not set" severity failure;
--				assert output_o = std_logic_vector(to_signed(result,size)) report "Incorrect result" severity failure;
--			end procedure;
		begin
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				read(L,operation);
				read(L,first_operand);
				read(L,second_operand);
				read(L,result);
				read(L,flags);
				reset_uut;
				perform_simulation(first_operand,second_operand,operation,result,flags);
			end loop;
		end procedure;
	begin
		data_a_i <= (data_a_i'range => '0');
		data_b_i <= (data_b_i'range => '0');
		valid_a_i <= '0';
		valid_b_i <= '0';
		opcode_i <= (opcode_i'range => '0');
		reset_uut;
		random_stimuli;
		-- Hack
		--assert false report "End of simulation" severity failure;
		running <= '0';
		wait;
	end process;

end architecture;