library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version1_defs.all;

entity cell_testbench is
end entity;

architecture cell_test of cell_testbench is
	constant size		: integer := 16;
	constant c_size		: integer := 2;
	constant config_size	: integer := 38;
	
	signal config_clk	: std_logic;
	signal data_clk		: std_logic;
	signal compute_clk	: std_logic;
	signal reset		: std_logic;
	
	signal test_clk		: std_logic;
	
	signal reconfigure_i	: std_logic;
	signal config_in_i	: std_logic_vector(c_size-1 downto 0);
	signal config_out_o	: std_logic_vector(c_size-1 downto 0);
	
	-- With room for valid signals
	signal north_i		: std_logic_vector(size downto 0);
	signal west_i		: std_logic_vector(size downto 0);
	signal south_o		: std_logic_vector(size downto 0);
	signal east_o		: std_logic_vector(size downto 0);
	
	signal north_idle_i	: std_logic;
	signal west_idle_i	: std_logic;
	signal south_idle_o	: std_logic;
	signal east_idle_o	: std_logic;
	
	signal running		: integer := 1;
begin
	uut : cell
		generic map(
			width => size,
			config_width => c_size
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			compute_clk => compute_clk,
			reset => reset,
			reconfigure => reconfigure_i,
			config_in => config_in_i,
			config_out => config_out_o,
			north_idle => north_idle_i,
			west_idle => west_idle_i,
			south_idle => south_idle_o,
			east_idle => east_idle_o,
			north => north_i,
			west => west_i,
			south => south_o,
			east => east_o
		);

	config_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			config_clk <= '1';
			--wait for 50 ns;
			wait for 73 ns;
			config_clk <= '0';
			wait for 74 ns;
			--wait for 50 ns;
		end loop;
		wait;
	end process;
	
	data_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			data_clk <= '1';
			wait for 73 ns;
			data_clk <= '0';
			wait for 2720 ns;
		end loop;
		wait;
	end process;
	
	compute_clock : process is 
	begin
		wait for 10 ns;
		while running = 1 loop
			compute_clk <= '1';
			wait for 67 ns;
			compute_clk <= '0';
			wait for 66 ns;
		end loop;
		wait;
	end process;
	
	
	-- For internal use in the testbench
	test_clock : process is
	begin
		while running = 1 loop
			test_clk <= '1';
			wait for 73 ns;
			test_clk <= '0';
			wait for 2720 ns;
		end loop;
		wait;
	end process;
	
	manual_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(data_clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		
		procedure perform_simulation (
			constant reconfigure	: std_logic;
			constant config_in	: std_logic_vector(config_size-1 downto 0);
			constant config_out	: std_logic_vector(config_size-1 downto 0);
			
			constant north		: std_logic_vector(size downto 0);
			constant west		: std_logic_vector(size downto 0);
			constant south		: std_logic_vector(size downto 0);
			constant east		: std_logic_vector(size downto 0);
			constant north_idle	: std_logic;
			constant west_idle	: std_logic
			)
		is 
			variable serial_in	: std_logic_vector(config_size-1 downto 0);
			variable serial_out	: std_logic_vector(config_size-1 downto 0);
			variable cycle_counter	: integer := (config_size/c_size);
		begin
			-- Don't play with the signals until the current async write is complete 
			serial_in := config_in;
			
			north_idle_i <= north_idle;
			west_idle_i <= west_idle;
			north_i <= north;
			west_i <= west;
			reconfigure_i <= reconfigure;
			wait until rising_edge(data_clk);
			north_i <= (others => '0');
			west_i <= (others => '0');
			reconfigure_i <= '0';
			
			north_idle_i <= '0';
			west_idle_i <= '0';
			
			
			config_in_i <= serial_in(c_size -1 downto 0);
			serial_in := (c_size -1 downto 0 => '0') & serial_in(config_size -1 downto c_size);
			while cycle_counter > 1 loop
				-- Check output before config_clk rising edge when not reconfiguring
				if reconfigure = '1' then
					wait until rising_edge(config_clk);
					serial_out := config_out_o & serial_out(config_size-1 downto c_size);
				else 
					-- Allow for propagation
					wait for 1 ns;
					serial_out := config_out_o & serial_out(config_size-1 downto c_size);
					wait until rising_edge(config_clk);
				end if;
				-- Put data on input
				config_in_i <= serial_in(c_size -1 downto 0);
				serial_in := (c_size -1 downto 0 => '0') & serial_in(config_size -1 downto c_size);
				
				cycle_counter := cycle_counter - 1;
			end loop;
			
			--report Image(serial_in);
			
			wait until rising_edge(test_clk); 
			-- Fetch the last bits on the serial_bus
			serial_out := config_out_o & serial_out(config_size-1 downto c_size);
			assert (north_idle or west_idle) = east_idle_o 
				report "East idle incorrect"
				severity failure;
			assert (north_idle or west_idle) = south_idle_o 
				report "South idle incorrect"
				severity failure;
			
			-- Only verify outputs when not reconfiguring 
			if reconfigure = '0' then
				if  east /= (east'range => '-') then
					assert east(east'high) = east_o(east_o'high) report "East valid incorrect. Expected "
						& std_logic'image(east(east'high))
						& " was " 
						& std_logic'image(east_o(east_o'high))
						severity failure;
					
					-- Only check valid flag if not valid
					if east(east'high) = '1' then
						assert east = east_o report "East output incorrect. Expected " 
							& Image(east)
							& " was "
							& Image(east_o)
							severity failure;
					end if;
				end if;
				if  south /= (south'range => '-') then
					assert south(south'high) = south_o(south_o'high) report "South valid incorrect. Expected "
						& std_logic'image(south(south'high))
						& " was " 
						& std_logic'image(south_o(south_o'high))
						severity failure;
					
					-- Only check valid flag if not valid
					if south(south'high) = '1' then
						assert south = south_o report "South output incorrect. Expected " 
							& integer'image(to_integer(unsigned(south)))
							& " was "
							& integer'image(to_integer(unsigned(south_o)))
							severity failure;
					end if;
				end if;
			-- Only verify config_out when reconfigure is active 
			end if;
			assert config_out = serial_out
				report "Config out incorrect. Expected " 
					& Image(config_out)
					& " was " 
					& Image(serial_out)
					severity failure;
			
			
		end procedure;
	begin
		reconfigure_i <= '0';
		config_in_i <= (others => '0');
		north_idle_i <= '0';
		west_idle_i <= '0';
		reset_uut;
		
		-- Check reset 
		assert config_out_o = (config_out_o'range => '0') report "Reset failed" severity failure;
		
		-- Sync with clocks
		wait until rising_edge(test_clk);
		
		-- Test that the configurationsystem doesn't loose any bits
		-- Check that the configuration output is zero.
		perform_simulation('1',
			"11111111111111111111111111111111111111",
			"00000000000000000000000000000000000000",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		-- Some computation.. Don't care
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"-----------------",
			"-----------------",
			"00000000000000000",
			"00000000000000000",
			'0','0'
		);
		-- State bit will have changed
		perform_simulation('1',
			"00000000000000000000000000000000000000",
			"11111111111111111111100000000000000000",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		
		
		-- Reconfigure with pass trough configuration. 
		-- West -> East
		-- North -> South
		-- Check that the configuration output is zero.
		perform_simulation('1',
			"00000000000000000000110000000000000000",
			"00000000000000000000000000000000000000",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		-- Test that the configuration was successful
		-- Inputs should have valid bits set. 
		perform_simulation('0',
			"00000000000000000000000000000000000000",
			"00000000000000000000000000000000000000",
			"10111001001001010",
			"11101010101011011",
			"10111001001001010",
			"11101010101011011",
			'0','0'
		);
		-- Check if the configuration is retained
		perform_simulation('0',
			"00000000000000000000000000000000000000",
			"00000000000000000000000000000000000000",
			"10101001001111011",
			"10101111010110101",
			"10101001001111011",
			"10101111010110101",
			'0','0'
		);
		
		-- Check config bus connection during computation
		perform_simulation('0',
			"10110010110100101011101001101010101010",
			"10110010110100101011101001101010101010",
			"10101111010110101",
			"10101001001111011",
			"10101111010110101",
			"10101001001111011",
			'0','0'
		);
		
		-- Check valid-flag propagation
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"1----------------",
			"0----------------",
			"1----------------",
			"0----------------",
			'0','0'
		);
		
		-- Check valid-flag propagation
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"0----------------",
			"1----------------",
			"0----------------",
			"1----------------",
			'0','0'
		);
		
		-- Check crosstalk
		perform_simulation('0',
			"00000000000000000000000000000000000000",
			"00000000000000000000000000000000000000",
			"00000000000000000",
			"10101111010110101",
			"00000000000000000",
			"10101111010110101",
			'0','0'
		);
		
		-- Retrieve configuration
		-- Configuration part should be static
		-- However the state register should have changed, as it is 
		-- not possible to disable writing on V1-arch
		perform_simulation('1',
			"00000000000000000000000000000000000000",
			"00000000000000000000110101111010110101",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		
		-- Check two successive reconfigurations
		perform_simulation('1',
			"00010000000000001100100000000000000000",
			"00000000000000000000000000000000000000",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		
		-- Check alu
		-- Config: 
		-- alu = north + west
		-- south = north
		-- east = alu
		-- (state = west)
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10001010101111011",
			"10001111010100110",
			"10001010101111011",
			"10011010000100001",
			'0','0'
		);
		-- Is the configuration retained ?
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10010001111000110",
			"10000111011011111",
			"10010001111000110",
			"10011001010100101",
			'0','0'
		);
		
		-- Check validflag propagation when alu is active
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"0----------------",
			"10000111011011111",
			"0----------------",
			"00000000000000000",
			'0','0'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"0----------------",
			"00000111011011111",
			"0----------------",
			"00000000000000000",
			'0','0'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"1----------------",
			"00000111011011111",
			"1----------------",
			"00000000000000000",
			'0','0'
		);
		
		-- Change ALU config to NOR (101)
		-- Config:
		-- alu = north nor west
		-- south = north
		-- east = alu
		-- (state = west)
		perform_simulation('1',
			"00011010000000001100110000000000000000",
			"00010000000000001100100000111011011111",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10000111101010101",
			"10111010100100011",
			"10000111101010101",
			"11000000010001000",
			'0','0'
		);
		
		-- Check multiplier
		-- Config: 
		-- mul = north*west
		-- south = north
		-- east = mul
		-- (state = west)
		-- Retrived state should contain west input from previous simulation
		perform_simulation('1',
			"00000000100000010000100000000000000000",
			"00011010000000001100110111010100100011",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10000011000000000",
			"10000000000001110",
			"10000011000000000",
			"10101010000000000",
			'0','0'
		);
		
		-- Check validflag propagation when multiplier is active
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"0----------------",
			"10010101010100111",
			"0----------------",
			"00000000000000000",
			'0','0'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"0----------------",
			"00010101010100111",
			"0----------------",
			"00000000000000000",
			'0','0'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"1----------------",
			"00010101010100111",
			"1----------------",
			"00000000000000000",
			'0','0'
		);
		
		-- Check state register
		-- Config:
		-- state = north;
		-- south = state;
		-- init_state = 42;
		-- ( east = west )
		perform_simulation('1',
			"00000000000000100001010000000000101010",
			"00000000100000010000100010101010100111",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10001010010100101",
			"-----------------",
			"10000000000101010",
			"-----------------",
			'0','0'
		);
		
		-- Check update on invalid data
		-- (State register should be updated regardless on V1-arch)
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"00101001010101011",
			"-----------------",
			"10001010010100101",
			"-----------------",
			'0','0'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10001011111010110",
			"-----------------",
			"00101001010101011",
			"-----------------",
			'0','0'
		);
		
		-- Check that internal data-flow works correctly
		-- Config: 
		-- alu = north - mul
		-- mul = state * west
		-- east = alu
		-- state = state
		-- south = state
		-- init_state = 28543
		perform_simulation('1',
			"01110011000001001101010000000000000101",
			"00000000000000100001010001011111010110",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10010111100001011",
			"10000000010000110",
			"10000000000000101",
			"10010110001101101",
			'0','0'
		);
		-- Check that the flow resets 
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10101110101001011",
			"10000111110110101",
			"10000000000000101",
			"10000111011000010",
			'0','0'
		);
		-- Check that the a missing valid input signal leads to no valid computation output
		-- State should remain valid
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10001010111100000",
			"0----------------",
			"10000000000000101",
			"00000000000000000",
			'0','0'
		);
		
		-- Check idle
		-- Config:
		-- alu = north + state;
		-- south = alu
		-- state = west
		-- east = state
		-- init_state = 0
		perform_simulation('1',
			"01100000000000001001110000000000000000",
			"01110011000001001101010000000000000101",
			"-----------------",
			"-----------------",
			"-----------------",
			"-----------------",
			'0','0'
		);
		
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10000001000001100",
			"10000000110110000",
			"10000001000001100",
			"10000000000000000",
			'0','0'
		);
		
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10000001010001110",
			"10010000000101010",
			"10000010000111110",
			"10000000110110000",
			'0','0'
		);
		-- Now.. idle
		-- The content of the state register should keep its value
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"1----------------",
			"1----------------",
			"0----------------",
			"0----------------",
			'1','0'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"1----------------",
			"1----------------",
			"0----------------",
			"0----------------",
			'0','1'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10000110011110101",
			"10000011011010111",
			"10010110100011111",
			"10010000000101010",
			'0','0'
		);
		perform_simulation('0',
			"--------------------------------------",
			"--------------------------------------",
			"10000000000000001",
			"10000001110111010",
			"10000011011011000",
			"10000011011010111",
			'0','0'
		);
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;
