library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version2_defs.all;

entity auto_cell_testbench is
end entity;

architecture cell_test of auto_cell_testbench is
	constant size		: integer := 16;
	constant c_size		: integer := 3;
	constant config_size	: integer := 57;
	
	signal config_clk	: std_logic;
	signal data_clk		: std_logic;
	signal compute_clk	: std_logic;
	signal reset		: std_logic;
	
	signal test_clk		: std_logic;
	
	signal reconfigure_i	: std_logic;
	signal config_in_i	: std_logic_vector(c_size-1 downto 0);
	signal config_out_o	: std_logic_vector(c_size-1 downto 0);
	
	-- With room for valid signals
	signal north_i		: std_logic_vector(size downto 0);
	signal west_i		: std_logic_vector(size downto 0);
	signal south_o		: std_logic_vector(size downto 0);
	signal east_o		: std_logic_vector(size downto 0);
	
	signal north_idle_i	: std_logic;
	signal west_idle_i 	: std_logic;
	signal south_idle_o	: std_logic;
	signal east_idle_o	: std_logic;
	
	signal running		: integer := 1;
begin
	uut : cell
		generic map(
			width => size,
			config_width => c_size
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			compute_clk => compute_clk,
			reset => reset,
			reconfigure => reconfigure_i,
			config_in => config_in_i,
			config_out => config_out_o,
			north_idle => north_idle_i,
			west_idle => west_idle_i,
			south_idle => south_idle_o,
			east_idle => east_idle_o,
			north => north_i,
			west => west_i,
			south => south_o,
			east => east_o
		);

	config_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			config_clk <= '1';
			wait for 73 ns;
			config_clk <= '0';
			wait for 74 ns;
		end loop;
		wait;
	end process;
	
	data_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			data_clk <= '1';
			wait for 73 ns;
			data_clk <= '0';
			wait for 2720 ns;
		end loop;
		wait;
	end process;
	
	compute_clock : process is 
	begin
		wait for 10 ns;
		while running = 1 loop
			compute_clk <= '1';
			wait for 67 ns;
			compute_clk <= '0';
			wait for 66 ns;
		end loop;
		wait;
	end process;
	
	
	-- For internal use in the testbench
	test_clock : process is
	begin
		while running = 1 loop
			test_clk <= '1';
			--wait for 50 ns;
			wait for 73 ns;
			test_clk <= '0';
			wait for 2720 ns;
			--wait for 1850 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(data_clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		
		procedure perform_simulation (
			constant reconfigure	: std_logic;
			constant config_in	: std_logic_vector(config_size-1 downto 0);
			constant config_out	: std_logic_vector(config_size-1 downto 0);
			
			constant north		: std_logic_vector(size downto 0);
			constant west		: std_logic_vector(size downto 0);
			constant south		: std_logic_vector(size downto 0);
			constant east		: std_logic_vector(size downto 0)
			)
		is 
			variable serial_in	: std_logic_vector(config_size-1 downto 0);
			variable serial_out	: std_logic_vector(config_size-1 downto 0);
			variable cycle_counter	: integer := (config_size/c_size);
		begin
			-- Don't play with the signals until the current async write is complete 
			serial_in := config_in;
			
			north_i <= north;
			west_i <= west;
			reconfigure_i <= reconfigure;
			wait until rising_edge(data_clk);
			north_i <= (others => '0');
			west_i <= (others => '0');
			reconfigure_i <= '0';
			
			config_in_i <= serial_in(c_size -1 downto 0);
			serial_in := (c_size -1 downto 0 => '0') & serial_in(config_size -1 downto c_size);
			while cycle_counter > 1 loop
				-- Check output before config_clk rising edge when not reconfiguring
				if reconfigure = '1' then
					wait until rising_edge(config_clk);
					serial_out := config_out_o & serial_out(config_size-1 downto c_size);
				else 
					-- Allow for propagation
					wait for 1 ns;
					serial_out := config_out_o & serial_out(config_size-1 downto c_size);
					wait until rising_edge(config_clk);
				end if;
				-- Put data on input
				config_in_i <= serial_in(c_size -1 downto 0);
				serial_in := (c_size -1 downto 0 => '0') & serial_in(config_size -1 downto c_size);
				
				cycle_counter := cycle_counter - 1;
			end loop;
			
			
			wait until rising_edge(test_clk); 
			-- Fetch the last bits on the serial_bus
			serial_out := config_out_o & serial_out(config_size-1 downto c_size);
			
			-- Only verify config_out when reconfigure is active 
			if reconfigure = '0' then
				assert east(east'high) = east_o(east_o'high) report "East valid incorrect. Expected "
					& std_logic'image(east(east'high))
					& " was " 
					& std_logic'image(east_o(east_o'high))
					severity failure;
				
				-- Only check valid flag if not valid
				if east(east'high) = '1' then
					assert east = east_o report "East output incorrect. Expected " 
						& Image(east)
						& " was "
						& Image(east_o)
						severity failure;
				end if;
				assert south(south'high) = south_o(south_o'high) report "South valid incorrect. Expected "
					& std_logic'image(south(south'high))
					& " was " 
					& std_logic'image(south_o(south_o'high))
					severity failure;
				
				-- Only check valid flag if not valid
				if south(south'high) = '1' then
					assert south = south_o report "South output incorrect. Expected " 
						& integer'image(to_integer(unsigned(south)))
						& " was "
						& integer'image(to_integer(unsigned(south_o)))
						severity failure;
				end if;
			end if;
			-- Only verify relevant config data
			if config_out(size) = '1' then
				assert config_out = serial_out
					report "Config out incorrect. Expected " 
						& Image(config_out)
						& " was " 
						& Image(serial_out)
						severity failure;
			else 
				assert config_out(config_out'high downto size) = serial_out(serial_out'high downto size)
					report "(Partial) Config out incorrect. Expected "
						& Image(config_out)
						& " was "
						& Image(serial_out)
						severity failure;
			end if;
			
		end procedure;
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-cell.txt";
			variable L: line;
			variable good: boolean;
			
			variable rc		:std_logic;
			
			variable config_in	:std_logic_vector(config_size-1 downto 0);
			variable config_out	:std_logic_vector(config_size-1 downto 0);
			variable north		:std_logic_vector(size downto 0);
			variable west		:std_logic_vector(size downto 0);
			variable south		:std_logic_vector(size downto 0);
			variable east		:std_logic_vector(size downto 0);
		begin
			
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				
				read(L,rc);
				read(L,config_in);
				read(L,config_out);
				read(L,north);
				read(L,west);
				read(L,south);
				read(L,east);
				
				perform_simulation(rc,config_in,config_out,north,west,south,east);
			end loop;
		end procedure;
		
	begin
		reconfigure_i <= '0';
		config_in_i <= (others => '0');
		north_idle_i <= '0';
		west_idle_i <= '0';
		reset_uut;
		
		-- Check reset 
		assert config_out_o = (config_out_o'range => '0') report "Reset failed" severity failure;
		
		-- Sync with clocks
		wait until rising_edge(test_clk);
		
		random_stimuli;
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;