#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#define op_t int16_t
#define uop_t uint16_t
/* CE CI CO W DI CDO NDO */
#define OP_LENGTH 16
#define CF_LENGTH 40
#define FULL_CONFIG_LENGTH (CF_LENGTH+1+OP_LENGTH)

#define IN_FILE "./auto.cell"
#define OUT_FILE "./auto.bit"
#define TRANSLATOR "../../../assembler/translator 2 1 1 auto.cell auto.bit"

enum alu {
	alu_add,
	alu_sub,
	alu_slr,
	alu_sll,
	alu_or,
	alu_nor,
	alu_and,
	alu_xor
};

const char* alu_ops[] = {
	"+",
	"-",
	"slr",
	"sll",
	"or",
	"nor",
	"and",
	"xor"
};

enum mul {
	mul_mul,
	mul_div
};

const char* mul_ops[] = {
	"*",
	"/"
};

enum source {
	west,
	north,
	state,
	mulout,
	aluout
};

const char* source_ops[] = {
	"west",
	"north",
	"state",
	"mulout",
	"aluout"
};

enum in_select {
	in_fixed,
	in_notvalid_other,
	in_other_valid
};

const char* in_select_ops[] = {
	"Fixed (default)",
	"not valid",
	"valid"
};

enum out_valid {
	out_fixed,
	out_zero,
	out_negative,
	out_never
};

const char* out_valid_ops[] = {
	"Fixed (default)",
	"zero",
	"negative",
	"never"
};

enum out_switch {
	sw_fixed,
	sw_zero,
	sw_negative
};

const char* out_switch_ops[] = {
	"Fixed (default)",
	"zero",
	"negative"
};

enum write_policy {
	wp_always,
	wp_onvalid
};

const char* write_policy_ops[] = {
	"always",
	"valid"
};

struct configuration {
	enum in_select sel_north;
	enum in_select sel_west;
	enum source alu_a;
	enum source alu_b;
	enum alu alu_op;
	enum source mul_a;
	enum source mul_b;
	enum mul mul_op;
	enum source east;
	enum source south;
	enum source state;
	enum out_switch sw_east_south;
	enum out_switch sw_south_state;
	enum out_switch sw_state_east;
	enum out_valid valid_east;
	enum out_valid valid_south;
	enum out_valid valid_state;
	enum write_policy state_wp;
	int32_t init_value;
};

struct cell_op {
	char valid;
	op_t value;
};

struct cell_state {
	struct cell_op in_north;
	struct cell_op north;
	struct cell_op in_west;
	struct cell_op west;
	struct cell_op state;
	struct cell_op out_state;
	struct cell_op alu_a;
	struct cell_op alu_b;
	struct cell_op alu_out;
	struct cell_op mul_a;
	struct cell_op mul_b;
	struct cell_op mul_out;
	struct cell_op east;
	struct cell_op out_east;
	struct cell_op south;
	struct cell_op out_south;
	struct cell_op cur_state; /* hm.. */
};

static void write_config(FILE* fd,struct configuration cfg);
static void write_switches(FILE* fd, struct configuration cfg);
static void write_select(FILE* fd, const char* target, const char* alt, enum in_select policy);
static void write_valid(FILE* fd, const char* target, enum out_valid policy);
static void update_old_config(char* old_config, struct cell_op state );
static void write_debug(FILE* fd, struct cell_state* cs);
static void write_stimuli(FILE* fd, char rc, struct cell_state* cs,char* config_in, char* config_out);
static void write_stim_config(FILE* fd, char* buffer, char rc);
static void write_operand(FILE* fd, struct cell_op co);
static void simulate_computation(struct cell_state* cs, const struct configuration* cfg, char rc);
static void auto_generate_config(struct configuration* cfg);
static void copy_to_mux(struct cell_state* cs, enum source conf, struct cell_op* target);
static void alu_compute(const struct configuration* cfg, struct cell_state* cs);
static void md_compute(const struct configuration* cfg, struct cell_state* cs);
static int generate_config_stream(struct configuration cfg, char** new_conf);
static void input_select(enum in_select, struct cell_op primary, struct cell_op secondary, struct cell_op* target);
static void output_switch(enum out_switch policy, struct cell_op alu_out, struct cell_op primary, struct cell_op secondary, 
				struct cell_op* prim_target, struct cell_op* sec_target);

static void output_valid(enum out_valid policy, struct cell_op alu_out, struct cell_op *target);
static void write_policy(enum write_policy policy, struct cell_op source, struct cell_op *target);

int main(int argc, char** argv) {
	long int vectors = 0;
	FILE* fd = NULL;
	char* current_config = NULL;
	char* old_config = NULL;
	struct configuration cfg = {0,};
	struct cell_state cs = {0,};
	
	time_t timeseed = time(NULL);
	srand(timeseed);
	
	if ( argc != 3 ) {
		printf("Usage: %s vectors <output filname>\n",argv[0]);
		return -1;
	}
	
	errno = 0;
	vectors = strtol(argv[1],NULL,10);
	if ( errno == ERANGE || errno == EINVAL ) {
		fprintf(stderr,"Parse error: %s\n",strerror(errno));
		return -1;
	}
	
	if ( (fd = fopen(argv[2],"w")) == NULL ) {
		perror("fopen");
		return -1;
	}
	
	current_config = malloc(FULL_CONFIG_LENGTH+1);
	if ( current_config == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		return -1;
	} else {
		memset(current_config,'0',FULL_CONFIG_LENGTH);
		current_config[FULL_CONFIG_LENGTH] = '\0';
	}
	
	for (;vectors > 0 ; vectors-- ) {
		char reconfigure = !(rand()%32);
		if ( reconfigure ) {
			auto_generate_config(&cfg);
			free(old_config);
			old_config = current_config;
			update_old_config(old_config,cs.cur_state);
			if ( generate_config_stream(cfg,&current_config) ) {
				return 1;
			}
		} else {
			cs.in_north.valid = !!(rand()%8);
			cs.in_north.value = rand()%0xFFFF;
			cs.in_west.valid = !!(rand()%8);
			cs.in_west.value = rand()%0xFFFF;
		}
		simulate_computation(&cs,&cfg,reconfigure);
		write_stimuli(fd,reconfigure,&cs,current_config,old_config);
	}
	free(old_config);
	free(current_config);
	fclose(fd);
	return 0;
}

static void update_old_config(char* old_config, struct cell_op state ) {
	int i = OP_LENGTH-1;
	*(old_config+CF_LENGTH) = '0'+state.valid;
	for ( ; i >= 0 ; i-- ) {
		*(old_config+FULL_CONFIG_LENGTH-1-i) = '0'+((state.value>>i)&0x01);
	}
}

static void write_debug(FILE* fd, struct cell_state* cs) {
	fprintf(fd,"(");
	fprintf(fd,"Real north: %hd\t",cs->in_north.value);
	fprintf(fd,"Comp north: %hd\t",cs->north.value);
	fprintf(fd,"Real west: %hd\t",cs->in_west.value);
	fprintf(fd,"Comp west: %hd\t",cs->west.value);
	fprintf(fd,"Cur state: %hd\t",cs->cur_state.value);
	fprintf(fd,"Comp state: %hd\t",cs->state.value);
	fprintf(fd,"Real state: %hd\t",cs->out_state.value);
	fprintf(fd,"ALU a: %hd\t",cs->alu_a.value);
	fprintf(fd,"ALU b: %hd\t",cs->alu_b.value);
	fprintf(fd,"ALU out: %hd\t",cs->alu_out.value);
	fprintf(fd,"MUL a: %hd\t",cs->mul_a.value);
	fprintf(fd,"MUL b: %hd\t",cs->mul_b.value);
	fprintf(fd,"MUL out: %hd\t",cs->mul_out.value);
	fprintf(fd,"Real east: %hd\t",cs->out_east.value);
	fprintf(fd,"Comp east: %hd\t",cs->east.value);
	fprintf(fd,"Real south: %hd\t",cs->out_south.value);
	fprintf(fd,"Comp south: %hd ",cs->south.value);
	fprintf(fd,")\n");
}

static void write_stimuli(FILE* fd, char rc, struct cell_state* cs, char* config_in, char* config_out) {
	fprintf(fd,"%c ",rc+'0');
	/* Insert cfg stuff */
	write_stim_config(fd,config_in,rc);
	write_stim_config(fd,config_out,rc);
	
	write_operand(fd,cs->in_north);
	write_operand(fd,cs->in_west);
	write_operand(fd,cs->out_south);
	write_operand(fd,cs->out_east);
	fprintf(fd,"\n");
}



static void write_stim_config(FILE* fd, char* buffer, char rc) {
	int i = 0;
	for ( ; i < FULL_CONFIG_LENGTH ; i++ ) {
		if ( rc ) {
			fprintf(fd,"%c",*(buffer+i));
		} else {
			fprintf(fd,"-");
		}
	}
	fprintf(fd," ");
}

static void write_operand(FILE* fd,struct cell_op co) {
	int i = OP_LENGTH-1;
	fprintf(fd,"%c",'0'+co.valid);
	for ( ; i >= 0 ; i-- ) {
		fprintf(fd,"%c",'0'+((co.value>>i)&0x01));
	}
	fprintf(fd," ");
}

static void simulate_computation(struct cell_state* cs, const struct configuration* cfg, char rc) {
	int i = 0;
	if ( rc ) {
		cs->cur_state.valid = 1;
		cs->cur_state.value = cfg->init_value;
	} else {
		/* Input switching */
		input_select(cfg->sel_north,cs->in_north,cs->in_west,&(cs->north));
		input_select(cfg->sel_west,cs->in_west,cs->in_north,&(cs->west));
		
		cs->state = cs->cur_state;
		
		/* Alu and mul.. go */
		for ( i = 0 ; i < 2 ; i++ ) {
			copy_to_mux(cs,cfg->alu_a,&(cs->alu_a));
			copy_to_mux(cs,cfg->alu_b,&(cs->alu_b));
			copy_to_mux(cs,cfg->mul_a,&(cs->mul_a));
			copy_to_mux(cs,cfg->mul_b,&(cs->mul_b));
			alu_compute(cfg,cs);
			md_compute(cfg,cs);
		}
		copy_to_mux(cs,cfg->east,&(cs->east));
		copy_to_mux(cs,cfg->south,&(cs->south));
		copy_to_mux(cs,cfg->state,&(cs->state));
		output_switch(cfg->sw_east_south,cs->alu_out,cs->east,cs->south,&(cs->out_east),&(cs->out_south));
		output_switch(cfg->sw_state_east,cs->alu_out,cs->state,cs->out_east,&(cs->out_state),&(cs->out_east));
		output_switch(cfg->sw_south_state,cs->alu_out,cs->out_south,cs->out_state,&(cs->out_south),&(cs->out_state));
		
		output_valid(cfg->valid_east,cs->alu_out,&(cs->out_east));
		output_valid(cfg->valid_south,cs->alu_out,&(cs->out_south));
		output_valid(cfg->valid_state,cs->alu_out,&(cs->out_state));
		
		write_policy(cfg->state_wp, cs->out_state, &(cs->cur_state));
	
	}
}

static void md_compute(const struct configuration* cfg, struct cell_state* cs) {
	if ( cs->mul_a.valid == 0 || cs->mul_b.valid == 0 ) {
		cs->mul_out.valid = 0;
		cs->mul_out.value = 0;
		return;
	}
	cs->mul_out.valid = 1;
	switch ( cfg->mul_op ) {
	case mul_mul:
		cs->mul_out.value = ((op_t) cs->mul_a.value) * ((op_t)cs->mul_b.value);
		break;
	case mul_div:
		if ( cs->mul_b.value == 0 ) {
			/* Div by zero */
			if ( cs->mul_a.value < 0 ) {
				cs->mul_out.value = 0x8000;
			} else {
				cs->mul_out.value = 0x7FFF;
			}
		} else {
			cs->mul_out.value = cs->mul_a.value / cs->mul_b.value;
		}
		break;
	default:
		fprintf(stderr,"Err. unknown mul op\n");
		exit(-1);
	}
}

static void alu_compute(const struct configuration* cfg, struct cell_state* cs) {
	
	if ( cs->alu_a.valid == 0 || cs->alu_b.valid == 0 ) {
		cs->alu_out.valid = 0;
		cs->alu_out.value = 0;
		return;
	}
	
	cs->alu_out.valid = 1;
	switch( cfg->alu_op) {
	case alu_add:
		cs->alu_out.value = cs->alu_a.value + cs->alu_b.value;
		break;
	case alu_sub:
		cs->alu_out.value = cs->alu_a.value - cs->alu_b.value;
		break;
	case alu_slr:
		if ( (uop_t)cs->alu_b.value <= 15 ) {
			cs->alu_out.value = (uop_t)cs->alu_a.value >> (uop_t)cs->alu_b.value;
		} else {
			cs->alu_out.value = 0;
		}
		break;
	case alu_sll:
		if ( (uop_t)cs->alu_b.value <= 15 ) {
			cs->alu_out.value = (uop_t)cs->alu_a.value << (uop_t)cs->alu_b.value;
		} else {
			cs->alu_out.value = 0;
		}
		break;
	case alu_or:
		cs->alu_out.value = cs->alu_a.value | cs->alu_b.value;
		break;
	case alu_nor:
		cs->alu_out.value = ~(cs->alu_a.value | cs->alu_b.value);
		break;
	case alu_and:
		cs->alu_out.value = cs->alu_a.value & cs->alu_b.value;
		break;
	case alu_xor:
		cs->alu_out.value = cs->alu_a.value ^ cs->alu_b.value;
		break;
	default:
		fprintf(stderr,"Huh.. unknown aluop\n");
		exit(-1);
	}
}

static void copy_to_mux(struct cell_state* cs, enum source conf, struct cell_op* target) {
	
	switch ( conf ) {
	case north:
		target->valid = cs->north.valid;
		target->value = cs->north.value;
		break;
	case west:
		target->valid = cs->west.valid;
		target->value = cs->west.value;
		break;
	case state:
		target->valid = cs->state.valid;
		target->value = cs->state.value;
		break;
	case mulout:
		target->valid = cs->mul_out.valid;
		target->value = cs->mul_out.value;
		break;
	case aluout:
		target->valid = cs->alu_out.valid;
		target->value = cs->alu_out.value;
		break;
	default:
		fprintf(stderr,"ehm.. fail!\n");
		exit(-1);
	}
}

static void auto_generate_config(struct configuration* cfg) {
	cfg->alu_op = rand()%8;
	cfg->mul_a = rand()%4;
	/* Don't allow backfeed */
	do {
		cfg->alu_a = rand()%4;
		cfg->alu_b = rand()%4;
		/* Don't allow loopbacks */
		if ( cfg->mul_a == 3 ) 
			cfg->mul_a++;
		cfg->mul_b = rand()%4;
		if ( cfg->mul_b == 3 ) 
			cfg->mul_b++;
	
	} while ( ( cfg->alu_a == mulout || cfg->alu_b == mulout ) 
		&& ( cfg->mul_a == aluout || cfg->mul_b == aluout ) );
	
	cfg->mul_op = rand()%2;
	cfg->east = rand()%5;
	cfg->south = rand()%5;
	cfg->state = rand()%5;
	cfg->init_value = rand()%(0xFFFF);
	
	cfg->sel_north = rand()%3;
	cfg->sel_west = rand()%3;
	
	cfg->sw_east_south = rand()%3;
	cfg->sw_south_state = rand()%3;
	cfg->sw_state_east = rand()%3;
	
	cfg->valid_east = rand()%4;
	cfg->valid_south = rand()%4;
	cfg->valid_state = rand()%4;
	
	cfg->state_wp = rand()%2;
}

static int generate_config_stream(struct configuration cfg, char** new_conf) {
	FILE* input;
	FILE* output;
	char file_buffer[512] = {0,};
	int res = 0;
	
	input = fopen(IN_FILE,"w");
	if ( input == NULL ) {
		fprintf(stderr,"Unable to open input file\n");
		return 1;
	}
	
	write_config(input,cfg);
	fclose(input);
	
	if ( ( res = system(TRANSLATOR)) ) {
		fprintf(stderr,"Execution of assembler failed (%d) \n", res);
		return 2;
	}
	
	output = fopen(OUT_FILE,"r");
	
	if ( output == NULL ) {
		fprintf(stderr,"Huh? No outputfile found\n");
		return 3;
	}
	
	/* Skip header */
	fgets(file_buffer,512,output);
	fgets(file_buffer,512,output);
	fclose(output);
	
	if ( new_conf == NULL ) {
		fprintf(stderr,"Internal pointer error\n");
		return 5;
	}
	
	*new_conf = strdup(file_buffer);
	if ( new_conf == NULL ) {
		fprintf(stderr,"Memory allocation error\n");
		return 4;
	}
	
	return 0;
}


static void write_config(FILE* fd,struct configuration cfg) {
	fprintf(fd,"cell [0][0] {\n");
	fprintf(fd,"\taluout = %s %s %s;\n",source_ops[cfg.alu_a],alu_ops[cfg.alu_op],source_ops[cfg.alu_b]);
	fprintf(fd,"\tmulout = %s %s %s;\n",source_ops[cfg.mul_a],mul_ops[cfg.mul_op],source_ops[cfg.mul_b]);
	fprintf(fd,"\teast = %s;\n",source_ops[cfg.east]);
	fprintf(fd,"\tsouth = %s;\n",source_ops[cfg.south]);
	fprintf(fd,"\tstate = %s;\n",source_ops[cfg.state]);
	fprintf(fd,"\tinit {\n");
	fprintf(fd,"\t\tstate = %d;\n",cfg.init_value);
	fprintf(fd,"\t}\n");
	write_select(fd,"north","west",cfg.sel_north);
	write_select(fd,"west","north",cfg.sel_west);
	write_switches(fd,cfg);
	write_valid(fd,"state",cfg.valid_state);
	write_valid(fd,"south",cfg.valid_south);
	write_valid(fd,"east",cfg.valid_east);
	fprintf(fd,"\twrite(state) = %s;\n",write_policy_ops[cfg.state_wp]);
	fprintf(fd,"}\n");
}

static void write_switches(FILE* fd, struct configuration cfg) {
	enum out_switch itr = sw_zero;
	while ( itr <= sw_negative ) {
		if ( 	cfg.sw_east_south != itr 
			&& cfg.sw_state_east != itr
			&& cfg.sw_south_state != itr
		) {
			itr++;
			continue;
		}
		
		fprintf(fd,"\tif( %s(aluout) ) {\n", out_switch_ops[itr]);
		if ( cfg.sw_east_south == itr ) {
			fprintf(fd,"\t\tswitch(east,south);\n");
		}
		if ( cfg.sw_state_east == itr ) {
			fprintf(fd,"\t\tswitch(east,state);\n");
		}
		if ( cfg.sw_south_state == itr ) {
			fprintf(fd,"\t\tswitch(south,state);\n");
		}
		fprintf(fd,"\t}\n");
		itr++;
	}

}

static void write_select(FILE* fd, const char* target, const char* alt, enum in_select policy) {
	if ( policy != in_fixed ) {
		fprintf(fd,"\tif( %s ", in_select_ops[policy]);
		if ( policy == in_notvalid_other ) {
			fprintf(fd,"( %s ) ){\n",target);
		} else if ( policy == in_other_valid ) {
			fprintf(fd,"( %s ) ){\n",alt);
		}
		fprintf(fd,"\t\t%s = %s;\n",target,alt);
		fprintf(fd,"\t}\n");
	}
}

static void write_valid(FILE* fd, const char* target, enum out_valid policy) {
	switch ( policy ) {
	case out_zero:
		fprintf(fd,"\tvalid(%s) = zero(aluout);\n",target);
		break;
	case out_negative:
		fprintf(fd,"\tvalid(%s) = negative(aluout);\n",target);
		break;
	case out_never:
		fprintf(fd,"\tvalid(%s) = never;\n",target);
		break;
	}
}

static void input_select(enum in_select policy, struct cell_op primary, struct cell_op secondary, struct cell_op* target) {
	
	switch ( policy ) {
	case in_fixed:
		*target = primary;
		break;
	case in_notvalid_other:
		fprintf(stderr,"Non fixed\n");
		if ( primary.valid == 0 ) {
			*target = secondary;
		} else {
			*target = primary;
		}
		break;
	case in_other_valid:
		fprintf(stderr,"Non fixed\n");
		if ( secondary.valid == 1 ) {
			*target = secondary;
		} else {
			*target = primary;
		}
		break;
	default:
		fprintf(stderr,"Unknown select policy\n");
		return;
	}
}

static void output_switch(enum out_switch policy, struct cell_op alu_out, struct cell_op primary, struct cell_op secondary, 
				struct cell_op* prim_target, struct cell_op* sec_target) {

	switch ( policy ) {
	case sw_fixed: 
		*prim_target = primary;
		*sec_target = secondary;
		break;
	case sw_zero:
		if ( alu_out.valid == 1 && alu_out.value == 0 ) {
			*prim_target = secondary;
			*sec_target = primary;
		} else {
			*prim_target = primary;
			*sec_target = secondary;
		}
		break;
	
	case sw_negative:
		if ( alu_out.valid == 1 && alu_out.value < 0 ) {
			*prim_target = secondary;
			*sec_target = primary;
		} else {
			*prim_target = primary;
			*sec_target = secondary;
		}
		break;
	default:
		fprintf(stderr,"Unknown switch policy\n");
		return;
	}
}

static void output_valid(enum out_valid policy, struct cell_op alu_out, struct cell_op *target ) {
	
	switch ( policy ) {
	case out_fixed:
		/* Valid remains unchanged */
		break;
	case out_zero:
		if ( alu_out.value == 0 ) {
			target->valid = target->valid & alu_out.valid;
		} else {
			target->valid = 0;
		}
		break;
	
	case out_negative:
		if ( alu_out.value < 0 ) {
			target->valid = target->valid & alu_out.valid;
		} else {
			target->valid = 0;
		}
		break;
	
	case out_never:
		target->valid = 0;
		break;
	
	
	default:
		fprintf(stderr,"Unknown valid policy\n");
		return;
	}

}

static void write_policy(enum write_policy policy, struct cell_op source, struct cell_op* target ) {

	switch ( policy ) {
	case wp_always:
		*target = source;
		break;
	case wp_onvalid:
		if ( source.valid == 1 ) {
			*target = source;
		}
		break;
	default:
		fprintf(stderr,"Unknown write policy\n");
		return;
	}

}
