library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.common_defs.all;

entity multiplydivide_testbench is
end entity;

architecture md_test of multiplydivide_testbench is
	constant size		: integer := 32;

	signal clk		: std_logic;
	signal reset		: std_logic;
	
	signal opcode_i		: std_logic;
	signal valid_a_i	: std_logic;
	signal valid_b_i	: std_logic;
	
	signal data_a_i		: std_logic_vector(size -1 downto 0);
	signal data_b_i		: std_logic_vector(size -1 downto 0);

	signal valid_output_o	: std_logic;
	signal output_o		: std_logic_vector(size -1 downto 0);
	signal running		: std_logic := '1';
begin
	uut : multiplydivider
		generic map(
			width => size
		)
		port map(
			clk => clk,
			reset => reset,
			opcode => opcode_i,
			valid_a => valid_a_i,
			valid_b => valid_b_i,
			data_a => data_a_i,
			data_b => data_b_i,
			valid_output => valid_output_o,
			output => output_o
		);
	
	
	clock_gen : process is
	begin
		while running='1' loop
			clk <= '1';
			wait for 50 ns;
			clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	manual_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(clk);
			wait until rising_edge(clk);
			valid_a_i <= '0';
			valid_b_i <= '0';
			reset <= '1';
		end procedure;
	
		
		procedure perform_simulation(
			constant da	:in std_logic_vector(size-1 downto 0);
			constant db	:in std_logic_vector(size-1 downto 0);
			constant op	:in std_logic;
			constant result	:in std_logic_vector(size-1 downto 0)
			)
		is
			variable counter	:integer	:= 0;
		begin
			reset_uut;
			opcode_i  <= op;
			valid_a_i <= '1';
			data_a_i  <= da;
			
			valid_b_i   <= '1';
			data_b_i  <= db;
			
			-- Should be in start state now.
			-- Result should be ready after size cycles +1
			while counter <= size +2 loop
				wait until rising_edge(clk);
				counter := counter + 1;
			end loop;
			
			wait for 1 ns; -- Allow for propagation of signals 
			assert valid_output_o = '1' report "Valid output not set" severity failure;
			
			assert output_o = result report "Incorrect result " 
				& integer'image(to_integer(signed(da)))
				& " "
				& integer'image(to_integer(signed(db)))
				& " = "
				& integer'image(to_integer(signed(result))) 
				& " != " 
				& integer'image(to_integer(signed(output_o))) severity failure;
		
		end procedure;
	begin
		data_a_i <= (data_a_i'range => '0');
		data_b_i <= (data_b_i'range => '0');
		valid_a_i <= '0';
		valid_b_i <= '0';
		opcode_i <= '0';
		reset_uut;
		
		-- Basic positive I/O-test X*1=X and X/1=X
		perform_simulation(X"00000001",X"00000001",'0',X"00000001");
		perform_simulation(X"00000001",X"00000001",'1',X"00000001");
		perform_simulation(X"00000232",X"00000001",'1',X"00000232");
		perform_simulation(X"04535BFD",X"00000001",'0',X"04535BFD");
		perform_simulation(X"00540232",X"00000001",'0',X"00540232");
		
		-- Basic negative I/O-test -X*1=-X and -X/1=-X
		perform_simulation(X"FFFFFFFF",X"00000001",'0',X"FFFFFFFF");
		perform_simulation(X"FFFFFFFF",X"00000001",'1',X"FFFFFFFF");
		perform_simulation(X"80000232",X"00000001",'1',X"80000232");
		perform_simulation(X"84535BFD",X"00000001",'0',X"84535BFD");
		perform_simulation(X"80540232",X"00000001",'0',X"80540232");
		
		-- Test zero multiplicand and dividend
		-- Result should be zero
		perform_simulation(X"00000000",X"00000001",'0',X"00000000");
		perform_simulation(X"00000000",X"00000001",'1',X"00000000");
		-- Random positive numbers as multiplier / dividend
		perform_simulation(X"00000000",X"00000242",'0',X"00000000");
		perform_simulation(X"00000000",X"00000543",'1',X"00000000");
		-- Random negative numbers as multiplier / dividend
		perform_simulation(X"00000000",X"80000012",'0',X"00000000");
		perform_simulation(X"00000000",X"80000053",'1',X"00000000");
		
		-- Test basic multiplication
		-- Positive operands 
		perform_simulation(X"00000002",X"00000003",'0',X"00000006");
		-- Operand order should not matter on multiply 
		perform_simulation(X"000034EF",X"000000EF",'0',X"00316B21");
		perform_simulation(X"000000EF",X"000034EF",'0',X"00316B21");
		perform_simulation(X"00FFFFFF",X"00000003",'0',X"02FFFFFD");
		-- Negative operands - Results should be positive
		perform_simulation(X"FFFFFFFE",X"FFFFFFFD",'0',X"00000006");
		-- Operand order should not matter on multiply 
		perform_simulation(X"FFFFCB11",X"FFFFFF11",'0',X"00316B21");
		perform_simulation(X"FFFFFF11",X"FFFFCB11",'0',X"00316B21");
		-- Test large positive result possible from two negative numbers
		perform_simulation(X"FF000001",X"FFFFFFFD",'0',X"02FFFFFD");
		
		
		-- Test zero multiplier
		perform_simulation(X"00000001",X"00000000",'0',X"00000000");
		perform_simulation(X"00000242",X"00000000",'0',X"00000000");
		perform_simulation(X"00000000",X"00000000",'0',X"00000000");
		perform_simulation(X"80000000",X"00000000",'0',X"00000000");
		
		-- Test true signed multiplication
		-- -1 * 1 = -1
		perform_simulation(X"00000001",X"FFFFFFFF",'0',X"FFFFFFFF");
		perform_simulation(X"FFFFFFFF",X"00000001",'0',X"FFFFFFFF");
		-- Positive * Negative and vice versa.
		perform_simulation(X"00000432",X"FFFFFF06",'0',X"FFFBE72C");
		perform_simulation(X"00654345",X"FFFFFFFD",'0',X"FED03631");
		perform_simulation(X"FFFF2F23",X"00000234",'0',X"FE33D91C");
		perform_simulation(X"FF3E4245",X"00000043",'0',X"CD4B580F");
		-- Large operands
		perform_simulation(X"80000000",X"00000001",'0',X"80000000");
		perform_simulation(X"00000001",X"80000000",'0',X"80000000");
		
		-- Trunkated overflow
		-- Positive
		perform_simulation(X"40000000",X"40000000",'0',X"00000000");
		perform_simulation(X"40000000",X"43453554",'0',X"00000000");
		perform_simulation(X"40000000",X"4345355F",'0',X"C0000000");
		perform_simulation(X"47867865",X"4345355F",'0',X"440C967B");
		-- Negative
		perform_simulation(X"80000000",X"80000000",'0',X"00000000");
		perform_simulation(X"80000000",X"89827358",'0',X"00000000");
		perform_simulation(X"80000000",X"8892735F",'0',X"80000000");
		perform_simulation(X"C9874395",X"39827358",'0',X"88852A38");
		
		
		
		-- Test basic division
		perform_simulation(X"00000002",X"00000002",'1',X"00000001");
		perform_simulation(X"000000FF",X"00000005",'1',X"00000033");
		-- Test division were remainder is non zero
		perform_simulation(X"00000008",X"00000003",'1',X"00000002");
		perform_simulation(X"00353453",X"00002343",'1',X"00000182");
		-- Operand order should matter on division
		perform_simulation(X"00003448",X"000000EF",'1',X"00000038");
		perform_simulation(X"000000EF",X"000034EF",'1',X"00000000");
		-- Negative operands - Results should be positive
		perform_simulation(X"FFFFFFFC",X"FFFFFFFE",'1',X"00000002");
		-- Operand ordher should matter on division 
		perform_simulation(X"FFFFCB11",X"FFFFFF11",'1',X"00000038");
		perform_simulation(X"FFFFFF11",X"FFFFCB11",'1',X"00000000");
		-- Test large positive result possible from two negative numbers
		perform_simulation(X"FF000001",X"FFFFFFFD",'1',X"00555555");
		
		-- Test division containing -MAX_INT as this value causes trouble
		-- during abs()
		perform_simulation(X"80000000",X"00000002",'1',X"C0000000");
		perform_simulation(X"80000000",X"00000023",'1',X"FC57C57D");
		perform_simulation(X"80000000",X"80000000",'1',X"00000001");
		perform_simulation(X"53456065",X"80000000",'1',X"00000000");
		perform_simulation(X"80000000",X"00000001",'1',X"80000000");
		perform_simulation(X"00000001",X"80000000",'1',X"00000000");
		
		-- Division by zero should yield signed MAX_INT as there is no
		-- exception handling
		perform_simulation(X"00003434",X"00000000",'1',X"7FFFFFFF");
		perform_simulation(X"80003434",X"00000000",'1',X"80000000");
		perform_simulation(X"80000000",X"00000000",'1',X"80000000");
		perform_simulation(X"00000000",X"00000000",'1',X"7FFFFFFF");
		
		-- Test Multiplication / division mix
		perform_simulation(X"FFFFFF11",X"FFFFCB11",'0',X"00316B21");
		perform_simulation(X"00000002",X"00000002",'1',X"00000001");
		perform_simulation(X"FF000001",X"FFFFFFFD",'0',X"02FFFFFD");
		perform_simulation(X"000000FF",X"00000005",'1',X"00000033");
		
		-- End of test bench
		running <= '0';
		wait for 500 ns;
		wait;
	end process;

end architecture;