library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version1_defs.all;

entity auto_rc_master_testbench is
end entity;

architecture rc_master_test of auto_rc_master_testbench is
	constant addr_width	: integer := 16;
	constant header_size	: integer := 24;
	constant rc_units	: integer := 8;
	constant system_version	: integer := 0;
	constant version_width	: integer := 4;
	constant flags_width	: integer := 4;
	constant size_width	: integer := 8;
	constant addr_consume	: integer := 3;
	
	signal clk		: std_logic;
	signal reset		: std_logic;
	
	signal test_clk		: std_logic;
	
	signal ctrl_i		: std_logic_vector(rcm_ctrl_size -1 downto 0);
	signal status_o		: std_logic_vector(rcm_status_size -1 downto 0);
	signal src_base_addr_i	: std_logic_vector(addr_width -1 downto 0);
	signal dest_base_addr_i : std_logic_vector(addr_width -1 downto 0);
	
	signal flags_o		: std_logic_vector(flags_width-1 downto 0);
	
	signal read_addr_o	: std_logic_vector(addr_width -1 downto 0);
	signal read_data_i	: std_logic_vector(header_size -1 downto 0);
	signal write_en_o	: std_logic;
	signal write_addr_o	: std_logic_vector(addr_width-1 downto 0);
	signal write_data_o	: std_logic_vector(header_size -1 downto 0);
	
	signal rcr_rc_o		: std_logic;
	signal rcr_addr_o	: std_logic_vector(addr_width -1 downto 0);
	signal rcr_pattern_o	: std_logic_vector(rc_units-1 downto 0);
	signal rcr_vertical_o	: std_logic;
	signal rcr_offset_o	: std_logic_vector(addr_width -1 downto 0);
	
	signal rcw_rc_o		: std_logic;
	signal rcw_addr_o	: std_logic_vector(addr_width -1 downto 0);
	signal rcw_pattern_o	: std_logic_vector(rc_units-1 downto 0);
	signal rcw_vertical_o	: std_logic;
	signal rcw_offset_o	: std_logic_vector(addr_width -1 downto 0);
	
	
	signal running		: integer := 1;
begin
	uut : rc_master
		generic map(
			addr_width => addr_width,
			header_size => header_size,
			rc_units => rc_units,
			system_version => system_version,
			version_width => version_width,
			flags_width => flags_width,
			size_width => size_width,
			addr_consume => addr_consume
		)
		port map(
			clk => clk,
			reset => reset,
			status => status_o,
			ctrl => ctrl_i,
			src_base_addr => src_base_addr_i,
			dest_base_addr => dest_base_addr_i,
			flags => flags_o,
			read_addr => read_addr_o,
			read_data => read_data_i,
			write_en => write_en_o,
			write_addr => write_addr_o,
			write_data => write_data_o,
			rcr_rc => rcr_rc_o,
			rcr_addr => rcr_addr_o,
			rcr_pattern => rcr_pattern_o,
			rcr_vertical => rcr_vertical_o,
			rcr_offset => rcr_offset_o,
			rcw_rc => rcw_rc_o,
			rcw_addr => rcw_addr_o,
			rcw_pattern => rcw_pattern_o,
			rcw_vertical => rcw_vertical_o,
			rcw_offset => rcw_offset_o
		);

	clock : process is
	begin
		clk <= '0';
		wait for 10 ns;
		while running = 1 loop
			clk <= '1';
			wait for 50 ns;
			clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	-- For internal use in the testbench
	test_clock : process is
	begin
		while running = 1 loop
			test_clk <= '1';
			wait for 50 ns;
			test_clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		
		variable rcr_v_pos	: integer := 0;
		variable rcw_v_pos	: integer := 0;
		
		procedure perform_simulation (
			constant configured	:std_logic;
			constant source		:std_logic;
			constant drain		:std_logic;
			constant clear_error	:std_logic;
			constant write_allow	:std_logic;
			
			constant flags		:std_logic_vector(flags_width-1 downto 0);
			
			constant src_base_addr	:std_logic_vector(addr_width-1 downto 0);
			constant read_address	:std_logic_vector(addr_width-1 downto 0);
			constant read_header	:std_logic_vector(header_size-1 downto 0);
			constant dest_base_addr	:std_logic_vector(addr_width-1 downto 0);
			constant write_address	:std_logic_vector(addr_width-1 downto 0);
			constant write_header	:std_logic_vector(header_size-1 downto 0);
			
			constant rcr_rc		:std_logic;
			constant rcr_addr	:std_logic_vector(addr_width-1 downto 0);
			constant rcr_pattern	:std_logic_vector(rc_units-1 downto 0);
			constant rcr_v_pattern	:std_logic_vector(rc_units-1 downto 0);
			constant rcr_offset	:std_logic_vector(addr_width-1 downto 0);
			
			constant rcw_rc		:std_logic;
			constant rcw_addr	:std_logic_vector(addr_width-1 downto 0);
			constant rcw_pattern	:std_logic_vector(rc_units-1 downto 0);
			constant rcw_v_pattern	:std_logic_vector(rc_units-1 downto 0);
			constant rcw_offset	:std_logic_vector(addr_width-1 downto 0)
			)
		is 
			variable r_addr		: std_logic_vector(addr_width-1 downto 0);
		begin
			ctrl_i(0) <= source;
			ctrl_i(1) <= drain;
			ctrl_i(2) <= clear_error;
			src_base_addr_i <= src_base_addr;
			dest_base_addr_i <= dest_base_addr;
			
			r_addr := read_addr_o;
			
			if clear_error = '1' then
				assert status_o(2) = '1' 
					report "Tried to clear non-error state"
					severity warning;
			end if;
			
			wait until rising_edge(clk);
			if read_address = r_addr then
				read_data_i <= read_header;
			else 
				read_data_i <= (others => '0');
			end if;
			ctrl_i <= (others => '0');
			
			wait for 2 ns; --propagation
			
			if write_allow = '0' or configured = '0' then
				assert write_en_o = '0' 
					report "Illegal write to memory"
					severity failure;
			end if;
			
			assert rcr_rc = rcr_rc_o 
				report "Reader RC signal incorrect. Expected " &
				std_logic'image(rcr_rc) &
				" was " &
				std_logic'image(rcr_rc_o)
				severity failure;
			
			if rcr_rc_o = '1' then 
				assert rcr_addr =  rcr_addr_o 
					report "Wrong reader address. Expected " &
					Image(rcr_addr) &
					" was " &
					Image(rcr_addr_o)
					severity failure;
				assert rcr_pattern =  rcr_pattern_o 
					report "Wrong reader pattern. Expected " &
					Image(rcr_pattern) &
					" was " &
					Image(rcr_pattern_o)
					severity failure;
				assert rcr_offset = rcr_offset_o
					report "Wrong reader offset. Expected " &
					Image(rcr_offset) &
					" was " &
					Image(rcr_offset_o)
					severity failure;
				assert rcr_v_pattern(rcr_v_pos) = rcr_vertical_o 
					report "Wrong vertical pattern"
					severity failure;
				assert flags = flags_o 
					report "Wrong flags output"
					severity failure;
				-- Post inc
				rcr_v_pos := rcr_v_pos +1;
			else 
				rcr_v_pos := 0;
			end if;
			
			assert rcw_rc = rcw_rc_o
				report "Writer RC signal incorrect"
				severity failure;
			
			if rcw_rc_o = '1' then
				assert rcw_addr =  rcw_addr_o 
					report "Wrong writer address. Expected " &
					Image(rcw_addr) &
					" was " &
					Image(rcw_addr_o)
					severity failure;
				assert rcw_pattern = rcw_pattern_o
					report "Wrong writer pattern. Expected " &
					Image(rcw_pattern) &
					" was " &
					Image(rcw_pattern_o)
					severity failure;
				assert rcw_offset = rcw_offset_o
					report "Wrong writer offset. Expected " &
					Image(rcw_offset) &
					" was " &
					Image(rcw_offset_o)
					severity failure;
				assert rcw_v_pattern(rcw_v_pos) = rcw_vertical_o
					report "Wrong vertical pattern"
					severity failure;
				rcw_v_pos := rcw_v_pos +1;
			else
				rcw_v_pos := 0;
			end if;
			
			wait until rising_edge(test_clk); 
			
			if write_address = write_addr_o and write_en_o = '1' then
				assert write_header = write_data_o
					report "Memory output incorrect. Expected " 
						& Image(write_header)
						& " was " 
						& Image(write_data_o)
						severity failure;
			end if;
			
		end procedure;
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-rc_master.txt";
			variable L: line;
			--variable good: boolean;
			
			variable configured	:std_logic;
			variable source		:std_logic;
			variable drain		:std_logic;
			variable clear_error	:std_logic;
			variable write_allow	:std_logic;
			
			variable flags		:std_logic_vector(flags_width-1 downto 0);
			
			variable src_base_addr	:std_logic_vector(addr_width-1 downto 0);
			variable read_address	:std_logic_vector(addr_width-1 downto 0);
			variable read_header	:std_logic_vector(header_size-1 downto 0);
			variable dest_base_addr	:std_logic_vector(addr_width-1 downto 0);
			variable write_address	:std_logic_vector(addr_width-1 downto 0);
			variable write_header	:std_logic_vector(header_size-1 downto 0);
			
			variable rcr_rc		:std_logic;
			variable rcr_addr	:std_logic_vector(addr_width-1 downto 0);
			variable rcr_pattern	:std_logic_vector(rc_units-1 downto 0);
			variable rcr_v_pattern	:std_logic_vector(rc_units-1 downto 0);
			variable rcr_offset	:std_logic_vector(addr_width-1 downto 0);
			
			variable rcw_rc		:std_logic;
			variable rcw_addr	:std_logic_vector(addr_width-1 downto 0);
			variable rcw_pattern	:std_logic_vector(rc_units-1 downto 0);
			variable rcw_v_pattern	:std_logic_vector(rc_units-1 downto 0);
			variable rcw_offset	:std_logic_vector(addr_width-1 downto 0);
			
		begin
			
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				
				read(L,configured);
				read(L,source);
				read(L,drain);
				read(L,clear_error);
				read(L,write_allow);
				
				read(L,flags);
				
				read(L,src_base_addr);
				read(L,read_address);
				read(L,read_header);
				read(L,dest_base_addr);
				read(L,write_address);
				read(L,write_header);
				
				read(L,rcr_rc);
				read(L,rcr_addr);
				read(L,rcr_pattern);
				read(L,rcr_v_pattern);
				read(L,rcr_offset);
				
				read(L,rcw_rc);
				read(L,rcw_addr);
				read(L,rcw_pattern);
				read(L,rcw_v_pattern);
				read(L,rcw_offset);
				
				perform_simulation(configured,source,drain,clear_error,write_allow,
					flags,src_base_addr,read_address,read_header,
					dest_base_addr,write_address,write_header,
					rcr_rc,rcr_addr,rcr_pattern,rcr_v_pattern,rcr_offset,
					rcw_rc,rcw_addr,rcw_pattern,rcw_v_pattern,rcw_offset);
			end loop;
		end procedure;
		
	begin
		ctrl_i <= (others => '0');
		src_base_addr_i <= (others => '0');
		dest_base_addr_i <= (others => '0');
		
		reset_uut;
		
		-- Check reset 
		
		-- Sync with clocks
		wait until rising_edge(test_clk);
		
		random_stimuli;
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;