library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version2_defs.all;

entity rc_master_testbench is
end entity;

architecture rc_master_test of rc_master_testbench is
	constant addr_width	: integer := 8;
	constant header_size	: integer := 24;
	constant rc_units	: integer := 8;
	constant system_version	: integer := 1;
	constant version_width	: integer := 4;
	constant flags_width	: integer := 4;
	constant size_width	: integer := 8;
	constant addr_consume	: integer := 5;
	
	signal clk		: std_logic;
	signal reset		: std_logic;
	
	signal test_clk		: std_logic;
	
	signal ctrl_i		: std_logic_vector(rcm_ctrl_size -1  downto 0);
	signal status_o		: std_logic_vector(rcm_status_size -1 downto 0);
	signal src_base_addr_i	: std_logic_vector(addr_width -1 downto 0);
	signal dest_base_addr_i : std_logic_vector(addr_width -1 downto 0);
	
	signal read_addr_o	: std_logic_vector(addr_width -1 downto 0);
	signal read_data_i	: std_logic_vector(header_size -1 downto 0);
	signal write_en_o	: std_logic;
	signal write_addr_o	: std_logic_vector(addr_width-1 downto 0);
	signal write_data_o	: std_logic_vector(header_size -1 downto 0);
	
	signal rcr_rc_o		: std_logic;
	signal rcr_addr_o	: std_logic_vector(addr_width -1 downto 0);
	signal rcr_pattern_o	: std_logic_vector(rc_units-1 downto 0);
	signal rcr_vertical_o	: std_logic;
	signal rcr_offset_o	: std_logic_vector(addr_width -1 downto 0);
	
	signal rcw_rc_o		: std_logic;
	signal rcw_addr_o	: std_logic_vector(addr_width -1 downto 0);
	signal rcw_pattern_o	: std_logic_vector(rc_units-1 downto 0);
	signal rcw_vertical_o	: std_logic;
	signal rcw_offset_o	: std_logic_vector(addr_width -1 downto 0);
	
	
	signal running		: integer := 1;
begin
	uut : rc_master
		generic map(
			addr_width => addr_width,
			header_size => header_size,
			rc_units => rc_units,
			system_version => system_version,
			version_width => version_width,
			flags_width => flags_width,
			size_width => size_width,
			addr_consume => addr_consume
		)
		port map(
			clk => clk,
			reset => reset,
			status => status_o,
			ctrl => ctrl_i,
			src_base_addr => src_base_addr_i,
			dest_base_addr => dest_base_addr_i,
			read_addr => read_addr_o,
			read_data => read_data_i,
			write_en => write_en_o,
			write_addr => write_addr_o,
			write_data => write_data_o,
			rcr_rc => rcr_rc_o,
			rcr_addr => rcr_addr_o,
			rcr_pattern => rcr_pattern_o,
			rcr_vertical => rcr_vertical_o,
			rcr_offset => rcr_offset_o,
			rcw_rc => rcw_rc_o,
			rcw_addr => rcw_addr_o,
			rcw_pattern => rcw_pattern_o,
			rcw_vertical => rcw_vertical_o,
			rcw_offset => rcw_offset_o
		);

	clock : process is
	begin
		clk <= '0';
		wait for 10 ns;
		while running = 1 loop
			clk <= '1';
			wait for 50 ns;
			clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	-- For internal use in the testbench
	test_clock : process is
	begin
		while running = 1 loop
			test_clk <= '1';
			wait for 50 ns;
			test_clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	manual_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		
		variable rcr_v_pos	: integer := 0;
		variable rcw_v_pos	: integer := 0;
		
		procedure perform_simulation (
			constant configured	:std_logic;
			constant source		:std_logic;
			constant drain		:std_logic;
			constant clear_error	:std_logic;
			
			constant allow_write	:std_logic;
			constant src_base_addr	:std_logic_vector(addr_width-1 downto 0);
			constant read_address	:std_logic_vector(addr_width-1 downto 0);
			constant read_header	:std_logic_vector(header_size-1 downto 0);
			constant dest_base_addr	:std_logic_vector(addr_width-1 downto 0);
			constant write_address	:std_logic_vector(addr_width-1 downto 0);
			constant write_header	:std_logic_vector(header_size-1 downto 0);
			
			constant rcr_rc		:std_logic;
			constant rcr_addr	:std_logic_vector(addr_width-1 downto 0);
			constant rcr_pattern	:std_logic_vector(rc_units-1 downto 0);
			constant rcr_v_pattern	:std_logic_vector(rc_units-1 downto 0);
			constant rcr_offset	:std_logic_vector(addr_width-1 downto 0);
			
			constant rcw_rc		:std_logic;
			constant rcw_addr	:std_logic_vector(addr_width-1 downto 0);
			constant rcw_pattern	:std_logic_vector(rc_units-1 downto 0);
			constant rcw_v_pattern	:std_logic_vector(rc_units-1 downto 0);
			constant rcw_offset	:std_logic_vector(addr_width-1 downto 0);
			constant status		:std_logic_vector(rcm_status_size -1 downto 0)
			)
		is 
			variable mem_output	: std_logic_vector(header_size-1 downto 0);
			variable r_addr		: std_logic_vector(addr_width-1 downto 0);
		begin
			ctrl_i(0) <= source;
			ctrl_i(1) <= drain;
			ctrl_i(2) <= clear_error;
			src_base_addr_i <= src_base_addr;
			dest_base_addr_i <= dest_base_addr;
			
			r_addr := read_addr_o;
			
			wait until rising_edge(clk);
			if read_address = r_addr then
				read_data_i <= read_header;
			else 
				read_data_i <= (others => '0');
			end if;
			ctrl_i <= (others => '0');
			mem_output := (others => '0');
			
			wait for 2 ns; --propagation
			
			if allow_write = '0' or configured = '0' then
				assert write_en_o = '0' 
					report "Illegal write to memory"
					severity failure;
			end if;
			
			assert rcr_rc = rcr_rc_o 
				report "Reader RC signal incorrect"
				severity failure;
			
			if rcr_rc_o = '1' then 
				assert rcr_addr =  rcr_addr_o 
					report "Wrong reader address. Expected " &
					Image(rcr_addr) &
					" was " &
					Image(rcr_addr_o)
					severity failure;
				assert rcr_pattern =  rcr_pattern_o 
					report "Wrong reader pattern. Expected " &
					Image(rcr_pattern) &
					" was " &
					Image(rcr_pattern_o)
					severity failure;
				assert rcr_offset = rcr_offset_o
					report "Wrong reader offset. Expected " &
					Image(rcr_offset) &
					" was " &
					Image(rcr_offset_o)
					severity failure;
				assert rcr_v_pattern(rcr_v_pos) = rcr_vertical_o 
					report "Wrong vertical pattern"
					severity failure;
				-- Post inc
				rcr_v_pos := rcr_v_pos +1;
			else 
				rcr_v_pos := 0;
			end if;
			
			assert rcw_rc = rcw_rc_o
				report "Writer RC signal incorrect"
				severity failure;
			
			if rcw_rc_o = '1' then
				assert rcw_addr =  rcw_addr_o 
					report "Wrong writer address. Expected " &
					Image(rcw_addr) &
					" was " &
					Image(rcw_addr_o)
					severity failure;
				assert rcw_pattern = rcw_pattern_o
					report "Wrong writer pattern. Expected " &
					Image(rcw_pattern) &
					" was " &
					Image(rcw_pattern_o)
					severity failure;
				assert rcw_offset = rcw_offset_o
					report "Wrong writer offset. Expected " &
					Image(rcw_offset) &
					" was " &
					Image(rcw_offset_o)
					severity failure;
				assert rcw_v_pattern(rcw_v_pos) = rcw_vertical_o
					report "Wrong vertical pattern"
					severity failure;
				rcw_v_pos := rcw_v_pos +1;
			else
				rcw_v_pos := 0;
			end if;
			
			if write_address = write_addr_o then 
				mem_output := write_data_o;
			end if;
			
			wait until rising_edge(test_clk); 
			
			if drain = '1' and write_address = write_addr_o and write_en_o = '1' then
				assert write_header = mem_output
					report "Memory output incorrect. Expected " 
						& Image(write_header)
						& " was " 
						& Image(mem_output)
						severity failure;
			end if;
			
			assert status = status_o 
				report "Incorrect status. Expected "
					& Image(status) 
					& " was "
					& Image(status_o)
					severity failure;
			
		end procedure;
	
	begin
		ctrl_i <= (others => '0');
		src_base_addr_i <= (others => '0');
		dest_base_addr_i <= (others => '0');
		
		reset_uut;
		
		-- Check reset
		assert write_en_o = '0' report "Write high after reset" severity failure;
		
		-- Sync with clocks
		wait until rising_edge(test_clk);
--		perform_simulation(configured,source,drain,clear_error,write_allow
--			src_base_addr,read_address,read_header,
--			dest_base_addr,write_address,write_header,
--			rcr_rc,rcr_addr,rcr_pattern,rcr_v_pattern,rcr_offset,
--			rcw_rc,rcw_addr,rcw_pattern,rcw_v_pattern,rcw_offset);
--		
		-- Check IDLE
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"100101",
			X"01",X"01",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"100101",
			X"01",X"01",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		
		-- Standard operation
		-- Full read
--		perform_simulation('0','1','0',
--			X"00",X"00",X"10808",
--			X"00",X"00",X"FFFFF",
		-- Start reconfiguration
		perform_simulation('0','1','0','0','0',
			X"10",X"00",X"FFFFFF",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		-- CTRL parameters should be locked down during reconfig
		-- Allow read of configuration header
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"10",X"000808",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'1',X"15",X"FF",X"FF",X"08",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"08",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"08",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"08",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"08",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"08",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"08",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"08",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		
		-- Read new configuration with writeback of previous
		perform_simulation('1','1','1','0','0',
			X"55",X"00",X"FFFFFF",
			X"AB",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		-- CTRL parameters should be locked down during reconfig
		-- Allow read of configuration header
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"55",X"000404",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		-- TODO: Use a separate flag to signalize allow write
		perform_simulation('1','0','0','0','1',
			X"00",X"00",X"000000",
			X"00",X"AB",X"000808",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'1',X"5A",X"55",X"55",X"04",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"04",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"04",
			'1',X"B0",X"FF",X"FF",X"08",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"04",
			'1',X"B0",X"FF",X"FF",X"08",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"04",
			'1',X"B0",X"FF",X"FF",X"08",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"04",
			'1',X"B0",X"FF",X"FF",X"08",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"04",
			'1',X"B0",X"FF",X"FF",X"08",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"04",
			'1',X"B0",X"FF",X"FF",X"08",
			"011");
		-- The writer RC should be active to cycles longer than
		-- the reader RC flag, to allow for config propagation
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"B0",X"FF",X"FF",X"08",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"B0",X"FF",X"FF",X"08",
			"001");
		
		
		-- Check configurated flag setting
		-- Drain config
		perform_simulation('1','0','1','0','0',
			X"00",X"00",X"FFFFFF",
			X"41",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		-- CTRL parameters should be locked down during reconfig
		-- Allow read of configuration header
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		-- Note: We skip two stages here. Fast forward when draining,
		-- no need to fetch new patterns
		-- TODO: Use a separate flag to signalize allow write
		perform_simulation('1','0','0','0','1',
			X"00",X"00",X"000000",
			X"00",X"41",X"000404",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"46",X"55",X"55",X"04",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"46",X"55",X"55",X"04",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"46",X"55",X"55",X"04",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"46",X"55",X"55",X"04",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"46",X"55",X"55",X"04",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"46",X"55",X"55",X"04",
			"011");
		-- The writer RC should be active to cycles longer than
		-- the reader RC flag, to allow for config propagation
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"46",X"55",X"55",X"04",
			"011");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"46",X"55",X"55",X"04",
			"000");
		
		
		-- Error checking. Try to drain twice
		perform_simulation('0','0','1','0','0',
			X"00",X"00",X"000000",
			X"87",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		perform_simulation('0','0','1','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		-- Check that the rc-master sticks to 
		-- error until cleared
		perform_simulation('0','1','1','0','0',
			X"23",X"54",X"364534",
			X"54",X"23",X"675ED0",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','1','0','0','0',
			X"23",X"54",X"364534",
			X"54",X"23",X"675ED0",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','1','0','0',
			X"23",X"54",X"364534",
			X"54",X"23",X"675ED0",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','0','1','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		-- Try other inputs that will generate errors
		-- Oversize configuration
		perform_simulation('0','1','0','0','0',
			X"23",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"23",X"000909",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','0','1','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		-- Newer version header
		perform_simulation('0','1','0','0','0',
			X"23",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"23",X"200404",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','0','1','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		-- Unmappable size
		perform_simulation('0','1','0','0','0',
			X"23",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"23",X"000303",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','0','1','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"100");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		-- Check if errors affect the current config
		perform_simulation('0','1','0','0','0',
			X"54",X"00",X"FFFFFF",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		-- CTRL parameters should be locked down during reconfig
		-- Allow read of configuration header
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"54",X"000202",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'1',X"59",X"11",X"11",X"02",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"59",X"11",X"11",X"02",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"59",X"11",X"11",X"02",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"59",X"11",X"11",X"02",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"59",X"11",X"11",X"02",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"59",X"11",X"11",X"02",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"59",X"11",X"11",X"02",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"59",X"11",X"11",X"02",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		-- Generate error
		perform_simulation('1','1','1','0','0',
			X"43",X"00",X"000000",
			X"12",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"43",X"000403",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"101");
		-- Check if the master changes the drain flags
		-- This should not happen
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"101");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"101");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"101");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"101");
		-- Clear error
		perform_simulation('1','0','0','1','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"101");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		-- Drain
		perform_simulation('1','0','1','0','0',
			X"00",X"00",X"000000",
			X"80",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','1',
			X"00",X"00",X"000000",
			X"00",X"80",X"000202",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'1',X"85",X"11",X"11",X"02",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'1',X"85",X"11",X"11",X"02",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'1',X"85",X"11",X"11",X"02",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'1',X"85",X"11",X"11",X"02",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'1',X"85",X"11",X"11",X"02",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'1',X"85",X"11",X"11",X"02",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'1',X"85",X"11",X"11",X"02",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'1',X"85",X"11",X"11",X"02",
			"000");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		
		reset_uut;
		wait until rising_edge(test_clk);
		
		-- Version 2 test
		perform_simulation('0','1','0','0','0',
			X"10",X"00",X"FFFFFF",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"000");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"10",X"100808",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'1',X"15",X"FF",X"FF",X"10",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"10",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"10",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"10",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"10",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"10",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"10",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"15",X"FF",X"FF",X"10",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"010");
		perform_simulation('0','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		-- V2 configuration writeback
		-- Read new configuration with writeback of previous
		perform_simulation('1','1','1','0','0',
			X"55",X"00",X"FFFFFF",
			X"AB",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"001");
		-- CTRL parameters should be locked down during reconfig
		-- Allow read of configuration header
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"55",X"100404",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		-- TODO: Use a separate flag to signalize allow write
		perform_simulation('1','0','0','0','1',
			X"00",X"00",X"000000",
			X"00",X"AB",X"100808",
			'0',X"00",X"00",X"00",X"00",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"000000",
			'1',X"5A",X"55",X"55",X"08",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"08",
			'0',X"00",X"00",X"00",X"00",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"08",
			'1',X"B0",X"FF",X"FF",X"10",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"08",
			'1',X"B0",X"FF",X"FF",X"10",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"08",
			'1',X"B0",X"FF",X"FF",X"10",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"08",
			'1',X"B0",X"FF",X"FF",X"10",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"08",
			'1',X"B0",X"FF",X"FF",X"10",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'1',X"5A",X"55",X"55",X"08",
			'1',X"B0",X"FF",X"FF",X"10",
			"011");
		-- The writer RC should be active to cycles longer than
		-- the reader RC flag, to allow for config propagation
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"B0",X"FF",X"FF",X"10",
			"011");
		perform_simulation('1','0','0','0','0',
			X"00",X"00",X"000000",
			X"00",X"00",X"FFFFFF",
			'0',X"00",X"00",X"00",X"00",
			'1',X"B0",X"FF",X"FF",X"10",
			"001");
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;