library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version1_defs.all;

entity auto_rc_reader_testbench is
end entity;

architecture rc_reader_test of auto_rc_reader_testbench is
	constant c_size		: integer := 32;
	constant c_width	: integer := 8;
	constant a_width	: integer := 32;
	constant p_width	: integer := 8;
	
	signal config_clk	: std_logic;
	signal data_clk		: std_logic;
	signal reset		: std_logic;
	
	signal test_clk		: std_logic;
	
	signal reconfigure_i	: std_logic;
	signal rc_next_o	: std_logic;
	signal load_address_i	: std_logic_vector(a_width-1 downto 0);
	
	signal vertical_i	: std_logic;
	signal vertical_next_o	: std_logic;
	
	signal address_next_o	: std_logic_vector(a_width-1 downto 0);
	
	signal rc_pattern_i	: std_logic_vector(p_width-1 downto 0);
	signal rc_pattern_n_o	: std_logic_vector(p_width-1 downto 0);
	
	signal read_address_o	: std_logic_vector(a_width-1 downto 0);
	signal read_data_i	: std_logic_vector(c_size-1 downto 0);
	
	signal reconf_out_o	: std_logic_vector(p_width-1 downto 0);
	signal config_out_o	: std_logic_vector(c_width-1 downto 0);
	
	signal unit_offset_i	: std_logic_vector(a_width-1 downto 0);
	signal unit_offset_n_o	: std_logic_vector(a_width-1 downto 0);
	
	signal running		: integer := 1;
begin
	uut : rc_reader
		generic map(
			config_width => c_width,
			config_size => c_size,
			address_width => a_width,
			pattern_width => p_width,
			dummy_config => "00000000000000000000000000000000"
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			reconfigure => reconfigure_i,
			load_address => load_address_i,
			rc_pattern => rc_pattern_i,
			rc_pattern_next => rc_pattern_n_o,
			vertical => vertical_i,
			vertical_next => vertical_next_o,
			read_address => read_address_o,
			read_data => read_data_i,
			reconf_out => reconf_out_o,
			config_out => config_out_o,
			rc_next => rc_next_o,
			address_next => address_next_o,
			unit_offset => unit_offset_i,
			unit_offset_next => unit_offset_n_o
		);

	config_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			config_clk <= '1';
			wait for 50 ns;
			config_clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	data_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			data_clk <= '1';
			wait for 50 ns;
			data_clk <= '0';
			wait for 350 ns;
		end loop;
		wait;
	end process;
	
	-- For internal use in the testbench
	test_clock : process is
	begin
		while running = 1 loop
			test_clk <= '1';
			wait for 50 ns;
			test_clk <= '0';
			wait for 350 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(data_clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		
		procedure perform_simulation (
			constant reconfigure	: std_logic;
			constant rc_pattern	: std_logic_vector(p_width-1 downto 0);
			constant rc_pattern_n	: std_logic_vector(p_width-1 downto 0);
			constant vertical	: std_logic;
			constant rc_to_units	: std_logic_vector(p_width-1 downto 0);
			constant load_address	: std_logic_vector(a_width-1 downto 0);
			constant read_address	: std_logic_vector(a_width-1 downto 0);
			constant read_data	: std_logic_vector(c_size-1 downto 0);
			constant config_data	: std_logic_vector(c_size-1 downto 0);
			constant address_next	: std_logic_vector(a_width-1 downto 0);
			constant unit_offset	: std_logic_vector(a_width-1 downto 0)
			)
		is 
			variable serial_out	: std_logic_vector(c_size-1 downto 0);
			variable next_mem_data	: std_logic_vector(c_size-1 downto 0);
			variable reg_rc		: std_logic;
			variable cycle_counter	: integer := (c_size/c_width);
		begin
			-- Don't play with the signals until the current async write is complete 
			reconfigure_i <= reconfigure;
			load_address_i <= load_address;
			rc_pattern_i <= rc_pattern;
			read_data_i <= (others => '0');
			vertical_i <= vertical;
			unit_offset_i <= unit_offset;
			if reconf_out_o /= (reconf_out_o'range => '0') then
				reg_rc := '1';
			else
				reg_rc := '0';
			end if;

			wait until rising_edge(data_clk);
			reconfigure_i <= '0';
			load_address_i <= (others => '0');
			rc_pattern_i <= (others => '0');
			next_mem_data := (others => '0');
			unit_offset_i <= (others => '0');
			
			while cycle_counter > 1 loop
				-- Check output before config_clk rising edge when not reconfiguring
				wait for 2 ns; --propagation
				read_data_i <= next_mem_data;
				if read_address_o = read_address then 
					next_mem_data := read_data;
				else
					next_mem_data := (others => '0');
				end if;
				wait until rising_edge(config_clk);
				serial_out := config_out_o & serial_out(c_size-1 downto c_width);
				
				cycle_counter := cycle_counter - 1;
			end loop;
			
			wait until rising_edge(test_clk); 
			-- Fetch the last bits on the serial_bus
			serial_out := config_out_o & serial_out(c_size-1 downto c_width);
			
			assert rc_pattern_n_o = rc_pattern_n 
				report "Incorrect pattern to next unit. Expected "
					& Image(rc_pattern_n)
					& " was "
					& Image(rc_pattern_n_o)
					severity failure;
			
			assert rc_to_units = reconf_out_o 
				report "Incorrect reconfiguration signal to cells. Expected "
					& Image(rc_to_units)
					& " was "
					& Image(reconf_out_o)
					severity failure;
			
			assert vertical = vertical_next_o
				report "Incorrect horizontal signal to next unit. Expected "
					& std_logic'image(vertical)
					& " was "
					& std_logic'image(vertical_next_o)
					severity failure;
			
			if reg_rc = '1' then
				assert config_data = serial_out
					report "Config out incorrect. Expected " 
						& Image(config_data)
						& " was " 
						& Image(serial_out)
						severity failure;
			end if;
			
			assert reconfigure = rc_next_o 
				report "RC-signal to next unit incorrect. Expected "
					& std_logic'image(reconfigure)
					& " was " 
					& std_logic'image(rc_next_o)
					severity failure;
			
			if rc_next_o = '1' then
				assert address_next = address_next_o
					report "Base address to next unit incorrect. Expected "
					& Image(address_next)
					& " was " 
					& Image(address_next_o)
					severity failure;
			end if;
			
			assert unit_offset = unit_offset_n_o 
				report "Unit offset propagation incorrect"
				severity failure;
			
		end procedure;
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-rc_reader.txt";
			variable L: line;
			variable good: boolean;
			
			variable rc		:std_logic;
			variable rc_pattern	:std_logic_vector(p_width-1 downto 0);
			variable rc_pattern_next :std_logic_vector(p_width-1 downto 0);
			variable rc_to_units	:std_logic_vector(p_width-1 downto 0);
			
			variable load_address	:std_logic_vector(a_width-1 downto 0);
			variable read_address	:std_logic_vector(a_width-1 downto 0);
			variable address_next	:std_logic_vector(a_width-1 downto 0);
			variable read_data	:std_logic_vector(c_size-1 downto 0);
			variable config_data	:std_logic_vector(c_size-1 downto 0);
			
			variable vertical	: std_logic;
			variable unit_offset	:std_logic_vector(a_width-1 downto 0);
		begin
			
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				
				read(L,rc);
				read(L,rc_pattern);
				read(L,rc_pattern_next);
				read(L,rc_to_units);
				read(L,load_address);
				read(L,read_address);
				read(L,read_data);
				read(L,config_data);
				read(L,address_next);
				
				read(L,vertical);
				
				read(L,unit_offset);
				
				perform_simulation(rc,rc_pattern,rc_pattern_next,vertical,rc_to_units,load_address,read_address,read_data,config_data,address_next,unit_offset);
			end loop;
		end procedure;
		
	begin
		reconfigure_i <= '0';
		reset_uut;
		
		-- Sync with clocks
		wait until rising_edge(test_clk);
		
		random_stimuli;
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;