library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version1_defs.all;

entity auto_rc_writer_testbench is
end entity;

architecture rc_writer_test of auto_rc_writer_testbench is
	constant c_size		: integer := 32;
	constant c_width	: integer := 8;
	constant a_width	: integer := 32;
	constant p_width	: integer := 8;
	
	signal config_clk	: std_logic;
	signal data_clk		: std_logic;
	signal reset		: std_logic;
	
	signal test_clk		: std_logic;
	
	signal reconfigure_i	: std_logic;
	signal rc_next_o	: std_logic;
	signal store_address_i	: std_logic_vector(a_width-1 downto 0);
	
	signal vertical_i	: std_logic;
	signal vertical_next_o	: std_logic;
	
	signal address_next_o	: std_logic_vector(a_width-1 downto 0);
	
	signal rc_pattern_i	: std_logic_vector(p_width-1 downto 0);
	signal rc_pattern_n_o	: std_logic_vector(p_width-1 downto 0);
	
	signal write_address_o	: std_logic_vector(a_width-1 downto 0);
	signal write_data_o	: std_logic_vector(c_size-1 downto 0);
	signal write_enable_o	: std_logic;
	
	signal config_in_i	: std_logic_vector(c_width-1 downto 0);
	
	signal unit_offset_i	: std_logic_vector(a_width-1 downto 0);
	signal unit_offset_n_o	: std_logic_vector(a_width-1 downto 0);
	
	signal running		: integer := 1;
begin
	uut : rc_writer
		generic map(
			config_width => c_width,
			config_size => c_size,
			address_width => a_width,
			pattern_width => p_width
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			reconfigure => reconfigure_i,
			store_address => store_address_i,
			rc_pattern => rc_pattern_i,
			rc_pattern_next => rc_pattern_n_o,
			vertical => vertical_i,
			vertical_next => vertical_next_o,
			write_address => write_address_o,
			write_data => write_data_o,
			write_enable => write_enable_o,
			config_in => config_in_i,
			rc_next => rc_next_o,
			address_next => address_next_o,
			unit_offset => unit_offset_i,
			unit_offset_next => unit_offset_n_o
		);

	config_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			config_clk <= '1';
			wait for 50 ns;
			config_clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	data_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			data_clk <= '1';
			wait for 50 ns;
			data_clk <= '0';
			wait for 350 ns;
		end loop;
		wait;
	end process;
	
	-- For internal use in the testbench
	test_clock : process is
	begin
		while running = 1 loop
			test_clk <= '1';
			wait for 50 ns;
			test_clk <= '0';
			wait for 350 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(data_clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		procedure perform_simulation (
			constant reconfigure	: std_logic;
			constant rc_pattern	: std_logic_vector(p_width-1 downto 0);
			constant rc_pattern_n	: std_logic_vector(p_width-1 downto 0);
			constant vertical	: std_logic;
			constant store_address	: std_logic_vector(a_width-1 downto 0);
			constant write_address	: std_logic_vector(a_width-1 downto 0);
			constant write_data	: std_logic_vector(c_size-1 downto 0);
			constant write_allow	: std_logic;
			constant config_data	: std_logic_vector(c_size-1 downto 0);
			constant address_next	: std_logic_vector(a_width-1 downto 0);
			constant unit_offset	: std_logic_vector(a_width-1 downto 0)
			)
		is 
			variable serial_in	: std_logic_vector(c_size-1 downto 0);
			variable mem_output	: std_logic_vector(c_size-1 downto 0);
			variable cycle_counter	: integer := (c_size/c_width);
		begin
			-- Don't play with the signals until the current async write is complete 
			serial_in := config_data;
			
			reconfigure_i <= reconfigure;
			store_address_i <= store_address;
			rc_pattern_i <= rc_pattern;
			vertical_i <= vertical;
			unit_offset_i <= unit_offset;
			
			wait until rising_edge(data_clk);
			reconfigure_i <= '0';
			store_address_i <= (others => '0');
			rc_pattern_i <= (others => '0');
			mem_output := (others => '0');
			unit_offset_i <= (others => '0');
			
			while cycle_counter > 1 loop
				-- Check output before config_clk rising edge when not reconfiguring
				config_in_i <= serial_in(c_width-1 downto 0);
				serial_in := (c_width-1 downto 0 => '0') & serial_in(c_size-1 downto c_width);
				wait for 2 ns; --propagation
				if write_allow = '0' then
					assert write_enable_o = '0' 
						report "Illegal write to memory"
						severity failure;
				end if;
				
				if write_address_o = write_address and write_enable_o = '1' then 
					mem_output := write_data_o;
				end if;
				wait until rising_edge(config_clk);
				
				cycle_counter := cycle_counter - 1;
			end loop;
			
			config_in_i <= serial_in(c_width-1 downto 0);
			
			wait until rising_edge(test_clk); 
			
			assert rc_pattern_n_o = rc_pattern_n 
				report "Incorrect pattern to next unit. Expected "
					& Image(rc_pattern_n)
					& " was "
					& Image(rc_pattern_n_o)
					severity failure;
			
			assert vertical = vertical_next_o
				report "Incorrect horizontal signal to next unit. Expected "
					& std_logic'image(vertical)
					& " was "
					& std_logic'image(vertical_next_o)
					severity failure;
			
			if reconfigure = '1' and write_allow = '1' then
				assert write_data = mem_output
					report "Config out incorrect. Expected " 
						& Image(write_data)
						& " was " 
						& Image(mem_output)
						severity failure;
			end if;
			
			assert reconfigure = rc_next_o 
				report "RC-signal to next unit incorrect. Expected "
					& std_logic'image(reconfigure)
					& " was " 
					& std_logic'image(rc_next_o)
					severity failure;
			
			if rc_next_o = '1' then
				assert address_next = address_next_o
					report "Base address to next unit incorrect. Expected "
					& Image(address_next)
					& " was " 
					& Image(address_next_o)
					severity failure;
			end if;
			
			assert unit_offset = unit_offset_n_o 
				report "Unit offset propagation incorrect"
				severity failure;
			
		end procedure;
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-rc_writer.txt";
			variable L: line;
			variable good: boolean;
			
			variable rc		:std_logic;
			variable rc_pattern	:std_logic_vector(p_width-1 downto 0);
			variable rc_pattern_next :std_logic_vector(p_width-1 downto 0);
			
			variable store_address	:std_logic_vector(a_width-1 downto 0);
			variable write_address	:std_logic_vector(a_width-1 downto 0);
			variable address_next	:std_logic_vector(a_width-1 downto 0);
			variable write_data	:std_logic_vector(c_size-1 downto 0);
			variable write_allow	:std_logic;
			variable config_data	:std_logic_vector(c_size-1 downto 0);
			
			variable vertical	:std_logic;
			variable unit_offset	:std_logic_vector(a_width-1 downto 0);
			
		begin
			
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				
				read(L,rc);
				read(L,rc_pattern);
				read(L,rc_pattern_next);
				read(L,store_address);
				read(L,write_address);
				read(L,write_data);
				read(L,write_allow);
				read(L,config_data);
				read(L,address_next);
				
				read(L,vertical);
				
				read(L,unit_offset);
				
				perform_simulation(rc,rc_pattern,rc_pattern_next,vertical,store_address,write_address,write_data,write_allow,config_data,address_next,unit_offset);
			end loop;
		end procedure;
		
	begin
		reconfigure_i <= '0';
		config_in_i <= (others => '0');
		reset_uut;
		
		-- Sync with clocks
		wait until rising_edge(test_clk);
		
		random_stimuli;
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;