library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version1_defs.all;

entity rc_writer_testbench is
end entity;

architecture rc_writer_test of rc_writer_testbench is
	constant c_size		: integer := 16;
	constant c_width	: integer := 4;
	constant a_width	: integer := 16;
	constant p_width	: integer := 4;
	
	signal config_clk	: std_logic;
	signal data_clk		: std_logic;
	signal reset		: std_logic;
	
	signal test_clk		: std_logic;
	
	signal reconfigure_i	: std_logic;
	signal rc_next_o	: std_logic;
	signal store_address_i	: std_logic_vector(a_width-1 downto 0);
	
	signal vertical_i	: std_logic;
	signal vertical_next_o	: std_logic;
	
	signal address_next_o	: std_logic_vector(a_width-1 downto 0);
	
	signal rc_pattern_i	: std_logic_vector(p_width-1 downto 0);
	signal rc_pattern_n_o	: std_logic_vector(p_width-1 downto 0);
	
	signal write_address_o	: std_logic_vector(a_width-1 downto 0);
	signal write_data_o	: std_logic_vector(c_size-1 downto 0);
	signal write_enable_o	: std_logic;
	
	signal config_in_i	: std_logic_vector(c_width-1 downto 0);
	
	signal unit_offset_i	: std_logic_vector(a_width-1 downto 0);
	signal unit_offset_n_o	: std_logic_vector(a_width-1 downto 0);
	
	signal running		: integer := 1;
begin
	uut : rc_writer
		generic map(
			config_width => c_width,
			config_size => c_size,
			address_width => a_width,
			pattern_width => p_width
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			reconfigure => reconfigure_i,
			store_address => store_address_i,
			rc_pattern => rc_pattern_i,
			rc_pattern_next => rc_pattern_n_o,
			vertical => vertical_i,
			vertical_next => vertical_next_o,
			write_address => write_address_o,
			write_data => write_data_o,
			write_enable => write_enable_o,
			config_in => config_in_i,
			rc_next => rc_next_o,
			address_next => address_next_o,
			unit_offset => unit_offset_i,
			unit_offset_next => unit_offset_n_o
		);

	config_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			config_clk <= '1';
			wait for 50 ns;
			config_clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	data_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			data_clk <= '1';
			wait for 50 ns;
			data_clk <= '0';
			wait for 350 ns;
		end loop;
		wait;
	end process;
	
	-- For internal use in the testbench
	test_clock : process is
	begin
		while running = 1 loop
			test_clk <= '1';
			wait for 50 ns;
			test_clk <= '0';
			wait for 350 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(data_clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		procedure read_std_logic(
			source		:inout line;
			res		:out std_logic
			)
		is
			variable char		:character;
			-- Todo: convert all testbenchs to verified mode
			variable good		:boolean;
		begin
			
			read(source,char,good);
			assert good report "Read failure" severity error;
			while char = ' ' loop
				read(source,char,good);
				assert good report "Read failure" severity error;
			end loop;
			
			if char = '0' then
				res := '0';
			else
				res := '1';
			end if;
		end procedure;

		
		procedure perform_simulation (
			constant reconfigure	: std_logic;
			constant rc_pattern	: std_logic_vector(p_width-1 downto 0);
			constant rc_pattern_n	: std_logic_vector(p_width-1 downto 0);
			constant vertical	: std_logic;
			constant store_address	: std_logic_vector(a_width-1 downto 0);
			constant write_address	: std_logic_vector(a_width-1 downto 0);
			constant write_data	: std_logic_vector(c_size-1 downto 0);
			constant write_allow	: std_logic;
			constant config_data	: std_logic_vector(c_size-1 downto 0);
			constant address_next	: std_logic_vector(a_width-1 downto 0);
			constant unit_offset	: std_logic_vector(a_width-1 downto 0)
			)
		is 
			--variable serial_in	: std_logic_vector(config_size-1 downto 0);
			variable serial_in	: std_logic_vector(c_size-1 downto 0);
			variable mem_output	: std_logic_vector(c_size-1 downto 0);
			--variable serial_in	: std_logic_vector(0 to config_size -1);
			--variable serial_out	: std_logic_vector(0 to config_size -1);
			variable cycle_counter	: integer := (c_size/c_width);
		begin
			-- Don't play with the signals until the current async write is complete 
			--config_en_i <= config_enable;
			--config_in_i <= config_in;
			serial_in := config_data;
			
			reconfigure_i <= reconfigure;
			store_address_i <= store_address;
			rc_pattern_i <= rc_pattern;
			vertical_i <= vertical;
			unit_offset_i <= unit_offset;
			
			wait until rising_edge(data_clk);
			reconfigure_i <= '0';
			store_address_i <= (others => '0');
			rc_pattern_i <= (others => '0');
			mem_output := (others => '0');
			unit_offset_i <= (others => '0');
			
			while cycle_counter > 1 loop
				-- Check output before config_clk rising edge when not reconfiguring
				config_in_i <= serial_in(c_width-1 downto 0);
				serial_in := (c_width-1 downto 0 => '0') & serial_in(c_size-1 downto c_width);
				wait for 2 ns; --propagation
				--read_data_i <= next_mem_data;
				if write_allow = '0' then
					assert write_enable_o = '0' 
						report "Illegal write to memory"
						severity failure;
				end if;
				
				if write_address_o = write_address and write_enable_o = '1' then 
					mem_output := write_data_o;
				end if;
--				if reg_rc = '1' then
--					wait until rising_edge(config_clk);
--					--read_data_i <= next_mem_data;
--					serial_out := config_out_o & serial_out(c_size-1 downto c_width);
--					wait for 1 ns;
--					wait for 10 ns;
--					if read_address_o = read_address then 
--						next_mem_data := read_data;
--					else
--						next_mem_data := (others => '0');
--					end if;
--				else 
					-- Allow for propagation
					--wait for 1 ns;
					--serial_out := config_out_o & serial_out(c_size-1 downto c_width);
--					wait until rising_edge(config_clk);
--				end if;
				wait until rising_edge(config_clk);
				
				cycle_counter := cycle_counter - 1;
			end loop;
			
			--report Image(serial_in);
			config_in_i <= serial_in(c_width-1 downto 0);
			
			wait until rising_edge(test_clk); 
			--serial_in := (c_width-1 downto 0 => '0') & serial_in(c_size-1 downto c_width);
			
			assert rc_pattern_n_o = rc_pattern_n 
				report "Incorrect pattern to next unit. Expected "
					& Image(rc_pattern_n)
					& " was "
					& Image(rc_pattern_n_o)
					severity failure;
			
			assert vertical = vertical_next_o
				report "Incorrect horizontal signal to next unit. Expected "
					& std_logic'image(vertical)
					& " was "
					& std_logic'image(vertical_next_o)
					severity failure;
			
			if reconfigure = '1' and write_allow = '1' then
				assert write_data = mem_output
					report "Config out incorrect. Expected " 
						--& integer'image(to_integer(unsigned(config_out)))
						& Image(write_data)
						& " was " 
						--& integer'image(to_integer(unsigned(serial_out)))
						& Image(mem_output)
						severity failure;
			end if;
			
			assert reconfigure = rc_next_o 
				report "RC-signal to next unit incorrect. Expected "
					& std_logic'image(reconfigure)
					& " was " 
					& std_logic'image(rc_next_o)
					severity failure;
			
			if rc_next_o = '1' then
				assert address_next = address_next_o
					report "Base address to next unit incorrect. Expected "
					& Image(address_next)
					& " was " 
					& Image(address_next_o)
					severity failure;
			end if;
			
			assert unit_offset = unit_offset_n_o 
				report "Unit offset propagation incorrect"
				severity failure;
			
		end procedure;
	
	begin
		reconfigure_i <= '0';
		config_in_i <= (others => '0');
		--config_in_i <= (others => '0');
		reset_uut;
		
		-- Check reset 
		-- Test critical output signals
		assert write_enable_o = '0' report "Reset failed. (Write asserted)" severity failure;
		--perform_simulation(rc,rc_pattern,rc_pattern_next,vertical,store_address,write_address,write_data,write_allow,config_data,address_next,unit_offset);
		
		-- Sync with clocks
		wait until rising_edge(test_clk);
		--perform_simulation(rc,rc_pattern,rc_pattern_next,vertical,store_address,write_address,write_data,write_allow,config_data,address_next,unit_offset);
		-- BASIC IO
		perform_simulation('0',X"0",X"0",'0',X"0000",X"0000",X"0000",'0',X"FFFF",X"0000",X"0000");
		perform_simulation('1',X"5",X"A",'1',X"0100",X"0100",X"FFFF",'1',X"0000",X"0104",X"0004");
		perform_simulation('1',X"5",X"A",'1',X"0100",X"0101",X"0000",'1',X"5555",X"0104",X"0004");
		perform_simulation('1',X"5",X"A",'1',X"0100",X"0102",X"5555",'1',X"AAAA",X"0104",X"0004");
		perform_simulation('1',X"5",X"A",'1',X"0100",X"0103",X"AAAA",'1',X"5555",X"0104",X"0004");
		perform_simulation('0',X"5",X"A",'1',X"0000",X"0000",X"0000",'0',X"6464",X"0000",X"0000");
		
		-- Check pattern
		-- Hortizontal
		-- Disable
		perform_simulation('1',X"2",X"1",'1',X"1210",X"0000",X"0000",'0',X"FFFF",X"1210",X"0004");
		-- Enable
		perform_simulation('1',X"5",X"A",'1',X"1210",X"1210",X"FFFF",'1',X"8495",X"1214",X"0004");
		-- Vertical
		perform_simulation('1',X"5",X"A",'0',X"1210",X"0000",X"0000",'0',X"543E",X"1214",X"0004");
		perform_simulation('1',X"5",X"A",'1',X"1210",X"1211",X"543E",'1',X"8495",X"1214",X"0004");
		perform_simulation('0',X"5",X"A",'1',X"0000",X"0000",X"0000",'0',X"6464",X"0000",X"0000");
		
		-- Check address propagation
		perform_simulation('1',X"0",X"0",'0',X"1500",X"0000",X"0000",'0',X"8495",X"1500",X"00CD");
		perform_simulation('1',X"1",X"8",'1',X"1500",X"1500",X"8495",'1',X"AB3D",X"15CD",X"00CD");
		perform_simulation('1',X"1",X"8",'0',X"1500",X"0000",X"0000",'0',X"DF32",X"15CD",X"00CD");
		
		-- Back to back reconfiguration
		perform_simulation('1',X"1",X"8",'1',X"1500",X"1501",X"DF32",'1',X"4354",X"15CD",X"00CD");
		perform_simulation('1',X"1",X"8",'1',X"3200",X"3200",X"4354",'1',X"BEEF",X"3208",X"0008");
		perform_simulation('1',X"1",X"8",'1',X"3200",X"3201",X"BEEF",'1',X"BABE",X"3208",X"0008");
		perform_simulation('1',X"1",X"8",'1',X"3200",X"3202",X"BABE",'1',X"DEAD",X"3208",X"0008");
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;