library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use work.common_defs.all;

entity sc_register_testbench is
end entity;

architecture sc_register_test2 of sc_register_testbench is
	constant size		: integer := 16;
	constant c_size		: integer := 4;

	signal config_clk	: std_logic;
	signal para_clk		: std_logic;
	signal reset		: std_logic;

	signal write_i		: std_logic;
	signal config_en_i	: std_logic;
	
	signal config_in_i	: std_logic_vector(c_size-1 downto 0);
	signal config_out_o	: std_logic_vector(c_size-1 downto 0);

	signal data_i		: std_logic_vector(size -1 downto 0);
	signal output_o		: std_logic_vector(size -1 downto 0);

	signal running		: integer := 1;
begin
	uut : sc_register
		generic map(
			width => size,
			config_width => c_size,
			allow_prohibit => '0'
		)
		port map(
			config_clk => config_clk,
			data_clk => para_clk,
			reset => reset,
			config_in => config_in_i,
			config_out => config_out_o,
			config_enable => config_en_i,
			data_in => data_i,
			data_out => output_o,
			write => write_i
		);

	serial_clock : process is
	begin
		while running = 1 loop
			config_clk <= '1';
			wait for 50 ns;
			config_clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	parallel_clock : process is
	begin
		while running = 1 loop
			-- size/c_size rising edges needed to empty register
			para_clk <= '1';
			wait for 50 ns;
			para_clk <= '0';
			wait for 350 ns;
		end loop;
		wait;
	end process;
	
	manual_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(config_clk);
			wait until rising_edge(para_clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		procedure perform_simulation (
			constant write		: std_logic;
			constant config_enable	: std_logic;
			constant config_in	: std_logic_vector(size-1 downto 0);
			constant data_in	: std_logic_vector(size-1 downto 0);
			
			constant config_out	: std_logic_vector(size-1 downto 0);
			constant data_out	: std_logic_vector(size-1 downto 0);
			constant new_data_out	: std_logic_vector(size-1 downto 0)
			)
		is 
			variable serial_in	: std_logic_vector(size-1 downto 0);
			variable serial_out	: std_logic_vector(size-1 downto 0);
			variable cycle_counter	: integer := (size/c_size);
		begin
			-- Don't play with the signals until the current async write is complete 
			wait until para_clk = '0';
			write_i <= write;
			serial_in := config_in;
			
			data_i <= data_in;
			config_en_i <= config_enable;
			
			while cycle_counter > 0 loop
				-- Allow the register to take appropriate action
				wait for 1 ns;
				-- Check for correct data_out, This should not change until para_clk'high
				if config_enable = '0' then
					assert data_out = output_o report "Fast data out incorrect. Expected " 
						& Image(data_out)
						& " was "
						& Image(output_o)
						severity failure;
				end if;
				-- Data should ready right away
				serial_out := config_out_o & serial_out(size-1 downto c_size);
				-- Put data on input
				config_in_i <= serial_in(c_size-1 downto 0);
				serial_in := (c_size -1 downto 0 => '0') & serial_in(size -1 downto c_size);
				cycle_counter := cycle_counter - 1;
				wait until rising_edge(config_clk);
			end loop;
			
			wait for 1 ns; -- Allow for propagation
			
			-- The value on the input shouldn't mather anymore
			config_in_i <= (others => '0');
			
			assert new_data_out = output_o report "End data out incorrect. Expected " 
				& Image(new_data_out)
				& " was "
				& Image(output_o)
				severity failure;
				-- Only verify config_out when config is active 
			if config_enable = '1' then
				assert config_out = serial_out
					report "Config out incorrect. Expected " 
						& Image(config_out)
						& " was " 
						& Image(serial_out)
						severity failure;
			end if;		
		end procedure;
	begin
		reset_uut;
		
		-- Check reset 
		assert output_o = (output_o'range => '0') report "Reset failed" severity failure;
		assert config_out_o = (config_out_o'range => '0') report "Reset failed" severity failure;
		
		
		-- Test basic clocking
		-- Basic IO para I/O
		perform_simulation('1','0',X"0000",X"0000",X"0000",X"0000",X"0000");
		perform_simulation('1','0',X"0000",X"FFFF",X"0000",X"0000",X"FFFF");
		perform_simulation('1','0',X"0000",X"AAAA",X"0000",X"FFFF",X"AAAA");
		perform_simulation('1','0',X"0000",X"5555",X"0000",X"AAAA",X"5555");
		
		reset_uut;
		-- Test write
		perform_simulation('1','0',X"0000",X"BEEF",X"0000",X"0000",X"BEEF");
		perform_simulation('0','0',X"0000",X"FFFF",X"0000",X"BEEF",X"BEEF");
		perform_simulation('0','0',X"0000",X"AAAA",X"0000",X"BEEF",X"BEEF");
		perform_simulation('0','0',X"0000",X"5555",X"0000",X"BEEF",X"BEEF");
		
		reset_uut;
		-- Test IO config
		perform_simulation('0','1',X"0000",X"0000",X"0000",X"0000",X"0000");
		perform_simulation('0','1',X"FA50",X"0000",X"0000",X"0000",X"FA50");
		perform_simulation('0','1',X"FFFF",X"0000",X"FA50",X"FA50",X"FFFF");
		perform_simulation('0','1',X"DDDD",X"0000",X"FFFF",X"DDDD",X"DDDD");
		perform_simulation('0','0',X"0000",X"0000",X"0000",X"DDDD",X"DDDD");
		perform_simulation('0','1',X"AAAA",X"5DD5",X"DDDD",X"DDDD",X"AAAA");
		perform_simulation('0','1',X"5555",X"EFDE",X"AAAA",X"AAAA",X"5555");
		perform_simulation('0','1',X"BEEF",X"5DD5",X"5555",X"5555",X"BEEF");
		
		reset_uut;
		-- Write + config combined
		-- Test writing and immediatly starting config after
		-- With allow_prohibit = '0' this should be allowed
		perform_simulation('1','0',X"0000",X"BEEF",X"0000",X"0000",X"BEEF");
		perform_simulation('0','1',X"DEAD",X"BEEF",X"BEEF",X"0000",X"DEAD");
		-- A write immediatly after a config should be ignored
		perform_simulation('1','1',X"C01D",X"DADA",X"DEAD",X"0000",X"DADA");
		-- A write should be possible now
		perform_simulation('1','0',X"0000",X"BABE",X"0000",X"DADA",X"BABE");
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;