library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.common_defs.all;

entity auto_sc_register_testbench is
end entity;

architecture sc_register_test of auto_sc_register_testbench is
	constant size		: integer := 32;
	constant c_size		: integer := 2;

	signal config_clk	: std_logic;
	signal para_clk		: std_logic;
	signal reset		: std_logic;

	signal write_i		: std_logic;
	signal config_en_i	: std_logic;
	
	signal config_in_i	: std_logic_vector(c_size-1 downto 0);
	signal config_out_o	: std_logic_vector(c_size-1 downto 0);

	signal data_i		: std_logic_vector(size -1 downto 0);
	signal output_o		: std_logic_vector(size -1 downto 0);

	signal running		: integer := 1;
begin
	uut : sc_register
		generic map(
			width => size,
			config_width => c_size
		)
		port map(
			config_clk => config_clk,
			data_clk => para_clk,
			reset => reset,
			config_in => config_in_i,
			config_out => config_out_o,
			config_enable => config_en_i,
			data_in => data_i,
			data_out => output_o,
			write => write_i
		);
	
	serial_clock : process is
	begin
		while running = 1 loop
			config_clk <= '1';
			wait for 50 ns;
			config_clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	parallel_clock : process is
	begin
		while running = 1 loop
			-- size/c_size rising edges needed to empty register
			para_clk <= '1';
			wait for 50 ns;
			para_clk <= '0';
			wait for 1550 ns;
		end loop;
		wait;
	end process;
	
	random_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(config_clk);
			wait until rising_edge(para_clk);
			reset <= '1';
		end procedure;
		
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "auto-sc_register.txt";
			variable L: line;
			variable good: boolean;
			
			
			variable write		:std_logic;
			variable config_enable	:std_logic;
			
			variable config_in	:integer;
			variable config_out	:integer;
			
			variable data_in	:integer;
			variable new_data_out	:integer;
			variable data_out	:integer;
		
			procedure perform_simulation is 
				variable serial_in	: std_logic_vector(size-1 downto 0);
				variable serial_out	: std_logic_vector(size-1 downto 0);
				variable cycle_counter	: integer := (size/c_size);
			begin
				-- Don't play with the signals until the current async write is complete 
				wait until para_clk = '0'; -- MEH.. this just might break everything..
				
				write_i <= write;
				config_en_i <= config_enable;
				
				serial_in := std_logic_vector(to_signed(config_in,size));
				
				data_i <= std_logic_vector(to_signed(data_in,size));
				
				while cycle_counter > 0 loop
					-- Allow the register to take appropriate action
					wait for 1 ns;
					-- Check for correct data_out, This should not change until para_clk'high
					if config_enable = '0' then
						assert std_logic_vector(to_signed(data_out,size)) = output_o report "Fast data out incorrect. Expected "
							& integer'image(data_out)
							& " was "
							& integer'image(to_integer(unsigned(output_o)))
							severity failure;
					end if;
					-- Data should ready right away 
					serial_out := config_out_o & serial_out(size-1 downto c_size);
					-- Put data on input
					config_in_i <= serial_in(c_size-1 downto 0);
					serial_in := (c_size -1 downto 0 => '0') & serial_in(size -1 downto c_size);
					cycle_counter := cycle_counter - 1;
					wait until rising_edge(config_clk);
				end loop;
				
				wait for 1 ns; -- Allow for propagation
				
				config_in_i <= (others => '0');
				
				assert std_logic_vector(to_signed(new_data_out,size)) = output_o report "End data out incorrect" severity failure;
				-- Only verify config_out when config is active 
				if config_enable = '1' then
					assert std_logic_vector(to_signed(config_out,size)) = serial_out 
						report "Config out incorrect. Expected " 
							& integer'image(config_out)
							& " was " 
							& integer'image(to_integer(signed(serial_out)))
							severity failure;
				end if;
			end procedure;
		begin
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				
				read(L,config_enable);
				read(L,config_in);
				read(L,config_out);
				
				read(L,write);
				read(L,data_in);
				read(L,data_out);
				read(L,new_data_out);
				
				perform_simulation;
			end loop;
		end procedure;
	begin
		reset_uut;
		
		random_stimuli;
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;