library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
use work.common_defs.all;

entity shift_register_testbench is
end entity;

architecture shift_register_test of shift_register_testbench is
	constant size		: integer := 32;

	signal clk		: std_logic;
	signal reset		: std_logic;

	signal write_i		: std_logic;
	signal shift_left_i	: std_logic;
	signal shift_right_i	: std_logic;
	signal carry_in_i	: std_logic;
	
	signal data_i		: std_logic_vector(size -1 downto 0);

	signal carry_out_o	: std_logic;
	signal output_o		: std_logic_vector(size -1 downto 0);

	signal running		: integer := 1;
begin
	uut : shift_register
		generic map(
			width => size
		)
		port map(
			clk => clk,
			reset => reset,
			write => write_i,
			shift_left => shift_left_i,
			shift_right => shift_right_i,
			carry_in => carry_in_i,
			carry_out => carry_out_o,
			data_in => data_i,
			data_out => output_o
		);
	
	clock : process is
	begin
		while running = 1 loop
			clk <= '1';
			wait for 50 ns;
			clk <= '0';
			wait for 50 ns;
		end loop;
		wait;
	end process;
	
	manual_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(clk);
			wait until rising_edge(clk);
			reset <= '1';
		end procedure;
		procedure perform_simulation (
			constant write		: std_logic;
			constant shift_left	: std_logic;
			constant shift_right	: std_logic;
			constant carry_in	: std_logic;
			constant operand	: std_logic_vector(size-1 downto 0);
			
			constant carry_out	: std_logic;
			constant result		: std_logic_vector(size-1 downto 0)
			)
		is 
		
		begin
			write_i <= write;
			carry_in_i <= carry_in;
			shift_left_i <= shift_left;
			shift_right_i <= shift_right;
			data_i <= operand;
			wait until rising_edge(clk);
			
			wait for 1 ns; -- Allow for propagation
			assert carry_out = carry_out_o 
				report "Carry out incorrect " 
				& std_logic'image(shift_left) 
				& std_logic'image(shift_right) 
				& std_logic'image(carry_in) 
				& " " 
				& integer'image(to_integer(signed(operand)))
				& " "
				& std_logic'image(carry_out) 
				& " was "
				& std_logic'image(carry_out_o)
				& integer'image(to_integer(signed(result))) 
				severity failure;
			assert result = output_o 
				report "Output incorrect "
				& std_logic'image(shift_left) 
				& std_logic'image(shift_right) 
				& std_logic'image(carry_in) 
				& " " 
				& integer'image(to_integer(signed(operand))) 
				& " "
				& std_logic'image(carry_out) 
				& " "
				& integer'image(to_integer(signed(result)))
				& " was "
				& integer'image(to_integer(signed(output_o))) 
				severity failure;
		
		end procedure;
	begin
		reset_uut;
		
		-- Check reset 
		assert output_o = (output_o'range => '0') report "Reset failed" severity failure;
		assert carry_out_o = '0' report "Reset failed" severity failure;
		
		-- Test basic clocking
		-- Basic IO 
		-- No shift
		perform_simulation('1','0','0','0',X"FFFFFFFF",'0',X"FFFFFFFF");
		perform_simulation('1','0','0','0',X"AAAAAAAA",'0',X"AAAAAAAA");
		perform_simulation('1','0','0','0',X"55555555",'0',X"55555555");
		perform_simulation('1','0','0','0',X"00000000",'0',X"00000000");
		
		-- Test write flag for data 
		perform_simulation('1','0','0','0',X"DEADBEEF",'0',X"DEADBEEF");
		perform_simulation('0','0','0','0',X"54535003",'0',X"DEADBEEF");
		perform_simulation('0','0','0','0',X"54535003",'0',X"DEADBEEF");
		-- Even with some random flags set
		perform_simulation('0','1','0','1',X"54535003",'0',X"DEADBEEF");
		perform_simulation('1','0','0','0',X"54535003",'0',X"54535003");
		
		-- Test shift left
		perform_simulation('1','1','0','0',X"00000002",'0',X"00000004");
		perform_simulation('1','1','0','0',X"00000008",'0',X"00000010");
		perform_simulation('1','1','0','0',X"55555555",'0',X"AAAAAAAA");
		
		-- Test shift right
		perform_simulation('1','0','1','0',X"00000002",'0',X"00000001");
		perform_simulation('1','0','1','0',X"00000010",'0',X"00000008");
		perform_simulation('1','0','1','0',X"AAAAAAAA",'0',X"55555555");
		
		-- Test that shift left has priority
		perform_simulation('1','1','1','0',X"55555555",'0',X"AAAAAAAA");
		
		-- Test carry in 
		-- Left
		perform_simulation('1','1','0','1',X"00000008",'0',X"00000011");
		perform_simulation('1','1','0','1',X"0000AAAA",'0',X"00015555");
		-- Right
		perform_simulation('1','0','1','1',X"20000008",'0',X"90000004");
		perform_simulation('1','0','1','1',X"55550000",'0',X"AAAA8000");
		
		-- Test carry out
		-- Left
		perform_simulation('1','1','0','0',X"90000000",'1',X"20000000");
		perform_simulation('1','1','1','1',X"AAAAAAAA",'1',X"55555555");
		-- Right
		perform_simulation('1','0','1','0',X"00000009",'1',X"00000004");
		perform_simulation('1','0','1','1',X"55555555",'1',X"AAAAAAAA");
		
		-- Test write flag for carry
		perform_simulation('1','1','0','0',X"AAAAAAAA",'1',X"55555554");
		perform_simulation('0','1','0','0',X"75984753",'1',X"55555554");
		perform_simulation('0','1','0','1',X"75984753",'1',X"55555554");
		perform_simulation('0','0','1','1',X"75984753",'1',X"55555554");
		perform_simulation('1','1','0','1',X"75984753",'0',X"EB308EA7");
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;