#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "generate.h"
#include "simulator_backend.h"

static void write_cell_function(struct cell* source, FILE* fd);
static void write_cell_map(struct cell** source_cell, 
		int size_x, int size_y, FILE* fd);
static int discover_order(struct cell* source_cell);
static void write_alu_valid(struct cell* source_cell, FILE* fd);
static void write_md_valid(struct cell* source_cell, FILE* fd);
static void write_valid(struct cell* source_cell, int order, FILE* fd);
static void write_computation(struct cell* source_cell, int order, FILE* fd);
static void write_outputs(struct cell* source_cell, FILE* fd);

const char* simulator_alu_inputs[] = {
	"west",
	"north",
	"self",
	"mulout"
};

const char simulator_alu_inputs_prefix[] = {
	' ',
	' ',
	'*',
	' '
};

const char* simulator_md_inputs[] = {
	"west",
	"north",
	"self",
	"aluout"
};

const char simulator_md_inputs_prefix[] = {
	' ',
	' ',
	'*',
	' '
};

const char* simulator_outputs[] = {
	"north",
	"west",
	"aluout",
	"mulout",
	"self"
};

const char simulator_outputs_prefix[] = {
	' ',
	' ',
	' ',
	' ',
	'*'
};


int write_simulator_file(struct cell** source, 
		int size_x, int size_y, FILE* fd) {
	int i = 0, j = 0;
	struct cell* source_cell = NULL;
	
	if ( fd == NULL ) {
		return 1;
	}
	for ( i = 0 ; i < size_y; i++ ) {
		for ( j = 0 ; j < size_x ; j++ ) {
			source_cell = MAP(source,size_y,i,j);
			write_cell_function(source_cell,fd);
		}
	}
	write_cell_map(source,size_x,size_y,fd);
	return 0;
}

static void write_cell_map(struct cell** source, 
		int size_x, int size_y, FILE* fd) {
	
	int i = 0, j = 0;
	struct cell* source_cell = NULL;
	fprintf(fd,"struct cell_rc_targets fabric[%d][%d] = {\n"
		,size_x,size_y
	);
	for ( i = 0 ; i < size_x; i++ ) {
		fprintf(fd,"{");
		for ( j = 0 ; j < size_y  ; j++) {
			source_cell = MAP(source,size_y,i,j);
			fprintf(fd,"{cell%d_%d,%d,%d}",source_cell->y,source_cell->x,source_cell->do_init,source_cell->init_value);
			if ( j < (size_y -1)) {
				fprintf(fd,",");
			}
		}
		if ( i < (size_x -1 ) ) {
			fprintf(fd,"},\n");
		} else {
			fprintf(fd,"}\n");
		}
	}
	fprintf(fd,"};\n");
	fprintf(fd,
		"struct cell_rc_targets* FABRIC_SYMBOL = (struct cell_rc_targets*) &fabric;\n");
	fprintf(fd,"grid_size_t GRID_DIMENTIONS = 2;\n");
	fprintf(fd,"grid_size_t GRID_SIZE[2] = { %d, %d };",size_x,size_y);
	fprintf(fd,"int random_seed = %d;\n", rand());
}


static void write_cell_function(struct cell* source_cell, FILE* fd) {
	int order = discover_order(source_cell);
	fprintf(fd,"void cell%d_%d(int ticks_left,", source_cell->y, source_cell->x);
	fprintf(fd,"cell_valid valid_north, cell_transfer north,\n");
	fprintf(fd,"\t\tcell_valid valid_west, cell_transfer west,\n");
	fprintf(fd,"\t\tcell_valid* valid_east, cell_transfer* east,\n");
	fprintf(fd,"\t\tcell_valid* valid_south, cell_transfer* south,\n");
	fprintf(fd,"\t\tcell_valid* valid_self, cell_transfer* self ) {\n");
	fprintf(fd,"\tcell_valid valid_aluout = 0, valid_mulout = 0;\n");
	fprintf(fd,"\tcell_transfer aluout = 0, mulout = 0;\n");
	fprintf(fd,"\tif (ticks_left == 1) {\n");
	write_valid(source_cell,order,fd);
	write_computation(source_cell,order,fd);
	write_outputs(source_cell,fd);
	
	fprintf(fd,"\t}\n");
	fprintf(fd,"}\n");
}

/* Discover executionorder of alu and mul */
/* Assumes cyclefree input */
static int discover_order(struct cell* source_cell ) {
	if ( 	source_cell->alu_a == alu_mulout 
		|| source_cell->alu_b == alu_mulout ) {
		
		return 0;
	} else {
		return 1;
	}
}

static void write_valid(struct cell* source_cell, int order, FILE* fd) {
	
	if ( order == 0 ) {
		write_md_valid(source_cell,fd);
		write_alu_valid(source_cell,fd);
	} else {
		write_alu_valid(source_cell,fd);
		write_md_valid(source_cell,fd);
	}
}

static void write_alu_valid(struct cell* source_cell, FILE* fd) {
	fprintf(fd,"\t\tvalid_aluout = %cvalid_%s & %cvalid_%s;\n",
		simulator_alu_inputs_prefix[source_cell->alu_a],
		simulator_alu_inputs[source_cell->alu_a],
		simulator_alu_inputs_prefix[source_cell->alu_b],
		simulator_alu_inputs[source_cell->alu_b]
	);
}

static void write_md_valid(struct cell* source_cell, FILE* fd) {
	fprintf(fd,"\t\tvalid_mulout = %cvalid_%s & %cvalid_%s;\n",
		simulator_md_inputs_prefix[source_cell->md_a],
		simulator_md_inputs[source_cell->md_a],
		simulator_md_inputs_prefix[source_cell->md_b],
		simulator_md_inputs[source_cell->md_b]
	);
}

static void write_computation(struct cell* source_cell, int order, FILE* fd) {
	if ( order == 0 ) {
		fprintf(fd,"\t\tmd_operation(%d, %c%s, %c%s, &mulout);\n",
			source_cell->md_op,
			simulator_md_inputs_prefix[source_cell->md_a],
			simulator_md_inputs[source_cell->md_a],
			simulator_md_inputs_prefix[source_cell->md_b],
			simulator_md_inputs[source_cell->md_b]);
		fprintf(fd,"\t\talu_operation(%d, %c%s, %c%s, &aluout);\n",
			source_cell->alu_op,
			simulator_alu_inputs_prefix[source_cell->alu_a],
			simulator_alu_inputs[source_cell->alu_a],
			simulator_alu_inputs_prefix[source_cell->alu_b],
			simulator_alu_inputs[source_cell->alu_b]);
	} else {
		fprintf(fd,"\t\talu_operation(%d, %c%s, %c%s, &aluout);\n",
			source_cell->alu_op,
			simulator_alu_inputs_prefix[source_cell->alu_a],
			simulator_alu_inputs[source_cell->alu_a],
			simulator_alu_inputs_prefix[source_cell->alu_b],
			simulator_alu_inputs[source_cell->alu_b]);
		fprintf(fd,"\t\tmd_operation(%d, %c%s, %c%s, &mulout);\n",
			source_cell->md_op,
			simulator_md_inputs_prefix[source_cell->md_a],
			simulator_md_inputs[source_cell->md_a],
			simulator_md_inputs_prefix[source_cell->md_b],
			simulator_md_inputs[source_cell->md_b]);
	}
}

static void write_outputs(struct cell* source_cell, FILE* fd) {
	fprintf(fd,"\t\t*valid_east = %cvalid_%s;\n",
		simulator_outputs_prefix[source_cell->east],
		simulator_outputs[source_cell->east]);
	fprintf(fd,"\t\t*east = %c%s;\n",
		simulator_outputs_prefix[source_cell->east],
		simulator_outputs[source_cell->east]);
	fprintf(fd,"\t\t*valid_south = %cvalid_%s;\n",
		simulator_outputs_prefix[source_cell->south],
		simulator_outputs[source_cell->south]);
	fprintf(fd,"\t\t*south = %c%s;\n",
		simulator_outputs_prefix[source_cell->south],
		simulator_outputs[source_cell->south]);
	fprintf(fd,"\t\t*valid_self = %cvalid_%s;\n",
		simulator_outputs_prefix[source_cell->state],
		simulator_outputs[source_cell->state]);
	fprintf(fd,"\t\t*self = %c%s;\n",
		simulator_outputs_prefix[source_cell->state],
		simulator_outputs[source_cell->state]);
}

