library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;
use std.textio.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version2_defs.all;

entity auto_wrpu_testbench is
	generic(
		cells_axis	: natural := 8
	);
end entity;

architecture wrpu_test of auto_wrpu_testbench is
	constant data_width	: positive := 16;
	constant addr_width	: positive := 8;
	constant config_width	: positive := 3;
	constant config_size	: positive := 57;
	constant header_size	: positive := 24;
	constant base_c_size	: positive := 38;
	
	signal config_clk	: std_logic;
	signal data_clk		: std_logic;
	signal compute_clk	: std_logic;
	signal reset		: std_logic;
	
	signal test_clk		: std_logic;
	
	signal rc_ctrl_i	: std_logic_vector(rcm_ctrl_size-1 downto 0);
	signal rc_status_o	: std_logic_vector(rcm_status_size -1 downto 0);
	signal src_base_addr_i	: std_logic_vector(addr_width-1 downto 0);
	signal dest_base_addr_i	: std_logic_vector(addr_width-1 downto 0);
	
	signal north_bank_i	: io_bank(cells_axis-1 downto 0);
	signal west_bank_i	: io_bank(cells_axis-1 downto 0);
	signal south_bank_o	: io_bank(cells_axis-1 downto 0);
	signal east_bank_o	: io_bank(cells_axis-1 downto 0);
	
	signal running		: integer := 1;
begin
	uut : wrpu
		generic map(
			data_width => data_width,
			addr_width => addr_width,
			config_width => config_width,
			config_size => config_size,
			base_config_size => base_c_size,
			header_size => header_size,
			cells_axis => cells_axis
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			compute_clk => compute_clk,
			reset => reset,
			rc_ctrl => rc_ctrl_i,
			rc_status => rc_status_o,
			src_base_addr => src_base_addr_i,
			dest_base_addr => dest_base_addr_i,
			north_bank => north_bank_i,
			west_bank => west_bank_i,
			south_bank => south_bank_o,
			east_bank => east_bank_o
		);

	config_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			config_clk <= '1';
			wait for 73 ns;
			config_clk <= '0';
			wait for 74 ns;
		end loop;
		wait;
	end process;
	
	data_clock : process is
	begin
		wait for 10 ns;
		while running = 1 loop
			data_clk <= '1';
			wait for 73 ns;
			data_clk <= '0';
			wait for 2720 ns;
		end loop;
		wait;
	end process;
	
	compute_clock : process is 
	begin
		wait for 10 ns;
		while running = 1 loop
			compute_clk <= '1';
			wait for 67 ns;
			compute_clk <= '0';
			wait for 66 ns;
		end loop;
		wait;
	end process;
	
	
	-- For internal use in the testbench
	test_clock : process is
	begin
		while running = 1 loop
			test_clk <= '1';
			wait for 73 ns;
			test_clk <= '0';
			wait for 2720 ns;
		end loop;
		wait;
	end process;
	
	simu_test : process is
		procedure reset_uut is 
		begin
			reset <= '0';
			wait until rising_edge(data_clk);
			reset <= '1';
		end procedure;
		
		-- Shamelessly copied from:
		-- VHDL : Answers to frequently asked questions
		-- 2nd edition
		-- Ben Cohen
		function Image(in_image : std_logic_vector ) return String is
			variable L: line;
			variable W: string(1 to In_Image'length) := (others => ' ');
		begin
			write(L,in_image);
			W(L.all'range) := L.all;
			Deallocate(L);
			return W;
		end Image;
		
		procedure read(
			source		:inout line;
			res		:out io_bank
			)
		is
			variable vector	: std_logic_vector(res(0)'high downto 0);
			variable good	: boolean;
		begin
			
			for i in res'low to res'high loop
				read(source,vector,good);
				res(i) := vector;
				assert good report "Read failure" severity error;
			end loop;
			
		end procedure;
		
		
		procedure perform_simulation (
			constant rc_ctrl	: std_logic_vector(rcm_ctrl_size-1 downto 0);
			constant rc_status	: std_logic_vector(rcm_status_size-1 downto 0);
			constant src_base_addr	: std_logic_vector(addr_width-1 downto 0);
			constant dest_base_addr	: std_logic_vector(addr_width-1 downto 0);
			
			constant north		: io_bank(cells_axis-1 downto 0);
			constant west		: io_bank(cells_axis-1 downto 0);
			constant south		: io_bank(cells_axis-1 downto 0);
			constant east		: io_bank(cells_axis-1 downto 0)
			)
		is 
		begin
			rc_ctrl_i <= rc_ctrl;
			src_base_addr_i <= src_base_addr;
			dest_base_addr_i <= dest_base_addr;
			for i in south'range loop
				-- Check valid is equal 
				assert south(i)(data_width) = south_bank_o(i)(data_width)
					report "South bank valid incorrect at "
						& integer'image(i)
						& " Expected "
						& std_logic'image(south(i)(data_width))
						& " was "
						& std_logic'image(south_bank_o(i)(data_width))
						& " Bank. Expected "
						& Image(south(i))
						& " was "
						& Image(south_bank_o(i))
						 severity failure;
				if south(i)(data_width) = '1' then
					assert south(i) = south_bank_o(i) 
						report "South bank output incorrect at "
						& integer'image(i)
						& " Expected "
						& Image(south(i))
						& " was "
						& Image(south_bank_o(i))
						 severity failure;
				end if;
			end loop;
			
			for i in east'range loop
				assert east(i)(data_width) = east_bank_o(i)(data_width)
					report "East bank valid incorrect at "
						& integer'image(i)
						& " Expected "
						& std_logic'image(east(i)(data_width))
						& " was "
						& std_logic'image(east_bank_o(i)(data_width))
						& " Bank. Expected "
						& Image(east(i))
						& " was "
						& Image(east_bank_o(i))
						 severity failure;
				if east(i)(data_width) = '1' then
					assert east(i) = east_bank_o(i) 
						report "East bank output incorrect at "
						& integer'image(i)
						& " Expected "
						& Image(east(i))
						& " was "
						& Image(east_bank_o(i))
						 severity failure;
				end if;
			end loop;
			north_bank_i <= north;
			west_bank_i <= west;
			wait until rising_edge(data_clk);
			north_bank_i <= (others => (others => '0'));
			west_bank_i <= (others => (others => '0'));
			src_base_addr_i <= (others => '0');
			dest_base_addr_i <= (others => '0');
			
			wait until rising_edge(test_clk); 
			
			
			
		end procedure;
	
		procedure random_stimuli is 
			file stimuli_file	:text open read_mode is "io-dump.txt";
			variable L: line;
			variable good: boolean;
			
			variable rc_ctrl	:std_logic_vector(rcm_ctrl_size-1 downto 0);
			variable src_addr	:std_logic_vector(addr_width-1 downto 0);
			variable drain_addr	:std_logic_vector(addr_width-1 downto 0);
			
			variable north		:io_bank(cells_axis-1 downto 0);
			variable west		:io_bank(cells_axis-1 downto 0);
			variable south		:io_bank(cells_axis-1 downto 0);
			variable east		:io_bank(cells_axis-1 downto 0);
		begin
			
			while not endfile(stimuli_file) loop
				readline(stimuli_file, L);
				
				read(L,rc_ctrl);
				read(L,src_addr);
				read(L,drain_addr);
				read(L,north);
				read(L,east);
				read(L,west);
				read(L,south);
				perform_simulation(rc_ctrl,"000",src_addr,drain_addr,north,west,south,east);
			end loop;
		end procedure;
		
	begin
		rc_ctrl_i <= (others => '0');
		src_base_addr_i <= (others => '0');
		dest_base_addr_i <= (others => '0');
		north_bank_i <= (others => (others => '0'));
		west_bank_i <= (others => (others => '0'));
		reset_uut;
		
		-- Sync with clocks
		wait until rising_edge(test_clk);
		
		random_stimuli;
		
		-- Hack
		running <= 0;
		wait;
	end process;

end architecture;