library ieee;
use ieee.std_logic_1164.all;
use work.common_defs.all;
use work.version1_defs.all;

entity cell is
	generic (
		width		: positive := 16;
		config_width	: positive := 2
	);
	port (
		config_clk	: in std_logic;
		data_clk	: in std_logic;
		compute_clk	: in std_logic;
		
		reset		: in std_logic;
		
		reconfigure	: in std_logic;
		
		config_in	: in std_logic_vector(config_width -1 downto 0);
		config_out	: out std_logic_vector(config_width -1 downto 0);
		
		north_idle	: in std_logic;
		west_idle	: in std_logic;
		south_idle	: out std_logic;
		east_idle	: out std_logic;
		
		-- With room for valid signal
		north		: in std_logic_vector(width downto 0);
		west		: in std_logic_vector(width downto 0);
		
		south		: out std_logic_vector(width downto 0);
		east		: out std_logic_vector(width downto 0)
		
	);
end entity;

architecture core of cell is
	attribute keep_hierarchy: string;
	attribute keep_hierarchy of core: architecture is "yes";
	
	constant exe_inputs_log	: positive	:= 2;
	constant data_outs_log	: positive	:= 3;
	constant config_bits	: positive	:= config_size-(width+1);
	
	signal north_reg	: std_logic_vector(width downto 0);
	signal west_reg		: std_logic_vector(width downto 0);
	
	signal out_south	: std_logic_vector(width downto 0);
	signal out_east		: std_logic_vector(width downto 0);

	-- Rooms for data + valid
	signal main_alu_input_a	: std_logic_vector(width downto 0);
	signal main_alu_input_b	: std_logic_vector(width downto 0);
	signal main_alu_op	: std_logic_vector(alu_op_size -1 downto 0);
	signal main_alu_out	: std_logic_vector(width downto 0);
	
	signal alu_input_a_sel	: std_logic_vector(exe_inputs_log - 1 downto 0);
	signal alu_input_b_sel	: std_logic_vector(exe_inputs_log - 1 downto 0);

	signal alu_mux_inputs	: std_logic_vector((2**exe_inputs_log)*(width+1) -1 downto 0);
	signal mul_mux_inputs	: std_logic_vector((2**exe_inputs_log)*(width+1) -1 downto 0);

	signal mul_input_a_sel	: std_logic_vector(exe_inputs_log - 1 downto 0);
	signal mul_input_b_sel	: std_logic_vector(exe_inputs_log - 1 downto 0);

	signal muldiv_input_a	: std_logic_vector(width downto 0);
	signal muldiv_input_b	: std_logic_vector(width downto 0);
	signal muldiv_op	: std_logic_vector(0 downto 0);
	signal muldiv_output	: std_logic_vector(width downto 0);

	signal state_output	: std_logic_vector(width downto 0);
	signal state_input	: std_logic_vector(width downto 0);
	
	signal state_mux_sel	: std_logic_vector(data_outs_log -1 downto 0);
	signal south_mux_sel	: std_logic_vector(data_outs_log -1 downto 0);
	signal east_mux_sel	: std_logic_vector(data_outs_log -1 downto 0);
	signal data_outputs	: std_logic_vector((2**data_outs_log)*(width+1) -1 downto 0);
	
	signal cs_bus		: std_logic_vector(config_width -1 downto 0);
	signal rc_bus		: std_logic_vector(config_width -1 downto 0);

	signal config		: std_logic_vector(config_bits-1 downto 0);
	signal config_data_in	: std_logic_vector(config_bits-1 downto 0);

	signal restart		: std_logic;
	signal reg_reconfigure	: std_logic;
	signal reg_idle		: std_logic;
	signal write_enable	: std_logic;
begin
	restart <= reset and (not data_clk);
	
	south_idle <= reg_idle;
	east_idle <= reg_idle;
	
	alu_input_mux_a: component mux
		generic map(
			path_width => (width +1),
			sel_width => exe_inputs_log
		)
		port map(
			input => alu_mux_inputs,
			sel => alu_input_a_sel,
			output => main_alu_input_a
		);
	
	alu_input_mux_b: component mux
		generic map(
			path_width => (width +1),
			sel_width => exe_inputs_log
		)
		port map(
			input => alu_mux_inputs,
			sel => alu_input_b_sel,
			output => main_alu_input_b
		);
	
	
	main_alu: component alu
		generic map(
			width => width
		)
		port map(
			clk => compute_clk,
			reset => restart,
			data_a => main_alu_input_a(width-1 downto 0),
			data_b => main_alu_input_b(width-1 downto 0),
			valid_a => main_alu_input_a(width),
			valid_b => main_alu_input_b(width),
			alu_op => main_alu_op,
			output => main_alu_out(width-1 downto 0),
			valid_output => main_alu_out(width)
		);
	
	muldiv_input_mux_a: component mux
		generic map(
			path_width => (width +1),
			sel_width => exe_inputs_log
		)
		port map(
			input => mul_mux_inputs,
			sel => mul_input_a_sel,
			output => muldiv_input_a
		);
	
	muldiv_input_mux_b: component mux
		generic map(
			path_width => (width +1),
			sel_width => exe_inputs_log
		)
		port map(
			input => mul_mux_inputs,
			sel => mul_input_b_sel,
			output => muldiv_input_b
		);
	
	muldiv: component multiplydivider
		generic map(
			width => width
		)
		port map(
			clk => compute_clk,
			reset => restart,
			opcode => muldiv_op(0),
			valid_a => muldiv_input_a(width),
			valid_b => muldiv_input_b(width),
			data_a => muldiv_input_a(width-1 downto 0),
			data_b => muldiv_input_b(width-1 downto 0),
			valid_output => muldiv_output(width),
			output => muldiv_output(width-1 downto 0)
		);
	
	state_input_mux: component mux
		generic map(
			path_width => (width+1),
			sel_width => data_outs_log
		)
		port map(
			input => data_outputs,
			sel => state_mux_sel,
			output => state_input
		);
	
	state_register: component sc_register
		generic map(
			width => (width+1),
			config_width => config_width
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			config_in => cs_bus,
			config_out => rc_bus,
			config_enable => reg_reconfigure,
			data_in => state_input,
			data_out => state_output,
			write => write_enable
		);
	write_enable <= not reg_idle;
	config_register: component sc_register
		generic map(
			width => config_bits,
			config_width => config_width
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			config_in => config_in,
			config_out => cs_bus,
			config_enable => reg_reconfigure,
			data_in => config_data_in,
			data_out => config,
			write => '0'
		);
	config_data_in <= (others => '0');
	
	
	south_output_mux: component mux
		generic map(
			path_width => (width+1),
			sel_width => data_outs_log
		)
		port map(
			input => data_outputs,
			sel => south_mux_sel,
			output => out_south
		);
	
	east_output_mux: component mux
		generic map(
			path_width => (width+1),
			sel_width => data_outs_log
		)
		port map(
			input => data_outputs,
			sel => east_mux_sel,
			output => out_east
		);
	
	regs : process(reset, data_clk)
	begin
		if reset = '0' then
			north_reg <= (north_reg'range => '0');
			west_reg <= (west_reg'range => '0');
			reg_reconfigure <= '0';
			reg_idle <= '0';
		elsif rising_edge(data_clk) then 
			north_reg <= north;
			west_reg <= west;
			reg_reconfigure <= reconfigure;
			reg_idle <= north_idle or west_idle;
		end if;
	end process;
	
	alu_mux_inputs <= muldiv_output & state_output & north_reg & west_reg;
	mul_mux_inputs <= main_alu_out & state_output & north_reg & west_reg;
	data_outputs(data_outputs'left downto 5*(width+1)) <= (others => '0');
	data_outputs(5*(width+1)-1 downto 0 ) <=  muldiv_output & main_alu_out & state_output & north_reg & west_reg;
	
	alu_input_a_sel <= config(20 downto 19);
	alu_input_b_sel <= config(18 downto 17);
	main_alu_op <= config(16 downto 14);
	mul_input_a_sel <= config(13 downto 12);
	mul_input_b_sel <= config(11 downto 10);
	muldiv_op <= config(9 downto 9);
	state_mux_sel <= config(8 downto 6);
	east_mux_sel <= config(5 downto 3);
	south_mux_sel <= config(2 downto 0);
	
	south(south'left) <= out_south(out_south'left) and not (reg_reconfigure or reg_idle);
	south(south'left -1 downto 0) <= out_south(out_south'left -1 downto 0);
	east(east'left) <= out_east(out_east'left) and not (reg_reconfigure or reg_idle);
	east(east'left -1 downto 0) <= out_east(out_east'left -1 downto 0);
	
	config_out <= rc_bus when reg_reconfigure = '1' else config_in;
	
end architecture;