library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version1_defs.all;

entity rc_writer is
	generic (
		config_width	: positive	:= 2;
		config_size	: positive	:= 38;
		address_width	: positive	:= 8;
		pattern_width	: positive	:= 4
	);
	port (
		data_clk	: in std_logic;
		config_clk	: in std_logic;
		reset		: in std_logic;
		
		-- Interface
		reconfigure	: in std_logic;
		store_address	: in std_logic_vector(address_width -1 downto 0);
		rc_pattern	: in std_logic_vector(pattern_width -1 downto 0);
		vertical	: in std_logic;
		unit_offset	: in std_logic_vector(address_width -1 downto 0);
		
		-- To next unit
		rc_next		: out std_logic;
		address_next	: out std_logic_vector(address_width -1 downto 0);
		rc_pattern_next : out std_logic_vector(pattern_width -1 downto 0);
		vertical_next	: out std_logic;
		unit_offset_next: out std_logic_vector(address_width -1 downto 0);
		
		-- Memory I/O
		write_address	: out std_logic_vector(address_width -1 downto 0);
		write_data	: out std_logic_vector(config_size -1 downto 0);
		write_enable	: out std_logic;
		
		config_in	: in std_logic_vector(config_width -1 downto 0)
	);
end entity;

architecture first of rc_writer is
	signal reg_pattern	: std_logic_vector(pattern_width -1 downto 0);
	signal reg_vertical	: std_logic;
	signal reg_base_address	: std_logic_vector(address_width -1 downto 0);
	signal reg_rc		: std_logic;
	signal reg_unit_offset	: std_logic_vector(address_width -1 downto 0);
	
	signal en_update	: std_logic;
	signal en_shift		: std_logic;
	signal reg_config	: std_logic_vector(config_size -1 downto 0);
	signal prev_config	: std_logic_vector(config_size -1 downto 0);
	
	signal en_address	: std_logic;
	signal reg_address	: std_logic_vector(address_width-1 downto 0);
	signal next_address	: std_logic_vector(address_width -1 downto 0);
	
	type memory_states	is (idle, retrieving, save, store, dummy, sleep);
	signal mem_state	: memory_states;
	signal next_mem_state	: memory_states;
	signal exit_counter	: integer range 0 to (config_size/config_width)-1;
	signal reconf_shifter	: std_logic_vector(pattern_width -1 downto 0);
	signal shifter_input	: std_logic;

	signal data_in		: std_logic_vector(config_size -1 downto 0);
begin
	
	shift_out: component sc_register 
		generic map(
			width => config_size,
			config_width => config_width,
			allow_prohibit => '0'
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			config_in => config_in,
			config_enable => '1', -- ?
			-- data_out not connected.. don't care
			data_in => data_in,
			data_out => prev_config,
			write => '0'
		);
	data_in <= (others => '0');
	
	ctrl_clk : process(data_clk, reset)
	begin
		if reset = '0' then
			reg_pattern <= (others => '0');
			reg_base_address <= (others => '0');
			reg_rc <= '0';
			reg_unit_offset <= (others => '0');
			reg_vertical <= '0';
		elsif rising_edge(data_clk) then
			reg_pattern <= rc_pattern;
			reg_vertical <= vertical;
			reg_base_address <= store_address;
			reg_rc <= reconfigure;
			reg_unit_offset <= unit_offset;
		end if;
	end process;
	
	perform_clk : process(config_clk, reset) 
	begin
		if reset = '0' then
			mem_state <= idle;
			exit_counter <= 0;
			reconf_shifter <= (others => '0');
			reg_config <= (others => '0');
		elsif rising_edge(config_clk) then
			mem_state <= next_mem_state;
			if en_update = '1' then
				reg_config <= prev_config;
			end if;
			if en_shift = '1' then
				reconf_shifter(reconf_shifter'left downto 1) <= reconf_shifter(reconf_shifter'left -1 downto 0);
				reconf_shifter(0) <= shifter_input;
			end if;
			if en_address = '1' then
				reg_address <= next_address;
			end if;
			if exit_counter = ((config_size/config_width) -1) or reg_rc = '0' then
				exit_counter <= 0;
			else
				exit_counter <= exit_counter + 1;
			end if;
		end if;
	end process;
	
	mem_state_next : process(mem_state,reg_pattern,reg_vertical,reg_base_address,reg_config,reg_rc,exit_counter,reconf_shifter,reg_address)
	begin
		write_address <= (others => '0');
		write_enable <= '0';
		write_data <= (others => '0');
		en_update <= '0';
		en_shift <= '0';
		shifter_input <= '0';
		next_address <= (others => '0');
		en_address <= '0';
		case mem_state is
		when idle => 
			-- Yes are using signals from a different clock domain here. But the clocks 
			-- should be in sync so this _should_ be okay without syncronization
			en_shift <= '1';
			en_update <= '1';
			if reg_rc = '1' then 
				shifter_input <= '1';
				en_address <= '1';
				next_address <= reg_base_address;
				if reg_pattern(0) = '0' or reg_vertical = '0' then
					next_mem_state <= dummy;
				else 
					next_mem_state <= save;
				end if;
			else 
				shifter_input <= '0';
				next_mem_state <= idle;
			end if;
		
		when retrieving => 
			en_shift <= '1'; -- ?
			en_update <= '1';
			if reg_pattern(0) = '0' or reg_vertical = '0' then
				next_mem_state <= dummy;
			else 
				next_mem_state <= save;
			end if;
		
		when dummy =>
			next_mem_state <= sleep;
		
		when save => 
			next_mem_state <= store;
		
		when store => 
			write_address <= reg_address;
			write_data <= reg_config;
			write_enable <= '1';
			en_address <= '1';
			next_address <= std_logic_vector(unsigned(reg_address) + 1);
			next_mem_state <= sleep;
		
		when sleep => 
			if exit_counter = ((config_size/config_width) -1) then
				if reconf_shifter(reconf_shifter'left) = '1' then
					next_mem_state <= idle;
				else
					next_mem_state <= retrieving;
				end if;
			else 
				next_mem_state <= sleep;
			end if;
		
		end case;
	end process;
	
	-- Rotate pattern to next unit
	rc_pattern_next(reg_pattern'left) <= reg_pattern(reg_pattern'right);
	rc_pattern_next(reg_pattern'left -1 downto 0) <= reg_pattern(reg_pattern'left downto 1);
	
	address_next <= std_logic_vector(unsigned(reg_base_address)+unsigned(reg_unit_offset)) when reg_pattern(0) = '1' else 
			reg_base_address;
	
	unit_offset_next <= reg_unit_offset;
	
	rc_next <= reg_rc;
	vertical_next <= reg_vertical;
	
end architecture;
