library ieee;
use ieee.std_logic_1164.all;
use work.common_defs.all;
use work.version1_defs.all;

entity wrpu is 
	generic(
		data_width	: positive	:= io_size;
		addr_width	: positive	:= 8;
		config_width	: positive	:= 2;
		config_size	: positive	:= config_size;
		header_size	: positive	:= 24;
		flags_size	: positive	:= 4;
		cells_axis	: natural	:= 4
	);
	port (
		data_clk	: in std_logic;
		config_clk	: in std_logic;
		compute_clk	: in std_logic;
		
		reset		: in std_logic;
		
		rc_ctrl		: in std_logic_vector(rcm_ctrl_size-1 downto 0);
		rc_status	: out std_logic_vector(rcm_status_size-1 downto 0);
		src_base_addr	: in std_logic_vector(addr_width-1 downto 0);
		dest_base_addr	: in std_logic_vector(addr_width-1 downto 0);
		
		north_bank	: in io_bank(cells_axis-1 downto 0);
		west_bank	: in io_bank(cells_axis-1 downto 0);
		
		south_bank	: out io_bank(cells_axis-1 downto 0);
		east_bank	: out io_bank(cells_axis-1 downto 0)
	);
end entity;

architecture wrpu_cellNxN of wrpu is
	signal rcm_read_addr	: std_logic_vector(addr_width -1 downto 0);
	signal rcm_read_data	: std_logic_vector(header_size-1 downto 0);
	signal rcm_write	: std_logic;
	signal rcm_write_addr	: std_logic_vector(addr_width -1 downto 0);
	signal rcm_write_data	: std_logic_vector(header_size -1 downto 0);
	
	signal rcm_flags	: std_logic_vector(flags_size-1 downto 0);
	
	signal m_rcr_rc		: std_logic_vector(cells_axis downto 0);
	signal m_rcr_vertical	: std_logic_vector(cells_axis downto 0);
	type addr_array		is array(cells_axis downto 0) of std_logic_vector(addr_width-1 downto 0);
	signal m_rcr_ld_addr	: addr_array;
	signal m_rcr_offset	: addr_array;
	
	type axis_array		is array(cells_axis downto 0) of std_logic_vector(cells_axis-1 downto 0);
	signal m_rcr_pattern	: axis_array;
	
	
	type rc_array		is array(cells_axis-1 downto 0) of std_logic_vector(cells_axis-1 downto 0);
	signal m_rcr_reconfigure: rc_array;
	type config_array	is array(cells_axis-1 downto 0) of std_logic_vector(config_width -1 downto 0);
	signal m_rcr_config	: config_array;
	
	signal m_rcram_write	: std_logic_vector(cells_axis-1 downto 0);
	type rcram_addr_array	is array(cells_axis-1 downto 0) of std_logic_vector(addr_width-1 downto 0);
	signal m_rcram_waddr	: rcram_addr_array;
	type rcram_bus_array	is array(cells_axis-1 downto 0) of std_logic_vector(config_size-1 downto 0);
	signal m_rcram_wdata	: rcram_bus_array;
	signal m_rcram_raddr	: rcram_addr_array;
	signal m_rcram_rdata	: rcram_bus_array;
	
	signal m_rcw_rc		: std_logic_vector(cells_axis downto 0);
	signal m_rcw_st_addr	: addr_array;
	signal m_rcw_pattern	: axis_array;
	signal m_rcw_vertical	: std_logic_vector(cells_axis downto 0);
	signal m_rcw_offset	: addr_array;
	signal m_rcw_config	: config_array;
	
	-- To stage yield easier io_bank connection
	type cell_connect	is array(cells_axis downto 0) of io_bank(cells_axis-1 downto 0);
	signal cell_north_con	: cell_connect;
	signal cell_west_con	: cell_connect;
	
	type idle_connect	is array(cells_axis downto 0) of std_logic_vector(cells_axis-1 downto 0);
	signal m_idle_row	: idle_connect;
	signal init_idle	: std_logic;
begin
	rcm : component rc_master
		generic map(
			addr_width => addr_width,
			rc_units => cells_axis,
			addr_consume => 0,
			header_size => header_size
		) 
		port map(
			clk => data_clk,
			reset => reset,
			ctrl => rc_ctrl,
			status => rc_status,
			src_base_addr => src_base_addr,
			dest_base_addr => dest_base_addr,
			flags => rcm_flags,
			read_addr => rcm_read_addr,
			read_data => rcm_read_data,
			write_en => rcm_write,
			write_addr => rcm_write_addr,
			write_data => rcm_write_data,
			rcr_rc => m_rcr_rc(0),
			rcr_addr => m_rcr_ld_addr(0),
			rcr_pattern => m_rcr_pattern(0),
			rcr_vertical => m_rcr_vertical(0),
			rcr_offset => m_rcr_offset(0),
			rcw_rc => m_rcw_rc(0),
			rcw_addr => m_rcw_st_addr(0),
			rcw_pattern => m_rcw_pattern(0),
			rcw_vertical => m_rcw_vertical(0),
			rcw_offset => m_rcw_offset(0)
		);
	rcm_ram : component rc_ram
		generic map(
			data_width => header_size,
			addr_width => addr_width
		)
		port map(
			clk => data_clk,
			write => rcm_write,
			waddr => rcm_write_addr,
			wdata => rcm_write_data,
			raddr => rcm_read_addr,
			rdata => rcm_read_data
		);
	
	rtm	: component runtime_manager
		generic map(
			axis => cells_axis
		)
		port map(
			clk => data_clk,
			reset => reset,
			
			override => rcm_flags(0),
			rc_active => m_rcr_rc(0),
			vertical => m_rcr_vertical(0),
			idle_out => init_idle
		);
	
	rc_readers : for i in 0 to cells_axis-1 generate
	begin
		
		rcr : component rc_reader
			generic map(
				config_width => config_width,
				config_size => config_size,
				address_width => addr_width,
				pattern_width => cells_axis
			)
			port map(
				data_clk => data_clk,
				config_clk => config_clk,
				reset => reset,
				reconfigure => m_rcr_rc(i),
				load_address => m_rcr_ld_addr(i),
				rc_pattern => m_rcr_pattern(i),
				vertical => m_rcr_vertical(i),
				unit_offset => m_rcr_offset(i),
				rc_next => m_rcr_rc(i+1),
				address_next => m_rcr_ld_addr(i+1),
				rc_pattern_next => m_rcr_pattern(i+1),
				vertical_next => m_rcr_vertical(i+1),
				unit_offset_next => m_rcr_offset(i+1),
				read_address => m_rcram_raddr(i),
				read_data => m_rcram_rdata(i),
				reconf_out => m_rcr_reconfigure(i),
				config_out => m_rcr_config(i)
			);
	end generate;
	
	rc_rams : for i in 0 to cells_axis -1 generate
	begin
		rcram : component rc_ram
			generic map(
				data_width => config_size,
				addr_width => addr_width
			)
			port map(
				clk => config_clk,
				write => m_rcram_write(i),
				waddr => m_rcram_waddr(i),
				wdata => m_rcram_wdata(i),
				raddr => m_rcram_raddr(i),
				rdata => m_rcram_rdata(i)
			);
	end generate;
	
	rc_writers: for i in 0 to cells_axis -1 generate
	begin
		rcw : component rc_writer
			generic map(
				config_width => config_width,
				config_size => config_size,
				address_width => addr_width,
				pattern_width => cells_axis
			)
			port map(
				data_clk => data_clk,
				config_clk => config_clk,
				reset => reset,
				reconfigure => m_rcw_rc(i),
				store_address => m_rcw_st_addr(i),
				rc_pattern => m_rcw_pattern(i),
				vertical => m_rcw_vertical(i),
				unit_offset => m_rcw_offset(i),
				rc_next => m_rcw_rc(i+1),
				address_next => m_rcw_st_addr(i+1),
				rc_pattern_next => m_rcw_pattern(i+1),
				vertical_next => m_rcw_vertical(i+1),
				unit_offset_next => m_rcw_offset(i+1),
				write_address => m_rcram_waddr(i),
				write_data => m_rcram_wdata(i),
				write_enable => m_rcram_write(i),
				config_in => m_rcw_config(i)
			);
	end generate;
	
	cell_column: for i in 0 to cells_axis -1 generate
		type cell_cfg_bus is array(0 to cells_axis) of std_logic_vector(config_width-1 downto 0);
		signal config_bus	: cell_cfg_bus;
		signal idle_col		: std_logic_vector(0 to cells_axis);
	begin
		-- Control via row input 
		idle_col(0) <= '0';
		config_bus(0) <= m_rcr_config(i);
		m_rcw_config(i) <= config_bus(config_bus'high);
		cell_row: for j in 0 to cells_axis -1 generate
		begin
			single_cell : component cell
				generic map(
					width => data_width,
					config_width => config_width
				)
				port map(
					config_clk => config_clk,
					data_clk => data_clk,
					compute_clk => compute_clk,
					reset => reset,
					reconfigure => m_rcr_reconfigure(i)(j),
					config_in => config_bus(j),
					config_out => config_bus(j+1),
					north_idle => idle_col(j),
					south_idle => idle_col(j+1),
					west_idle => m_idle_row(i)(j),
					east_idle => m_idle_row(i+1)(j),
					north => cell_north_con(j)(i),
					west => cell_west_con(i)(j),
					south => cell_north_con(j+1)(i),
					east => cell_west_con(i+1)(j)
				);
		end generate;
	end generate;
	
	cell_north_con(0)(cells_axis-1 downto 0) <= north_bank;
	cell_west_con(0)(cells_axis-1 downto 0) <= west_bank;
	south_bank <= cell_north_con(cells_axis)(cells_axis-1 downto 0);
	east_bank <= cell_west_con(cells_axis)(cells_axis-1 downto 0);
	
	m_idle_row(0) <= ( 0 => init_idle, others => '0');
	
end architecture;