library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
use work.common_defs.all;
use work.version2_defs.all;

entity rc_master is 
	generic (
		addr_width	:positive	:= 8;
		header_size	:positive	:= 24;
		rc_units	:positive	:= 4;
		system_version	:natural	:= 1;
		version_width	:positive	:= 4;
		flags_width	:positive	:= 4;
		size_width	:positive	:= 8;
		addr_consume	:integer	:= 3
	);
	port (
		clk		:in std_logic;
		reset		:in std_logic;
		
		-- Interface
		status		:out std_logic_vector(rcm_status_size-1 downto 0);
		ctrl		:in std_logic_vector(rcm_ctrl_size-1 downto 0);
		src_base_addr	:in std_logic_vector(addr_width -1 downto 0);
		dest_base_addr	:in std_logic_vector(addr_width -1 downto 0);
		
		-- General
		flags		:out std_logic_vector(flags_width -1 downto 0);
		
		-- For configuration headers
		read_addr	:out std_logic_vector(addr_width-1 downto 0);
		read_data	:in std_logic_vector(header_size -1 downto 0);
		write_en	:out std_logic;
		write_addr	:out std_logic_vector(addr_width-1 downto 0);
		write_data	:out std_logic_vector(header_size -1 downto 0);
		
		-- Read units
		rcr_rc		:out std_logic;
		rcr_addr	:out std_logic_vector(addr_width -1 downto 0);
		rcr_pattern	:out std_logic_vector(rc_units-1 downto 0);
		rcr_vertical	:out std_logic;
		rcr_offset	:out std_logic_vector(addr_width -1 downto 0);
		rcr_version	:out std_logic;
		
		-- Write units
		rcw_rc		:out std_logic;
		rcw_addr	:out std_logic_vector(addr_width -1 downto 0);
		rcw_pattern	:out std_logic_vector(rc_units-1 downto 0);
		rcw_vertical	:out std_logic;
		rcw_offset	:out std_logic_vector(addr_width -1 downto 0);
		rcw_version	:out std_logic
	);
end entity;


architecture rcm of rc_master is
	type rc_master_states	is (idle, read_header, parse_header, vertical_pattern, start_rc, reconfiguring, reconfigured, rc_error);
	signal next_state	: rc_master_states;
	signal master_state	: rc_master_states;
	
	signal counter		: integer range 0 to rc_units;
	signal clear_rc_counter : std_logic;
	
	signal update_ctrl	: std_logic;
	signal reg_conf		: std_logic;
	signal reg_ctrl		: std_logic_vector(rcm_ctrl_size-1 downto 0);
	signal reg_dest		: std_logic_vector(addr_width-1 downto 0);
	signal add_dest		: std_logic_vector(addr_width-1 downto 0);
	signal reg_src		: std_logic_vector(addr_width-1 downto 0);
	signal add_src		: std_logic_vector(addr_width-1 downto 0);
	signal reg_add_src	: std_logic_vector(addr_width-1 downto 0);
	
	signal read_version	: std_logic_vector(version_width-1 downto 0);
	signal read_flags	: std_logic_vector(flags_width-1 downto 0);
	signal read_x_size	: std_logic_vector(size_width-1 downto 0);
	signal read_y_size	: std_logic_vector(size_width-1 downto 0);
	
	signal rom_addr		: std_logic_vector(integer(ceil(log2(real(rc_units+1))))-1 downto 0);
	signal pattern_output	: std_logic_vector(rc_units-1 downto 0);
	
	signal unit_offset	: std_logic_vector(addr_width-1 downto 0);
	signal unit_offset_v1	: std_logic_vector(addr_width-1 downto 0);
	signal unit_offset_v2	: std_logic_vector(addr_width-1 downto 0);
	
	signal update_hdr_prs	: std_logic;
	signal reg_y_size	: std_logic_vector(size_width-1 downto 0);
	
	signal set_vertical	: std_logic;
	
	signal set_rcr_vertical : std_logic;
	signal set_rcw_vertical : std_logic;
	
	
	signal reg_header	: std_logic_vector(header_size-1 downto 0);
	signal reg_pattern	: std_logic_vector(rc_units-1 downto 0);
	signal reg_offset	: std_logic_vector(addr_width-1 downto 0);
	signal reg_vertical	: std_logic_vector(rc_units-1 downto 0);

	signal reg_rcr_vertical	: std_logic_vector(rc_units-1 downto 0);
	signal reg_rcw_vertical	: std_logic_vector(rc_units-1 downto 0);
	
	signal cur_header	: std_logic_vector(header_size-1 downto 0);
	signal cur_pattern	: std_logic_vector(rc_units-1 downto 0);
	signal cur_vertical	: std_logic_vector(rc_units-1 downto 0);
	signal cur_offset	: std_logic_vector(addr_width-1 downto 0);
	
	signal rcw_init_rc	: std_logic;
	signal rcw_d1_rc	: std_logic;
	signal rcw_d1_addr	: std_logic_vector(addr_width-1 downto 0);
	signal rcw_d1_pattern	: std_logic_vector(rc_units-1 downto 0);
	signal rcw_d1_vertical	: std_logic;
	signal rcw_d1_offset	: std_logic_vector(addr_width-1 downto 0);
	signal rcw_d1_version	: std_logic;
	signal rcw_d2_rc	: std_logic;
	signal rcw_d2_addr	: std_logic_vector(addr_width-1 downto 0);
	signal rcw_d2_pattern	: std_logic_vector(rc_units-1 downto 0);
	signal rcw_d2_vertical	: std_logic;
	signal rcw_d2_offset	: std_logic_vector(addr_width-1 downto 0);
	signal rcw_d2_version	: std_logic;
	
	signal set_current	: std_logic;
begin
	read_version <= read_data(read_data'left downto read_data'left+1-version_width);
	read_flags <= read_data(read_data'left - version_width downto read_data'left+1 -version_width - flags_width);
	read_y_size <= read_data(read_data'left - version_width - flags_width downto read_data'left+1 - version_width - flags_width - size_width);
	read_x_size <= read_data(read_data'left - version_width - flags_width - size_width downto 0);
	
	pattern_rom : rc_master_rom
		generic map(
			data_width => rc_units,
			elements => (rc_units +1),
			init_source => "roms/master_prom"
		)
		port map(
			addr => rom_addr,
			output => pattern_output
		);
	
	v1_offset_rom : rc_master_rom
		generic map(
			data_width => addr_width,
			elements => (rc_units +1),
			init_source => "roms/master_orom"
		)
		port map(
			addr => rom_addr,
			output => unit_offset_v1
		);
	v2_offset_rom : rc_master_rom
		generic map(
			data_width => addr_width,
			elements => (rc_units +1),
			init_source => "roms/master_orom_v2"
		)
		port map(
			addr => rom_addr,
			output => unit_offset_v2
		);
	
	unit_offset <= unit_offset_v2 when reg_header(header_size - version_width) = '1' else unit_offset_v1;
	
	reg_y_size <= reg_header(reg_header'left - version_width - flags_width downto reg_header'left+1 - version_width - flags_width - size_width);	
	rcm_states : process(clk,reset)
	begin
		if reset = '0' then
			master_state <= idle;
			reg_conf <= '0';
			reg_offset <= (others => '0');
			reg_header <= (others => '0');
			reg_pattern <= (others => '0');
			reg_vertical <= (others => '0');
			reg_dest <= (others => '0');
			reg_src <= (others => '0');
			reg_add_src <= (others => '0');
			reg_ctrl <= (others => '0');
			reg_rcr_vertical <= (others => '0');
			reg_rcw_vertical <= (others => '0');
			cur_header <= (others => '0');
			cur_pattern <= (others => '0');
			cur_vertical <= (others => '0');
			cur_offset <= (others => '0');
			rcw_d1_rc <= '0';
			rcw_d1_addr <= (others => '0');
			rcw_d1_pattern <= (others => '0');
			rcw_d1_offset <= (others => '0');
			rcw_d1_vertical <= '0';
			rcw_d1_version <= '0';
			rcw_d2_rc <= '0';
			rcw_d2_addr <= (others => '0');
			rcw_d2_pattern <= (others => '0');
			rcw_d2_offset <= (others => '0');
			rcw_d2_vertical <= '0';
			rcw_d2_version <= '0';
		elsif rising_edge(clk) then
			master_state <= next_state;
			if update_ctrl = '1' then
				reg_ctrl <= ctrl;
			end if;
			if next_state = idle then
				reg_dest <= dest_base_addr;
				reg_src	<= src_base_addr;
			end if;
			
			if clear_rc_counter = '1' then
				counter <= 0;
			elsif counter = rc_units then
				counter <= 0;
			else
				counter <= counter + 1;
			end if;
			
			if update_hdr_prs = '1' then
				reg_header <= read_data;
				reg_pattern <= pattern_output;
				reg_add_src <= add_src;
			end if;
			if set_vertical = '1' then
				reg_vertical <= pattern_output;
				reg_offset <= unit_offset;
			end if;
			
			if set_rcr_vertical = '1' then
				reg_rcr_vertical <= reg_vertical;
			else 
				reg_rcr_vertical <= "0" & reg_rcr_vertical(reg_rcr_vertical'left downto 1);
			end if;
			
			if set_rcw_vertical = '1' then
				reg_rcw_vertical <= cur_vertical;
			else
				reg_rcw_vertical <= "0" & reg_rcw_vertical(reg_rcw_vertical'left downto 1);
			end if;
			
			if set_current = '1' then
				cur_header <= reg_header;
				cur_pattern <= reg_pattern;
				cur_vertical <= reg_vertical;
				cur_offset <= reg_offset;
				-- Store the system status
				reg_conf <= reg_ctrl(0);
			end if;
			
			rcw_d1_rc <= rcw_init_rc;
			rcw_d1_addr <= add_dest;
			rcw_d1_pattern <= cur_pattern;
			rcw_d1_offset <= cur_offset;
			rcw_d1_vertical <= reg_rcw_vertical(0);
			rcw_d1_version <= cur_header(header_size - version_width); 
			rcw_d2_rc <= rcw_d1_rc;
			rcw_d2_addr <= rcw_d1_addr;
			rcw_d2_pattern <= rcw_d1_pattern;
			rcw_d2_offset <= rcw_d1_offset;
			rcw_d2_vertical <= rcw_d1_vertical;
			rcw_d2_version <= rcw_d1_version;
		end if;
	end process;
	
	rcm_action : process(master_state,reg_ctrl, read_version, read_x_size, reg_y_size, reg_src, reg_dest, counter, cur_header, pattern_output, reg_conf) 
	
	begin
		status(status'high downto 1) <= (others => '0');
		write_en <= '0';
		write_addr <= (others => '0');
		write_data <= (others => '0');
		read_addr <= (others => '0');
		rom_addr <= (others => '0');
		update_hdr_prs <= '0';
		set_vertical <= '0';
		rcr_rc <= '0';
		rcw_init_rc <= '0';
		set_rcr_vertical <= '0';
		set_rcw_vertical <= '0';
		clear_rc_counter <= '0';
		set_current <= '0';
		update_ctrl <= '0';
		
		case master_state is
		when idle =>
			if reg_ctrl(0) = '1' or reg_ctrl(1) = '1' then
				if reg_ctrl(1) = '1' and reg_conf = '0' then
					next_state <= rc_error;
				else 
					next_state <= read_header;
				end if;
			else 
				next_state <= idle;
				update_ctrl <= '1';
			end if;
		when read_header =>
			if reg_ctrl(0) = '1' then -- new configuration
				read_addr <= reg_src; -- Request configuration header
				next_state <= parse_header;
			else
				next_state <= start_rc;
			end if;
			status(1) <= '1'; -- Busy
		when parse_header =>
			rom_addr <= read_x_size(rom_addr'left downto 0);
			if unsigned(read_version) > system_version or read_x_size(read_x_size'high downto rom_addr'high+1) /= (read_x_size'high downto rom_addr'high+1 => '0') or pattern_output = (pattern_output'range => '0') then
				-- Device to old 
				next_state <= rc_error;
			else 
				update_hdr_prs <= '1';
				next_state <= vertical_pattern;
			end if;
			status(1) <= '1'; -- Busy
		when vertical_pattern =>
			rom_addr <= reg_y_size(rom_addr'left downto 0);
			if reg_y_size(reg_y_size'high downto rom_addr'high+1) /= (reg_y_size'high downto rom_addr'high+1 => '0') or pattern_output = (pattern_output'range => '0') then
				next_state <= rc_error;
			else
				set_vertical <= '1';
				next_state <= start_rc;
			end if;
			status(1) <= '1'; -- Busy
		
		when start_rc =>
			clear_rc_counter <= '1';
			set_rcr_vertical <= '1';
			set_rcw_vertical <= '1';
			if reg_ctrl(1) = '1' and reg_conf = '1' then
				write_addr <= reg_dest;
				write_data <= cur_header;
				write_en <= '1';
			end if;
			
			status(1) <= '1'; -- Busy
			next_state <= reconfiguring;
		
		when reconfiguring =>
			rcr_rc <= reg_ctrl(0);
			rcw_init_rc <= reg_ctrl(1);
			if counter = rc_units-1 then
				next_state <= reconfigured;
			else
				next_state <= reconfiguring;
			end if;
			status(1) <= '1'; -- Busy
		when reconfigured => 
			-- Move register values 
			set_current <= '1';
			update_ctrl <= '1';
			next_state <= idle;
			status(1) <= '1'; -- Busy
		when rc_error =>
			update_ctrl <= '1';
			status(2) <= '1';
			if reg_ctrl(2) = '1' then
				next_state <= idle;
			else
				next_state <= rc_error;
			end if;
		end case;
	end process;
	
	add_dest <= std_logic_vector(unsigned(reg_dest)+addr_consume);
	add_src <= std_logic_vector(unsigned(reg_src)+addr_consume);
	
	rcr_pattern <= reg_pattern;
	rcr_addr <= reg_add_src;
	rcr_vertical <= reg_rcr_vertical(0);
	rcr_offset <= reg_offset;
	rcr_version <= reg_header(header_size - version_width); 
	
	rcw_rc <= rcw_d2_rc;
	rcw_pattern <= rcw_d2_pattern;
	rcw_addr <= rcw_d2_addr;
	rcw_vertical <= rcw_d2_vertical;
	rcw_offset <= rcw_d2_offset;
	rcw_version <= rcw_d2_version;
	
	flags <= reg_header(reg_header'left - version_width downto reg_header'left+1 -version_width - flags_width);
	
	status(0) <= reg_conf;
end architecture;