library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common_defs.all;
use work.version2_defs.all;

entity rc_reader is
	generic (
		config_width	: positive	:= 3;
		config_size	: positive	:= 57;
		address_width	: positive	:= 8;
		pattern_width	: positive	:= 4;
		base_config_size: positive	:= 38;
		dummy_config	: std_logic_vector := "00000000000000000000100000000000000000"
	);
	port (
		data_clk	: in std_logic;
		config_clk	: in std_logic;
		reset		: in std_logic;
		
		-- Interface
		reconfigure	: in std_logic;
		load_address	: in std_logic_vector(address_width -1 downto 0);
		rc_pattern	: in std_logic_vector(pattern_width -1 downto 0);
		vertical	: in std_logic;
		unit_offset	: in std_logic_vector(address_width -1 downto 0);
		version		: in std_logic;
		
		-- To next unit
		rc_next		: out std_logic;
		address_next	: out std_logic_vector(address_width -1 downto 0);
		rc_pattern_next : out std_logic_vector(pattern_width -1 downto 0);
		vertical_next	: out std_logic;
		unit_offset_next: out std_logic_vector(address_width -1 downto 0);
		version_next	: out std_logic;
		
		-- Memory I/O
		read_address	: out std_logic_vector(address_width -1 downto 0);
		read_data	: in std_logic_vector(base_config_size -1 downto 0);
		
		reconf_out	: out std_logic_vector(pattern_width -1 downto 0);
		config_out	: out std_logic_vector(config_width -1 downto 0)
	);
end entity;

architecture second of rc_reader is
	signal reg_pattern	: std_logic_vector(pattern_width -1 downto 0);
	signal reg_vertical	: std_logic;
	signal reg_base_address	: std_logic_vector(address_width -1 downto 0);
	signal reg_rc		: std_logic;
	signal reg_unit_offset	: std_logic_vector(address_width -1 downto 0);
	signal reg_version	: std_logic;
	
	signal config_bus_in	: std_logic_vector(config_width -1 downto 0);
	
	--signal en_update_v1	: std_logic;
	signal en_update	: std_logic;
	signal en_shift		: std_logic;
	signal en_dummy		: std_logic;
	signal reg_v1_config	: std_logic_vector(base_config_size-1 downto 0);
	signal reg_v2_config	: std_logic_vector(config_size-base_config_size-1 downto 0);
	signal full_config	: std_logic_vector(config_size -1 downto 0);
	signal next_v1_config	: std_logic_vector(base_config_size-1 downto 0);
	signal next_v2_config	: std_logic_vector(base_config_size-1 downto 0);
	
	signal en_address	: std_logic;
	signal reg_address	: std_logic_vector(address_width-1 downto 0);
	signal next_address	: std_logic_vector(address_width -1 downto 0);
	
	type memory_states	is (idle, reconfiguring, mem_access, dummy,loaded, sleep);
	signal mem_state	: memory_states;
	signal next_mem_state	: memory_states;
	signal exit_counter	: integer range 0 to (config_size/config_width)-1;
	signal reconf_shifter	: std_logic_vector(pattern_width -1 downto 0);
	signal shifter_input	: std_logic;
	
	signal config_shifter	: std_logic_vector(1 downto 0);
	signal en_cshifter	: std_logic;
	signal cshifter_in	: std_logic;
begin
	
	shift_out: component sc_register 
		generic map(
			width => config_size,
			config_width => config_width,
			allow_prohibit => '0'
		)
		port map(
			config_clk => config_clk,
			data_clk => data_clk,
			reset => reset,
			config_in => config_bus_in,
			config_out => config_out,
			config_enable => '1', -- ?
			-- data_out not connected.. don't care
			data_in => full_config,
			write => '1'
		);
	config_bus_in <= (others => '0');
	
	full_config <= reg_v2_config & reg_v1_config;
	
	next_v2_config <= read_data when (reg_version = '1' and en_dummy = '0') else (others => '0');
	next_v1_config <= read_data when en_dummy = '0' else dummy_config;
	
	ctrl_clk : process(data_clk, reset)
	begin
		if reset = '0' then
			reg_pattern <= (others => '0');
			reg_base_address <= (others => '0');
			reg_rc <= '0';
			reg_unit_offset <= (others => '0');
			reg_vertical <= '0';
			reg_version <= '0';
		elsif rising_edge(data_clk) then
			reg_pattern <= rc_pattern;
			reg_vertical <= vertical;
			reg_base_address <= load_address;
			reg_rc <= reconfigure;
			reg_unit_offset <= unit_offset;
			reg_version <= version;
		end if;
	end process;
	
	perform_clk : process(config_clk, reset) 
	begin
		if reset = '0' then
			mem_state <= idle;
			exit_counter <= 0;
			reconf_shifter <= (others => '0');
			--reg_config <= (others => '0');
			reg_v1_config <= (others => '0');
			reg_v2_config <= (others => '0');
		elsif rising_edge(config_clk) then
			mem_state <= next_mem_state;
			if en_update = '1' then
				if config_shifter(0) = '1' then
					reg_v1_config <= next_v1_config;
				elsif config_shifter(1) = '1' then
					reg_v2_config <= next_v2_config(config_size-base_config_size-1 downto 0);
				end if;
			end if;
			if en_shift = '1' then
				reconf_shifter(reconf_shifter'left downto 1) <= reconf_shifter(reconf_shifter'left -1 downto 0);
				reconf_shifter(0) <= shifter_input;
			end if;
			if en_address = '1' then
				reg_address <= next_address;
			end if;
			
			if en_cshifter = '1' then
				config_shifter(config_shifter'left downto 1) <= config_shifter(config_shifter'left -1 downto 0);
				config_shifter(0) <= cshifter_in;
			end if;
			
			if exit_counter = ((config_size/config_width) -1) or reg_rc = '0' then
				exit_counter <= 0;
			else
				exit_counter <= exit_counter + 1;
			end if;
		end if;
	end process;
	
	mem_state_next : process(mem_state,reg_pattern,reg_vertical,reg_base_address,read_data,reg_rc,exit_counter,reconf_shifter,reg_address)
	begin
		read_address <= (others => '0');
		en_update <= '0';
		en_shift <= '0';
		shifter_input <= '0';
		next_address <= (others => '0');
		en_address <= '0';
		en_dummy <= '0';
		en_cshifter <= '0';
		cshifter_in <= '0';
		case mem_state is
		when idle => 
			-- Yes are using signals from a different clock domain here. But the clocks 
			-- should be in sync so this _should_ be okay without syncronization
			en_shift <= '1';
			en_cshifter <= '1';
			if reg_rc = '1' then 
				shifter_input <= '1';
				cshifter_in <= '1';
				en_address <= '1';
				next_address <= reg_base_address;
				if reg_pattern(0) = '0' or reg_vertical = '0' then
					next_mem_state <= dummy;
				else 
					next_mem_state <= mem_access;
				end if;
			else 
				--en_update <= '1';
				shifter_input <= '0';
				next_mem_state <= idle;
			end if;
		
		when reconfiguring => 
			en_shift <= '1';
			en_cshifter <= '1';
			cshifter_in <= '1';
			
			if reg_pattern(0) = '0' or reg_vertical = '0' then
				next_mem_state <= dummy;
			else 
				next_mem_state <= mem_access;
			end if;
		
		
		when dummy =>
			en_update <= '1';
			-- Dummy config
			en_dummy <= '1';
			en_cshifter <= '1';
			if config_shifter(config_shifter'high) = '1' then
				next_mem_state <= sleep;
			else 
				next_mem_state <= dummy;
			end if;
		
		when mem_access =>
			-- Put address on bus 
			read_address <= reg_address;
			next_mem_state <= loaded;
		
		when loaded => 
			-- Read configuration from bus
			en_update <= '1';
			en_address <= '1';
			next_address <= std_logic_vector(unsigned(reg_address) + 1);
			en_cshifter <= '1';
			if config_shifter(config_shifter'high) = '1' then
				next_mem_state <= sleep;
			elsif reg_version = '0' then 
				next_mem_state <= dummy;
			else
				next_mem_state <= mem_access;
			end if;
		
		when sleep => 
			if exit_counter = ((config_size/config_width) -1) then
				if reconf_shifter(reconf_shifter'left) = '1' then
					next_mem_state <= idle;
				else
					next_mem_state <= reconfiguring;
				end if;
			else 
				next_mem_state <= sleep;
			end if;
		
		end case;
	end process;
	
	-- Rotate pattern to next unit
	rc_pattern_next(reg_pattern'left) <= reg_pattern(reg_pattern'right);
	rc_pattern_next(reg_pattern'left -1 downto 0) <= reg_pattern(reg_pattern'left downto 1);
	
	rc_next <= reg_rc;
	vertical_next <= reg_vertical;
	
	address_next <= std_logic_vector(unsigned(reg_base_address)+unsigned(reg_unit_offset)) when reg_pattern(0) = '1' else 
			reg_base_address;
	
	unit_offset_next <= reg_unit_offset;
	
	version_next <= reg_version;
	
	reconf_out <= reconf_shifter;
end architecture;
