library ieee;
use ieee.std_logic_1164.all;
use work.common_defs.all;

entity sc_register is
	generic(
		width		: positive := 8;
		config_width	: positive := 2;
		allow_prohibit	: std_logic := '1'
	);
	port(
		config_clk	: in std_logic;
		data_clk	: in std_logic;
		reset		: in std_logic;
		
		config_in	: in std_logic_vector(config_width -1 downto 0);
		config_out	: out std_logic_vector(config_width -1 downto 0);
		config_enable	: in std_logic;
		
		data_in		: in std_logic_vector(width -1 downto 0);
		data_out	: out std_logic_vector(width -1 downto 0);
		write		: in std_logic
	);
end entity;

architecture shifter of sc_register is 
	signal store		:std_logic_vector(width -1 downto 0);
	signal write_pulse	:std_logic;
	signal write_prohibit	:std_logic;
	signal in_register	:std_logic_vector(width -1 downto 0);
	signal write_buffer	:std_logic;
	signal anti_clk		:std_logic;
	
begin
	anti_clk <= (not data_clk) and config_clk;
	
	write_pulse <= ((data_clk and config_clk) and write_buffer) and (not (write_prohibit and allow_prohibit));
	
	clocked_action : process(write_pulse,config_clk,in_register,reset)
	begin
		if reset = '0' then
			store <= (store'range => '0');
			write_prohibit <= '0';
			in_register <= (in_register'range => '0');
		elsif write_pulse = '1' then
			store <= in_register;
		elsif rising_edge(config_clk) then
			-- It doesn't hurt that the config_out bus 
			-- value changes when config_enable is deasserted
			-- Config enable overrides write
			
			-- The data clock must be fase-shifted for this to work.
			-- This might break in any number of ways.. Clock data across 
			-- domains without any sync-circuits.
			if config_enable = '1' then
				--config_out <= store(config_width -1 downto 0);
				store <= config_in & store(width -1 downto config_width);
			end if;
			in_register <= data_in;
			-- Nuke write_pulse on end
			write_prohibit <= config_enable;
		end if;
	end process;
	data_out <= store;
	config_out <= store(config_width -1 downto 0);
	
	-- Lock down write input signal during async load
	-- Clear when data clk is not asserted
	buffer_action : process(data_clk,anti_clk,reset) 
	begin
		if reset = '0' then
			write_buffer <= '0';
		elsif anti_clk = '1' then
			write_buffer <= '0';
		elsif rising_edge(data_clk) then
			write_buffer <= write;
		end if;
	end process;
	
end architecture;