library ieee;
use ieee.std_logic_1164.all;
use ieee.math_real.all;

package version2_defs is 
	-- Cell specific defs
	constant config_width	: positive := 3;
	constant config_size	: positive := 57; -- padded with two bits
	constant base_c_size	: positive := 38;
	
	-- RC Master defs
	constant rcm_status_size: positive	:= 3;
	constant rcm_ctrl_size	: positive	:= 3;
	
	component cell is
	generic (
		width		: positive := 16;
		config_width	: positive := 3
	);
	port (
		config_clk	: in std_logic;
		data_clk	: in std_logic;
		compute_clk	: in std_logic;
		
		reset		: in std_logic;
		
		reconfigure	: in std_logic;
		
		config_in	: in std_logic_vector(config_width -1 downto 0);
		config_out	: out std_logic_vector(config_width -1 downto 0);
		
		north_idle	: in std_logic;
		west_idle	: in std_logic;
		south_idle	: out std_logic;
		east_idle	: out std_logic;
		
		-- With room for valid signal
		north		: in std_logic_vector(width downto 0);
		west		: in std_logic_vector(width downto 0);
		
		south		: out std_logic_vector(width downto 0);
		east		: out std_logic_vector(width downto 0)
		
	);
	end component;
	
	component rc_reader is
	generic (
		config_width	: positive	:= 3;
		config_size	: positive	:= 57;
		address_width	: positive	:= 8;
		pattern_width	: positive	:= 4;
		base_config_size: positive	:= 38;
		dummy_config	: std_logic_vector := "00000000000000000000100000000000000000"
	);
	port (
		data_clk	: in std_logic;
		config_clk	: in std_logic;
		reset		: in std_logic;
		
		-- Interface
		reconfigure	: in std_logic;
		load_address	: in std_logic_vector(address_width -1 downto 0);
		rc_pattern	: in std_logic_vector(pattern_width -1 downto 0);
		vertical	: in std_logic;
		unit_offset	: in std_logic_vector(address_width -1 downto 0);
		version		: in std_logic;
		
		-- To next unit
		rc_next		: out std_logic;
		address_next	: out std_logic_vector(address_width -1 downto 0);
		rc_pattern_next : out std_logic_vector(pattern_width -1 downto 0);
		vertical_next	: out std_logic;
		unit_offset_next: out std_logic_vector(address_width -1 downto 0);
		version_next	: out std_logic;
		
		-- Memory I/O
		read_address	: out std_logic_vector(address_width -1 downto 0);
		read_data	: in std_logic_vector(base_config_size -1 downto 0);
		
		reconf_out	: out std_logic_vector(pattern_width -1 downto 0);
		config_out	: out std_logic_vector(config_width -1 downto 0)
	);
	end component;
	
	component rc_writer is
	generic (
		config_width	: positive	:= 2;
		config_size	: positive	:= 57;
		address_width	: positive	:= 8;
		pattern_width	: positive	:= 4;
		base_config_size: positive	:= 38
	);
	port (
		data_clk	: in std_logic;
		config_clk	: in std_logic;
		reset		: in std_logic;
		
		-- Interface
		reconfigure	: in std_logic;
		store_address	: in std_logic_vector(address_width -1 downto 0);
		rc_pattern	: in std_logic_vector(pattern_width -1 downto 0);
		vertical	: in std_logic;
		unit_offset	: in std_logic_vector(address_width -1 downto 0);
		version		: in std_logic;
		
		-- To next unit
		rc_next		: out std_logic;
		address_next	: out std_logic_vector(address_width -1 downto 0);
		rc_pattern_next : out std_logic_vector(pattern_width -1 downto 0);
		vertical_next	: out std_logic;
		unit_offset_next: out std_logic_vector(address_width -1 downto 0);
		version_next	: out std_logic;
		
		-- Memory I/O
		write_address	: out std_logic_vector(address_width -1 downto 0);
		write_data	: out std_logic_vector(base_config_size -1 downto 0);
		write_enable	: out std_logic;
		
		config_in	: in std_logic_vector(config_width -1 downto 0)
	);
	end component;
	
	component rc_master is 
	generic (
		addr_width	:positive	:= 8;
		header_size	:positive	:= 20;
		rc_units	:positive	:= 4;
		system_version	:natural	:= 1;
		version_width	:positive	:= 4;
		flags_width	:positive	:= 4;
		size_width	:positive	:= 8;
		addr_consume	:integer	:= 3
	);
	port (
		clk		:in std_logic;
		reset		:in std_logic;
		
		-- Interface
		status		:out std_logic_vector(rcm_status_size-1 downto 0);
		ctrl		:in std_logic_vector(rcm_ctrl_size-1 downto 0);
		src_base_addr	:in std_logic_vector(addr_width -1 downto 0);
		dest_base_addr	:in std_logic_vector(addr_width -1 downto 0);
		
		-- General
		flags		:out std_logic_vector(flags_width -1 downto 0);
		
		-- For configuration headers
		read_addr	:out std_logic_vector(addr_width-1 downto 0);
		read_data	:in std_logic_vector(header_size -1 downto 0);
		write_en	:out std_logic;
		write_addr	:out std_logic_vector(addr_width-1 downto 0);
		write_data	:out std_logic_vector(header_size -1 downto 0);
		
		-- Read units
		rcr_rc		:out std_logic;
		rcr_addr	:out std_logic_vector(addr_width -1 downto 0);
		rcr_pattern	:out std_logic_vector(rc_units-1 downto 0);
		rcr_vertical	:out std_logic;
		rcr_offset	:out std_logic_vector(addr_width -1 downto 0);
		rcr_version	:out std_logic;
		
		-- Write units
		rcw_rc		:out std_logic;
		rcw_addr	:out std_logic_vector(addr_width -1 downto 0);
		rcw_pattern	:out std_logic_vector(rc_units-1 downto 0);
		rcw_vertical	:out std_logic;
		rcw_offset	:out std_logic_vector(addr_width -1 downto 0);
		rcw_version	:out std_logic
	);
	end component;
	
	constant io_size	: positive	:= 16;
	type io_bank		is array(natural range<>) of std_logic_vector(io_size downto 0);
	
	component wrpu is 
	generic(
		data_width	: positive	:= io_size;
		addr_width	: positive	:= 8;
		config_width	: positive	:= config_width;
		config_size	: positive	:= config_size;
		base_config_size: positive	:= base_c_size;
		header_size	: positive	:= 24;
		cells_axis	: natural	:= 4
	);
	port (
		data_clk	: in std_logic;
		config_clk	: in std_logic;
		compute_clk	: in std_logic;
		
		reset		: in std_logic;
		
		rc_ctrl		: in std_logic_vector(rcm_ctrl_size-1 downto 0);
		rc_status	: out std_logic_vector(rcm_status_size-1 downto 0);
		src_base_addr	: in std_logic_vector(addr_width-1 downto 0);
		dest_base_addr	: in std_logic_vector(addr_width-1 downto 0);
		
		north_bank	: in io_bank(cells_axis-1 downto 0);
		west_bank	: in io_bank(cells_axis-1 downto 0);
		
		south_bank	: out io_bank(cells_axis-1 downto 0);
		east_bank	: out io_bank(cells_axis-1 downto 0)
	);
	end component;
	
end package;
